import ujson
from os.path import isabs
import edge.eventlog as eventlog

import edge.db as edge_db


class ArgsError(RuntimeError):
    pass


class DbArgs:
    def __init__(self, db_host, db_port, db_name, db_user, db_pass, db_link_min=1, db_link_max=1):
        """
        :type db_host: str
        :type db_port: int
        :type db_name: str
        :type db_user: str
        :type db_pass: str
        :type db_link_min: int
        :type db_link_max: int
        """
        self.db_host = db_host
        self.db_port = db_port
        self.db_name = db_name
        self.db_user = db_user
        self.db_pass = db_pass

        self.db_link_min = db_link_min
        self.db_link_max = db_link_max

    @staticmethod
    def deserialize(db_jnode):
        """
        :type db_jnode: dict
        :rtype: DbArgs
        :raise: ArgsError
        """
        try:
            args = edge_db.DbArgs.deserialize(db_jnode)

        except KeyError as x:
            eventlog.log_debug('fail: param #/%s has invalid value' % x.message)
            raise ArgsError

        return DbArgs(args.db_host, args.db_port, args.db_name, args.db_user, args.db_pass)


class ReactProducerArgs:
    def __init__(self, endpoint):
        """
        :type endpoint: str
        """
        self.endpoint = endpoint

    @staticmethod
    def deserialize(endpoint):
        """
        :type endpoint: str
        :rtype: ReactProducerArgs
        """
        return ReactProducerArgs(endpoint)


class EventConsumerArgs:
    MODE_DEFAULT = 1
    MODE_WITH_JSON = 2

    def __init__(self, mode, udp_peer_host, udp_peer_port, post_ident, chan_ident):
        """
        :type mode: int
        :type udp_peer_host: str
        :type udp_peer_port: int
        :type post_ident: str
        :type chan_ident: int
        """
        self.mode = mode
        self.udp_peer_host = udp_peer_host
        self.udp_peer_port = udp_peer_port

        self.post_ident = post_ident
        self.chan_ident = chan_ident

    @staticmethod
    def deserialize(fact_event_jnode, post_ident, chan_ident):
        """
        :type fact_event_jnode: (int, str, int)
        :type post_ident: str
        :type chan_ident: int
        :rtype: EventConsumerArgs
        """
        if len(fact_event_jnode) != 3:
            eventlog.log_debug('fail: param #/ has invalid value')
            raise ArgsError

        mode, udp_peer_host, udp_peer_port = fact_event_jnode
        if mode not in [EventConsumerArgs.MODE_DEFAULT, EventConsumerArgs.MODE_WITH_JSON]:
            eventlog.log_debug('fail: param #/[0] has invalid value (mode=%s)' % mode)
            raise ArgsError

        return EventConsumerArgs(mode, udp_peer_host, udp_peer_port, post_ident, chan_ident)


class CountersArgs:
    def __init__(self, stat_filepath, stat_tout_sec):
        """
        :type stat_filepath: str
        :type stat_tout_sec: int
        """
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @staticmethod
    def deserialize(monitoring_jnode):
        """
        :type monitoring_jnode: dict
        :rtype: CountersArgs
        :raise: ArgsError
        """
        stat_filepath = monitoring_jnode.get('stat-path')
        if not (unicode == type(stat_filepath) and len(stat_filepath) < 64 and isabs(stat_filepath)):
            eventlog.log_debug('fail: param #/stat-path has invalid value')
            raise ArgsError

        stat_tout_sec = monitoring_jnode.get('stat-tout')
        if not (int == type(stat_tout_sec) and 0 <= stat_tout_sec <= 255):
            eventlog.log_debug('fail: param #/stat-tout has invalid value')
            raise ArgsError

        return CountersArgs(stat_filepath, stat_tout_sec)


class KernelArgs:
    def __init__(self, db_args, db_post, db_chan, react_producer, event_consumer, counters):
        """
        :type db_args: DbArgs
        :type db_post: edge_db.DbPostArgs
        :type db_chan: edge_db.DbChannelArgs
        :type react_producer: ReactProducerArgs
        :type event_consumer: EventConsumerArgs | None
        :type counters: CountersArgs
        """
        self.db = db_args
        self.db_post = db_post
        self.db_chan = db_chan

        self.react_producer = react_producer
        self.event_consumer = event_consumer
        self.counters = counters

    @staticmethod
    def deserialize(main_conf_jnode):
        """
        :type main_conf_jnode: dict
        :rtype: KernelArgs
        :raise: ArgsError
        """
        # load base conf
        db_conf_abspath = main_conf_jnode.get('base')
        if not db_conf_abspath:
            eventlog.log_error('fail: param #base has invalid value')
            raise ArgsError

        try:
            with open(db_conf_abspath) as base_conf:
                db_conf_jnode = ujson.load(base_conf)

        except Exception as x:
            eventlog.log_error('fail: load %s (%s)' % (db_conf_abspath, x))
            raise ArgsError

        # deserialize db-args
        db_args_jnode = db_conf_jnode.get('db')
        if not db_args_jnode:
            eventlog.log_error('fail: param #/base has invalid value (#/db is absent)')
            raise ArgsError

        db = DbArgs.deserialize(db_args_jnode)

        # deserialize db-post-args
        db_post_jnode = db_conf_jnode.get('post')
        if not db_post_jnode:
            eventlog.log_error('fail: param #/base has invalid value (#/post is absent)')
            raise ArgsError

        db_post = edge_db.DbPostArgs.deserialize(db_post_jnode)

        # deserialize db-chan-args
        db_chan_jnode = main_conf_jnode['store']
        if not db_chan_jnode:
            eventlog.log_error('fail: param #/store has invalid value')
            raise ArgsError

        db_chan = edge_db.DbChannelArgs.deserialzie(db_chan_jnode)

        # deserialize react-producer-args
        fact_react_jnode = main_conf_jnode.get('fact-react')
        if not fact_react_jnode:
            eventlog.log_error('fail: param #/fact-react has invalid value')
            raise ArgsError

        react_producer = ReactProducerArgs.deserialize(fact_react_jnode)

        # deserialize event-consumer-args
        socket_api_jnode = main_conf_jnode.get('fact-event')
        if socket_api_jnode:
            event_consumer = EventConsumerArgs.deserialize(
                socket_api_jnode, db_post.ident, db_chan.ident)
        else:
            event_consumer = None

        # deserialize counters-args
        counters_jnode = main_conf_jnode.get('monitoring')
        if not counters_jnode:
            eventlog.log_error('fail: param #/monitoring has invalid value')
            raise ArgsError

        counters = CountersArgs.deserialize(counters_jnode)

        kernel_args = KernelArgs(db, db_post, db_chan, react_producer, event_consumer, counters)
        return kernel_args
