import logging
from logging.handlers import SysLogHandler

from edge.bin_fs import binfs
import edge.eventlog as eventlog

from _models import FactImages


class BlobStoreError(RuntimeError):
    pass


class BlobStore:
    def __init__(self, plate_fs, vehicle_fs, thumb_fs, scene_fs, scene_alt_fs):
        """
        :type plate_fs: binfs
        :type vehicle_fs: binfs
        :type thumb_fs: binfs
        :type scene_fs: binfs
        :type scene_alt_fs: binfs
        """
        self._plate_fs = plate_fs

        self._vehicle_fs = vehicle_fs
        self._thumb_fs = thumb_fs

        self._scene_fs = scene_fs
        self._scene_alt_fs = scene_alt_fs

    @staticmethod
    def deserialize(im_profiles_jnode):
        """
        :type im_profiles_jnode: list of dict
        :rtype: BlobStore
        :raise: ValueError
        """
        def mk_binfs(name, root):
            """
            :type name: str
            :type root: str
            :rtype: binfs
            """
            fs = binfs(name=name, root=root, oCreate=True, setGID='postgres', setUID='postgres')

            py_logger = logging.getLogger().getChild('binfs').getChild(name)
            py_logger.setLevel(logging.INFO)
            py_logger.addHandler(SysLogHandler(address='/dev/log', facility=SysLogHandler.LOG_DAEMON))
            fs.setlogger(py_logger)

            return fs

        body_fs = scene_fs = plate_fs = body_thumb_fs = scene_alt_fs = None

        for im_profile_jnode in im_profiles_jnode:
            root_path = im_profile_jnode['root']
            profile_name = im_profile_jnode['image']

            if profile_name == 'plate':
                plate_fs = mk_binfs(name=profile_name, root=root_path)

            elif profile_name == 'body':
                body_thumb_fs = mk_binfs(name=profile_name, root=root_path)

            elif profile_name == 'stampedbody':
                body_fs = mk_binfs(name=profile_name, root=root_path)

            elif profile_name == 'frame':
                scene_fs = mk_binfs(name=profile_name, root=root_path)

            elif profile_name == 'stampedframe':
                scene_alt_fs = mk_binfs(name=profile_name, root=root_path)

        if None in [plate_fs, body_fs,  body_thumb_fs, scene_fs, scene_alt_fs]:

            eventlog.log_debug('invalid im-profile (%s)' % map(
                lambda fs: fs.name if fs else None,
                [plate_fs, body_fs,  body_thumb_fs, scene_fs, scene_alt_fs]))

            raise BlobStoreError

        return BlobStore(plate_fs, body_fs, body_thumb_fs, scene_fs, scene_alt_fs)

    def save(self, blob_ident, images):
        """
        :type blob_ident: int
        :type images: FactImages
        :raise BlobStoreError:
        """
        def write_blob(bin_fs, blob_data, blob_meta):
            """
            :type bin_fs: binfs
            :type blob_data: str | None
            :type blob_meta: str | None
            :raise BlobStoreError:
            """
            # write blob-data
            size = 0 if blob_data is None else len(blob_data)
            if size > 0:
                if size != bin_fs.save(blob_ident, blob_data):
                    eventlog.log_debug('fail: write blob-data to the binfs')
                    raise BlobStoreError

                # write blob-meta
                size = 0 if blob_meta is None else len(blob_meta)
                if size > 0:
                    if size != bin_fs.meta_set(blob_ident, blob_meta):
                        eventlog.log_debug('fail: write blob-meta to the binfs')
                        raise BlobStoreError

        #
        write_blob(self._plate_fs, images.plate_blob, images.plate_meta)
        write_blob(self._thumb_fs, images.body_thumb_blob, images.body_thumb_meta)
        write_blob(self._vehicle_fs, images.body_blob, images.body_meta)
        write_blob(self._scene_fs, images.scene_blob, images.scene_meta)
        write_blob(self._scene_alt_fs, images.scene_alt_blob, images.scene_alt_blob)
