import time
from ujson import dump as ujson_dump
from os import rename

import edge.eventlog as eventlog

from _args import CountersArgs
from _models import Fact


class CountersError(RuntimeError):
    pass


class DbCounters:
    """
    :type link_present: long
    :type link_absent: long
    """
    def __init__(self):
        self.link_present = long()
        self.link_absent = long()

    def inc_link_present(self):
        self.link_present += 1

    def inc_link_absent(self):
        self.link_absent += 1


class ReactProducerCounters:
    def __init__(self):
        self.fact_count = long()
        self.speed_count = long()
        self.plate_count = long()

    def inc_fact_count(self, fact):
        """
        :type fact:Fact 
        """
        self.fact_count += 1
        if fact.vehicle.plate is not None:
            self.plate_count += 1

        if fact.vehicle.speed is not None:
            self.speed_count += 1


class EventConsumerCounters:
    def __init__(self):
        self.event_count = 0

    def inc_event_count(self):
        self.event_count += 1


class StatCounters:
    """
    :type db: DbCounters
    :type react_producer: ReactProducerCounters
    :type event_consumer: EventConsumerCounters | None
    """
    def __init__(self, twisted_reactor, args, stat_event_counsumer):
        """
        :type args: CountersArgs
        :type stat_event_counsumer: bool
        """
        self.db = DbCounters()
        self.react_producer = ReactProducerCounters()
        self.event_consumer = EventConsumerCounters() if stat_event_counsumer else None

        if args.stat_tout_sec:
            twisted_reactor.callLater(
                args.stat_tout_sec, self._log_stat_counters,
                twisted_reactor, args.stat_filepath, args.stat_tout_sec)

    def _log_stat_counters(self, twisted_reactor, stat_filepath, stat_tout_sec):
        """
        :type stat_filepath: str
        :type stat_tout_sec: int
        """
        # noinspection PyBroadException
        try:
            self._serialize(stat_filepath, stat_tout_sec)
        except CountersError:
            pass
        except:
            eventlog.log_debug_last_exc()
        finally:
            twisted_reactor.callLater(
                stat_tout_sec, self._log_stat_counters,
                twisted_reactor, stat_filepath, stat_tout_sec)

    def _serialize(self, stat_filepath, stat_tout_sec):
        """
        :type stat_filepath: str
        :type stat_tout_sec: int
        :raises: CountersError
        """
        sys_utc_time = time.time()

        stat = {
            "stat": {
                "mtime": [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                "tout-sec": stat_tout_sec
            },
            "db": {
                "link-present": self.db.link_present,
                "link-absent": self.db.link_absent,
            },
            "react": {
                "count": self.react_producer.fact_count,
                "vehicle": {
                    "plate": self.react_producer.plate_count,
                    "speed": self.react_producer.speed_count
                }
            }
        }

        if self.event_consumer is not None:
            stat['event'] = {
                "count": self.event_consumer.event_count
            }

        try:
            tmp_filepath = stat_filepath + '.tmp'
            with open(tmp_filepath, 'w') as json_file:
                # Use ensure_ascii=false to output UTF-8.
                ujson_dump(stat, json_file, ensure_ascii=False)

            rename(tmp_filepath, stat_filepath)

        except IOError as x:
            eventlog.log_warning('fail: dump stat-counters to %s (%s)' % (stat_filepath, x))
            raise CountersError
