import psycopg2
import psycopg2.pool
import psycopg2.extras

import edge.eventlog as eventlog
import edge.db as edge_db

from _args import DbArgs
from _counters import DbCounters
from _models import Fact
from _blob_store import BlobStore, BlobStoreError


class Db:
    """
    :type _blob_store: BlobStore | None
    """
    def __init__(self, stat_counters, link_pool, post_args, chan_args):
        """
        :type stat_counters: DbCounters
        :type link_pool: psycopg2.pool.SimpleConnectionPool
        :type post_args: edge_db.DbPostArgs
        :type chan_args: edge_db.DbChannelArgs
        """
        self._counters = stat_counters
        self._link_pool = link_pool

        self._post_args = post_args
        self._chan_args = chan_args

        self._blob_store = None

    @staticmethod
    def create(twisted_reactor, stat_counters, db_args, post_args, channel_args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorCore
        :type stat_counters: DbCounters
        :type db_args: DbArgs
        :type post_args: edge_db.DbPostArgs
        :type channel_args: edge_db.DbChannelArgs
        :rtype: Db
        """
        link_pool = psycopg2.pool.SimpleConnectionPool(
            db_args.db_link_min,
            db_args.db_link_max,
            user=db_args.db_user,
            password=db_args.db_pass,
            database=db_args.db_name,
            host=db_args.db_host,
            port=db_args.db_port)

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', link_pool.closeall)
        return Db(stat_counters, link_pool, post_args, channel_args)

    def _add_fact_args(self, db_link, fact):
        """
        :type fact: Fact
        :rtype: (int, int)
        :return: Tuple (platear-id, media-id)
        :raises: db.DbError
        """
        api_query_jnode = {
            '@header': {
                'vendor': 'VIT',
                'class': 'vitml.edge.api',
                'version': '1.0',
                'object': 'plate',
                'action': 'add'
            },
            'post': self._post_args.ident,
            'channel': self._chan_args.ident,

            'guid': fact.uuid,
            'ctime': fact.best_ts,
            'stime': fact.catch_ts,
            'etime': fact.lost_ts,

            'vehicle_length': fact.vehicle.length,
            'direction': fact.vehicle.direction
        }

        if fact.vehicle.plate:
            api_query_jnode.update({
                'rect': {
                    'left': fact.vehicle.plate.rect.left,
                    'top': fact.vehicle.plate.rect.top,
                    'right': fact.vehicle.plate.rect.right,
                    'bottom': fact.vehicle.plate.rect.bottom
                },
                'number': fact.vehicle.plate.text,
                'country': fact.vehicle.plate.template.country,
                'county': fact.vehicle.plate.template.region,
                'numtype': fact.vehicle.plate.template.index,
                'validity': fact.vehicle.plate.validity
            })

        if fact.vehicle.speed is not None:
            api_query_jnode['velocity'] = fact.vehicle.speed

            if fact.speed_limit is not None:
                api_query_jnode['overspeed'] = fact.vehicle.speed - fact.speed_limit
            else:
                api_query_jnode['overspeed'] = 0

        if fact.vehicle.lane is not None:
            api_query_jnode['lane'] = fact.vehicle.lane

        if fact.geolocation:
            api_query_jnode.update({
                'geodata': {
                    'latitude': fact.geolocation.latitude,
                    'longitude': fact.geolocation.longitude,
                    'altitude': fact.geolocation.altitude,
                    'speed': fact.geolocation.speed
                }
            })

        with db_link.cursor() as db_cursor:
            db_cursor.execute("SELECT api_plate_add(%s)", [edge_db.DbJsonArg(api_query_jnode)])

            api_reply_jnode = db_cursor.fetchone()[0]
            if int(api_reply_jnode['errcode']) == 0:
                platear_id = api_reply_jnode['id']
                scroll_id = api_reply_jnode['scrollid']
                return platear_id, scroll_id

            else:
                eventlog.log_debug('fail: api_plate_add (code=%s, text=<%s>)'
                                   % (api_reply_jnode['errcode'], api_reply_jnode['errmsg']))
                raise edge_db.DbError

    def _get_im_profiles(self):
        """
        :rtype: list of dict
        :raises db.DbError
        """
        api_query_jnode = {
            '@header': {
                'vendor': 'VIT',
                'class': 'vitml.edge.api',
                'version': '1.0',
                'object': 'mediastore',
                'action': 'discovery'
            },
            'post': self._post_args.ident,
            'channel': self._chan_args.ident
        }

        try:
            db_link = self._link_pool.getconn()
            self._counters.inc_link_present()
            try:
                with db_link, db_link.cursor() as db_cursor:
                    db_cursor.execute("SELECT * from api_discovery(%s)", [edge_db.DbJsonArg(api_query_jnode)])
                    api_reply_jnode = db_cursor.fetchone()[0]

            finally:
                self._link_pool.putconn(db_link)

        except psycopg2.pool.PoolError as x:
            eventlog.log_debug('fail: define db-link (%s)' % x)
            self._counters.inc_link_absent()
            raise edge_db.DbError

        if int(api_reply_jnode['errcode']) != 0:
            eventlog.log_debug(
                'fail: api_discovery (code=%s, text=<%s>)'
                % (api_reply_jnode['errcode'], api_reply_jnode['errmsg']))
            raise edge_db.DbError

        return api_reply_jnode['mediastore']

    def add_fact(self, fact):
        """
        :type fact: Fact
        :rtype: int, int
        :return: Tuple (platear-id, media-id)
        :raises: db.DbError
        """
        if not self._blob_store:
            im_profiles_jnode = self._get_im_profiles()
            try:
                self._blob_store = BlobStore.deserialize(im_profiles_jnode)
            except BlobStoreError:
                eventlog.log_debug('fail: mk blob-store')
                raise edge_db.DbError

        try:
            db_link = self._link_pool.getconn()
            self._counters.inc_link_present()
            try:
                with db_link:
                    platear_id, scroll_id = self._add_fact_args(db_link, fact)

                    try:
                        self._blob_store.save(blob_ident=scroll_id, images=fact.images)

                    except BlobStoreError:
                        eventlog.log_debug('fail: save images')
                        raise edge_db.DbError

                    return platear_id, scroll_id

            finally:
                self._link_pool.putconn(db_link)

        except psycopg2.pool.PoolError as x:
            eventlog.log_debug('fail: define db-link (%s)' % x)
            self._counters.inc_link_absent()
            raise edge_db.DbError
