import socket

from _args import EventConsumerArgs
from _models import Fact
from _counters import EventConsumerCounters
import base64
import ujson


class IEventConsumer:
    def __init__(self):
        pass

    def consume_fact(self, fact, platear_id, media_id):
        """
        :type fact: Fact
        :type platear_id: int
        :type media_id: int
        """
        raise NotImplemented


class StubEventConsumer(IEventConsumer):
    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwargs):
        IEventConsumer.__init__(self)

    def consume_fact(self, fact, platear_id, media_id):
        """
        :type fact: Fact
        :type platear_id: int
        :type media_id: int
        """
        pass


class EventConsumer(IEventConsumer):
    def __init__(self, twisted_reactor, stat_counters, args):
        """
        :type stat_counters: EventConsumerCounters
        :type args: EventConsumerArgs
        """
        IEventConsumer.__init__(self)
        self._mode = args.mode
        self._counters = stat_counters
        self._udp_peer = (args.udp_peer_host, args.udp_peer_port)
        self._udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        self._post_ident = args.post_ident
        self._channel_ident = args.chan_ident

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._udp_socket.close)

    def consume_fact(self, fact, platear_id, media_id):
        """
        :type fact: Fact
        :type platear_id: int
        :type media_id: int
        """
        # <post-guid>,<post-channel>,<image-ts>,<plate-text>,<media-id>,
        # <speed>,<speed-limit>,<platear-id>,<geodata>,<fact-guid>

        geodata_arg = '-'
        if fact.geolocation:
            lat_char = 'N' if fact.geolocation.latitude >= 0 else 'S'
            long_char = 'E' if fact.geolocation.longitude >= 0 else 'W'

            # geodata format : <latChar><DDD.uuuuuu><longChar><ddd.uuuuuu>
            geodata_arg = '%s%.6f%s%.6f' % (lat_char, abs(fact.geolocation.latitude),
                                            long_char, abs(fact.geolocation.longitude))

        arg_list = [
            self._post_ident,
            self._channel_ident,
            fact.best_ts,
            fact.vehicle.plate.text,

            media_id,
            '-' if fact.vehicle.speed is None else fact.vehicle.speed,
            '-' if fact.speed_limit is None else fact.speed_limit,
            platear_id,
            geodata_arg,
            fact.uuid
        ]

        if self._mode == EventConsumerArgs.MODE_WITH_JSON:
            fact_jnode = {
                "id": platear_id,
                "uuid": fact.uuid,
                "catch-ts": fact.catch_ts,
                "lost-ts": fact.lost_ts,
                "vehicle": {
                    "dir": fact.vehicle.direction
                },
                "image-ref": {
                    "ts": fact.best_ts,
                    "id": media_id
                }
            }

            if fact.vehicle.length:
                fact_jnode['vehicle']['lenght'] = fact.vehicle.length

            if fact.vehicle.lane:
                fact_jnode['vehicle']['lane'] = fact.vehicle.lane

            if fact.vehicle.speed:
                fact_jnode['vehicle']['speed'] = fact.vehicle.speed

            if fact.vehicle.plate:
                template = fact.vehicle.plate.template
                plate_type = [template.country, template.region, template.index]

                rect = fact.vehicle.plate.rect
                plate_rect = [rect.left, rect.top, rect.right, rect.bottom]

                fact_jnode['vehicle']['plate'] = {
                    "rect": plate_rect,
                    "type": plate_type,
                    "text": fact.vehicle.plate.text,
                    "validity": fact.vehicle.plate.validity
                }

            if fact.geolocation:
                fact_jnode['geodata'] = {
                    "latitude": fact.geolocation.latitude,
                    "longitude": fact.geolocation.longitude
                }

                if fact.geolocation.altitude is not None:
                    fact_jnode['geodata']['altitude'] = fact.geolocation.altitude

                if fact.geolocation.speed is not None:
                    fact_jnode['geodata']['speed'] = fact.geolocation.speed

            fact_utf8 = ujson.dumps(fact_jnode, ensure_ascii=False)
            fact_base64 = base64.b64encode(fact_utf8)
            arg_list.append(fact_base64)

        datagram = u','.join(map(unicode, arg_list)).encode('utf-8')
        self._udp_socket.sendto(datagram, self._udp_peer)
        self._counters.inc_event_count()
