import edge.eventlog as eventlog
import ujson


DIR_UNKNOWN = 1
DIR_APPROACHING = 2
DIR_RECEDING = 3


class Geodata:
    def __init__(self, latitude, longitude, altitude, speed):
        """
        :param latitude: Latitude in degrees: +/- signifies North/South
        :type latitude: float
        :param longitude: Longitude in degrees: +/- signifies East/West
        :type longitude: float
        :param altitude: Altitude in meters
        :type altitude: float | None
        :param speed: Speed over ground, meters per second
        :type speed: float | None
        """

        # http://catb.org/gpsd/gpsd_json.html
        self.latitude = latitude
        self.longitude = longitude
        self.altitude = altitude
        self.speed = speed

    @staticmethod
    def deserialize(geolocation_jnode):
        """
        :param geolocation_jnode: list of float
        :rtype: Geodata
        :raise: ValueError
        """
        latitude = None
        longitude = None
        altitude = None
        speed = None

        entry_count = len(geolocation_jnode)
        if entry_count == 2:
            latitude, longitude = geolocation_jnode

        elif entry_count == 3:
            latitude, longitude, altitude = geolocation_jnode

        elif entry_count == 4:
            latitude, longitude, altitude, speed = geolocation_jnode

        else:
            raise ValueError

        return Geodata(latitude, longitude, altitude, speed)


class FactImages:
    def __init__(
            self,
            plate_blob=None,
            plate_meta=None,
            body_thumb_blob=None,
            body_thumb_meta=None,
            body_blob=None,
            body_meta=None,
            scene_blob=None,
            scene_meta=None,
            scene_alt_blob=None,
            scene_alt_meta=None):
        """
        :type plate_blob: str | None
        :type plate_meta: str | None
        :type body_thumb_blob: str | None
        :type body_thumb_meta: str | None
        :type body_blob: str | None
        :type body_meta: str | None
        :type scene_blob: str | None
        :type scene_meta: str | None
        :type scene_alt_blob: str | None
        :type scene_alt_meta: str | None
        """
        self.plate_blob = plate_blob
        self.plate_meta = plate_meta

        self.body_thumb_blob = body_thumb_blob
        self.body_thumb_meta = body_thumb_meta

        self.body_blob = body_blob
        self.body_meta = body_meta

        self.scene_blob = scene_blob
        self.scene_meta = scene_meta

        self.scene_alt_blob = scene_alt_blob
        self.scene_alt_meta = scene_alt_meta

    @staticmethod
    def deserialize(images_jnode):
        """
        :type images_jnode: list of dict
        :rtype: FactImages
        :raise ValueError:
        """
        def read_im(im_path):
            """
            :type im_path: str
            :rtype: str | None
            """
            try:
                with open(im_path, 'r') as im_file:
                    im_bytes = im_file.read()
                    return im_bytes if len(im_bytes) else None

            except IOError as x:
                eventlog.log_debug('fail: read image %s (%s)' % (im_path, x))

        images = FactImages()

        for image_jnode in images_jnode:
            name = image_jnode['name']
            meta = image_jnode.get('meta')

            if name == 'plate':
                images.plate_blob = read_im(image_jnode['blob'])
                if images.plate_blob and meta:
                    images.plate_meta = ujson.dumps(meta)

            elif name == 'body.thumb':
                images.body_thumb_blob = read_im(image_jnode['blob'])
                if images.body_thumb_blob and meta:
                    images.body_thumb_meta = ujson.dumps(meta)

            elif name == 'body':
                images.body_blob = read_im(image_jnode['blob'])
                if images.body_blob and meta:
                    images.body_meta = ujson.dumps(meta)

            elif name == 'scene':
                images.scene_blob = read_im(image_jnode['blob'])
                if images.scene_blob and meta:
                    images.scene_meta = ujson.dumps(meta)

            elif name == 'scene.alt':
                images.scene_alt_blob = read_im(image_jnode['blob'])
                if images.scene_alt_blob and meta:
                    images.scene_alt_meta = ujson.dumps(meta)

        if not all([images.plate_blob, images.body_thumb_blob, images.body_blob]):
            eventlog.log_debug('fail: read all images of [plate, body, thumb]')
            raise ValueError

        return images


class PlateRect:
    def __init__(self, left, top, right, bottom):
        """
        :type left: float
        :type top: float
        :type right: float
        :type bottom: float
        """
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

    @staticmethod
    def deserialize(rect_jnode):
        """
        :type rect_jnode: list of (float | int)
        :rtype: PlateRect
        """
        l, t, r, b = map(float, rect_jnode)
        return PlateRect(l, t, r, b)


class PlateTemplate:
    def __init__(self, country, region, index):
        """
        :type country: int
        :type region: int
        :type index: int
        """
        self.country = country
        self.region = region
        self.index = index

    @staticmethod
    def deserialize(plate_jnode):
        """
        :type plate_jnode: dict
        :rtype: PlateTemplate
        """
        country, region, index = plate_jnode['type']

        return PlateTemplate(country, region, index)


class Plate:
    def __init__(self, text, template, rect, validity):
        """
        :type text: str
        :type template: PlateTemplate
        :type rect: PlateRect
        :type validity: int
        """
        self.text = text
        self.template = template
        self.rect = rect
        self.validity = validity

    @staticmethod
    def deserialize(plate_jnode):
        """
        :type plate_jnode: dict
        :rtype: Plate
        """
        text = plate_jnode['text']
        template = PlateTemplate.deserialize(plate_jnode)
        rect = PlateRect.deserialize(plate_jnode['rect'])
        validity = plate_jnode['validity']

        plate = Plate(text, template, rect, validity)
        return plate


class Vehicle:
    def __init__(self, plate, speed, length, lane, direction):
        """
        :type plate: Plate | None
        :type speed: int | None
        :type length: int
        :type lane: int | None
        :type direction: int
        """
        self.plate = plate
        self.speed = speed
        self.length = length
        self.lane = lane
        self.direction = direction

    @staticmethod
    def deserialize(fact_jnode):
        """
        :type fact_jnode: dict
        :rtype: Vehicle
        """
        vehicle_jnode = fact_jnode['vehicle']
        plate = Plate.deserialize(vehicle_jnode['plate'])
        speed = vehicle_jnode.get('speed')
        lane = vehicle_jnode.get('lane')
        direction = vehicle_jnode.get('dir', DIR_UNKNOWN)
        length = vehicle_jnode.get('length', 0)

        vehicle = Vehicle(plate, speed, length, lane, direction)
        return vehicle


class Fact:
    def __init__(self, uuid, geolocation, catch_ts, best_ts, lost_ts, vehicle, speed_limit, images):
        """
        :type uuid: str
        :type geolocation: Geodata | None
        :type lost_ts: int
        :type catch_ts: int
        :type best_ts: int
        :type vehicle: Vehicle
        :type speed_limit: int | None
        """
        self.uuid = uuid
        self.geolocation = geolocation

        self.catch_ts = catch_ts
        self.best_ts = best_ts
        self.lost_ts = lost_ts

        self.vehicle = vehicle
        self.speed_limit = speed_limit

        self.images = images

    @staticmethod
    def deserialize(fact_jnode):
        """
        :type fact_jnode: dict
        :rtype: Fact
        :raise: ValueError
        """

        # deserialize fact-args
        fact_args = fact_jnode['fact']

        uuid = fact_args['uuid']
        geolocation_jnode = fact_args.get('geolocation')
        geolocation = Geodata.deserialize(geolocation_jnode) if geolocation_jnode else None

        catch_ts = fact_args['catch-ts']
        best_ts = fact_args['best-ts']
        lost_ts = fact_args['lost-ts']

        vehicle = Vehicle.deserialize(fact_args)
        speed_limit = fact_args.get('speed-limit')

        # deserialize fact-images
        images = FactImages.deserialize(fact_jnode['images'])

        fact = Fact(uuid, geolocation, catch_ts, best_ts, lost_ts, vehicle, speed_limit, images)
        return fact
