import ujson
from twisted.internet import inotify
from twisted.python.filepath import FilePath

import edge.eventlog as eventlog
from edge.fs_helpers import silent_mkdir_r, silent_rm_r

import _models
from _args import ReactProducerArgs
from _counters import ReactProducerCounters


class IReactProducer:
    def __init__(self):
        pass

    def set_fact_handler(self, handler):
        raise NotImplemented


# noinspection PyUnusedLocal
def stub_fact_react_handler(react_handler):
    """
    :type react_handler: (, _models.Fact) -> None
    :rtype: bool
    """
    print 'stub_fact_react_handler(react_handler)'


def _load_react_args_jnode(react_args_filepath):
    """
    :type react_args_filepath: str | None
    :rtype: dict
    """
    react_jnode = None
    if not react_args_filepath:
        return react_jnode

    # noinspection PyBroadException
    try:
        with open(react_args_filepath) as json_file:
            react_jnode = ujson.load(json_file)

    except IOError as x:
        eventlog.log_debug('fail: read %s (%s)' % (react_args_filepath, x))

    except ValueError as x:
        eventlog.log_debug('fail: deserialize %s (%s)' % (react_args_filepath, x))

    except:
        eventlog.log_debug_last_exc()

    return react_jnode


class ReactProducer(IReactProducer):
    def __init__(self, twisted_reactor, stat_counters, args):
        """
        :type stat_counters: ReactProducerCounters
        :type args: ReactProducerArgs
        """
        IReactProducer.__init__(self)

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stat_counters
        self._react_endpoint = args.endpoint

        self._react_handler = None
        self._inotify = inotify.INotify()
        self._inotify.startReading()

    def set_fact_handler(self, react_handler):
        """
        :type react_handler: (, _models.TrTrack) -> bool
        :rtype: bool
        """
        # Allow to set the handler only once.
        if react_handler and not self._react_handler:
            self._react_handler = react_handler
            self._reset_inotify_watch()
            return True

        return False

    def _reset_inotify_watch(self):
        try:
            self._inotify.ignore(FilePath(self._react_endpoint))
        except KeyError:
            pass

        silent_mkdir_r(self._react_endpoint)

        try:
            self._inotify.watch(
                FilePath(self._react_endpoint),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._handle_fact_react])
        except inotify.INotifyError as x:
            eventlog.log_debug('fail: reset inotify watch (%s)' % x)

    # noinspection PyUnusedLocal
    def _handle_fact_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_warning('react dir was changed to %s' % self._react_endpoint)
            return

        fact = None
        try:
            fact_jnode = _load_react_args_jnode(filepath.path)
            if fact_jnode is None:
                return

            fact = _models.Fact.deserialize(fact_jnode)

        except ValueError:
            eventlog.log_debug_last_exc()
            return

        finally:
            silent_rm_r(filepath.path)

        # noinspection PyBroadException
        try:
            self._react_handler(fact)
            self._stat_counters.inc_fact_count(fact)
        except:
            eventlog.log_debug_last_exc()
