import traceback
import ujson
from twisted.internet import inotify
from twisted.python.filepath import FilePath

import edge_rawutil.rawutil as rawutil

import _eventlog
from _models import TrScene
from _react_producer import IAddSceneReactProducer
from _kernel_param_set import AddSceneReact
import edge.fs_helpers as fs_helpers


def _listen_react_dir(react_dir, react_handler):
    fs_helpers.silent_rm_r(react_dir)
    fs_helpers.silent_mkdir_r(react_dir)

    notifier = inotify.INotify()
    notifier.startReading()
    notifier.watch(FilePath(react_dir),
                   mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE,
                   callbacks=[react_handler])


def _load_json_file(react_abspath):
    """
    :type react_abspath: str | None
    :rtype: dict
    """
    if not react_abspath:
        return None

    react_jnode = None
    try:
        with open(react_abspath) as json_file:
            react_jnode = ujson.load(json_file)

    except ValueError as value_error:
        _eventlog.log_invalid_json_file(value_error, react_abspath)

    except IOError as io_error:
        _eventlog.log_error(traceback.format_exc())
        _eventlog.log_exception(io_error)

    return react_jnode


class INotifyAddSceneReactProducer(IAddSceneReactProducer):

    def __init__(self, twisted_reactor, scene_react):
        """
        :type scene_react: AddSceneReact
        """
        IAddSceneReactProducer.__init__(self)

        self._react_dir = scene_react.react_dir
        self._react_handler = None

        self._twisted_reactor = twisted_reactor
        self._inotify = inotify.INotify()
        self._inotify.startReading()
        self._rawutil = rawutil.RawUtil.create(scene_react.image_wh[0], scene_react.image_wh[1], scene_react.log_level)
        self._twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._rawutil.destroy)

    def _reset_inotify_watch(self):
        try:
            self._inotify.ignore(FilePath(self._react_dir))
        except KeyError:
            pass

        fs_helpers.silent_mkdir_r(self._react_dir)

        success = False
        try:
            self._inotify.watch(
                FilePath(self._react_dir),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._on_notify])
            success = True
        except inotify.INotifyError:
            pass

        if not success:
            _eventlog.log_error('Failed to reset inotify watch')

    def set_addscene_react_handler(self, handler):
        self._react_handler = handler
        self._reset_inotify_watch()

    # noinspection PyUnusedLocal
    def _on_notify(self, ignored, path, mask):
        if 0 != mask & (inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            _eventlog.log_warning('React dir was changed to %s' % self._react_dir)
            return

        try:
            tr_scene_jnode = _load_json_file(path.path)
            tr_scene = TrScene.deserialize(tr_scene_jnode)

            if tr_scene.img_data.codec in ['MONO8T', 'I420PLN', 'NV12PLN', 'YUV422T', 'YUV411T', 'BayerRG8T',
                                           'BayerGR8T', 'BayerBG8T', 'BayerGB8T']:
                rgba_pil_image = self._rawutil.handle_image(
                    tr_scene.img_data.url, tr_scene.img_data.codec,
                    tr_scene.img_data.w, tr_scene.img_data.h)

                if not rgba_pil_image:
                    _eventlog.log_trace('fail: get image from librawutil')
                    raise ValueError

            else:
                raise NotImplemented

            self._react_handler(tr_scene, rgba_pil_image)
        except Exception as x:
            _eventlog.log_debug('fail: handle tr_scene')
            pass

        finally:
            fs_helpers.silent_rm_r(path.path)
