import _eventlog
import ujson
from os.path import isabs


class ImageQueueArgs:
    def __init__(self, endpoint, min_file_count, max_file_count):
        """
        :type endpoint: str
        :type min_file_count: int
        :type max_file_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_file_count
        self.max_file_count = max_file_count


class JsonQueueArgs:
    def __init__(self, endpoint, min_file_count, max_file_count):
        """
        :type endpoint: str
        :type min_file_count: int
        :type max_file_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_file_count
        self.max_file_count = max_file_count


class AddSceneReact:
    def __init__(self, react_dir):
        """
        :type react_dir: str
        """
        self.react_dir = react_dir
        self.log_level = 'info'
        self.image_wh = (None, None)

    @staticmethod
    def deserialize(source_jnode):
        """
        :type source_jnode: dict
        :rtype: AddSceneReact
        """
        react_dir = source_jnode['Url']

        add_scene_react = AddSceneReact(react_dir)
        return add_scene_react


class ImageStoreArgs:
    _IMAGE_CODEC_LIST = ['JPEG', 'RGBA']
    _IMAGE_MAX_WIDTH_LO = 640
    _IMAGE_MAX_WIDTH_HI = 65535

    _IMAGE_MAX_HEIGHT_LO = 480
    _IMAGE_MAX_HEIGHT_HI = 65535

    def __init__(self, codec, image_max_wh, image_queue_args):
        """
        :type codec: str
        :type image_max_wh: (int, int) | None
        :type image_queue_args: ImageQueueArgs

        """
        self.codec = codec
        self.image_max_wh = image_max_wh
        self.image_queue_args = image_queue_args

    @staticmethod
    def deserialize(image_store_jnode):
        """
        :type image_store_jnode: dict
        :rtype: _ImageResizer
        """
        max_file_count = image_store_jnode.get('max-file-count', 4)
        codec = image_store_jnode['format']

        if codec not in ImageStoreArgs._IMAGE_CODEC_LIST:
            _eventlog.log_error("Codec value is not valid")

        if not (int == type(max_file_count) and 4 <= max_file_count <= 8):
            raise ValueError('main.conf#/event-store/max-file-count')

        max_width = image_store_jnode.get('max-width', None)
        if not (max_width and ImageStoreArgs._IMAGE_MAX_WIDTH_LO <= max_width <= ImageStoreArgs._IMAGE_MAX_WIDTH_HI):
            max_width = None

        max_height = image_store_jnode.get('max-height', None)
        if not (max_height and ImageStoreArgs._IMAGE_MAX_HEIGHT_LO <= max_height <= ImageStoreArgs._IMAGE_MAX_HEIGHT_HI):
            max_height = None

        image_max_wh = None
        if max_width and max_height:
            image_max_wh = (max_width, max_height)

        image_store_dir = image_store_jnode['endpoint']
        if not (unicode == type(image_store_dir) and len(image_store_dir) < 64 and isabs(image_store_dir)):
            raise ValueError('main.conf#/store/endpoint')

        image_queue_args = ImageQueueArgs(image_store_dir, max_file_count - 2, max_file_count)

        return ImageStoreArgs(codec, image_max_wh, image_queue_args)


class SceneEventArgs:
    def __init__(self, json_queue_args_list):
        """
        :type json_queue_args_list: list of JsonQueueArgs
        """
        self.json_queue_args_list = json_queue_args_list
        self.log_level = 'info'

    @staticmethod
    def deserialize(scene_event_jnode):
        """
        :type scene_event_jnode: dict
        :rtype: SceneEventArgs
        """
        max_file_count = scene_event_jnode.get('max-file-count', 4)

        if not (int == type(max_file_count) and 4 <= max_file_count <= 8):
            raise ValueError('main.conf#/event-store/max-file-count')

        json_queue_args_list = map(
            lambda scene_endpoint: JsonQueueArgs(scene_endpoint, max_file_count - 5, max_file_count),
            scene_event_jnode['endpoints'])

        scene_event = SceneEventArgs(json_queue_args_list)
        return scene_event


class RendererArgs:
    _LR = ["always", "never"]
    _RA = ["always", "never", "if-plate"]

    def __init__(self, image_time, speed_limit, overspeed_limit, lr_plate, lr_plate_trace, lr_plate_upscale, lr_dir,
                 lr_lane, ra_position, ra_speed, ra_dir, ra_length, image_time_tz, lr_store_tout):
        """
        :type image_time: bool
        :type speed_limit: int | None
        :type overspeed_limit: int | None
        :type lr_plate: str
        :type lr_plate_trace: str
        :type lr_plate_upscale: str
        :type lr_dir: str
        :type lr_lane: str
        :type ra_position: str
        :type ra_speed: str
        :type ra_dir: str
        :type ra_length: str
        :type image_time_tz: str
        :type lr_store_tout: int
        """
        self.image_time = image_time
        self.speed_limit = speed_limit
        self.overspeed_limit = overspeed_limit
        self.lr_plate = lr_plate
        self.lr_plate_trace = lr_plate_trace
        self.lr_plate_upscale = lr_plate_upscale
        self.lr_dir = lr_dir
        self.lr_lane = lr_lane
        self.ra_position = ra_position
        self.ra_speed = ra_speed
        self.ra_dir = ra_dir
        self.ra_length = ra_length
        self.image_time_tz = image_time_tz
        self.lr_store_tout = lr_store_tout

    @staticmethod
    def deserialize(renderer_jnode):
        image_time = renderer_jnode.get('ImageTime', False)
        speed_limit = renderer_jnode.get('SpeedLimit', None)
        overspeed_limit = renderer_jnode.get('OverSpeedLimit', None)

        # lr1
        video_plate = 'always'
        lr_plate = renderer_jnode.get('LrPlate', video_plate)
        if lr_plate not in RendererArgs._LR:
            lr_plate = video_plate

        # lr2
        video_plate_trace = 'always'
        lr_plate_trace = renderer_jnode.get('LrPlateTrace', video_plate_trace)
        if lr_plate_trace not in RendererArgs._LR:
            lr_plate_trace = video_plate_trace

        # lr3
        video_body_image = 'never'
        lr_plate_upscale = renderer_jnode.get('LrPlateUpscale', video_body_image)
        if lr_plate_upscale not in RendererArgs._LR:
            lr_plate_upscale = video_plate_trace

        # lr4
        video_dir = 'always'
        lr_dir = renderer_jnode.get('LrDir', video_dir)
        if lr_dir not in RendererArgs._LR:
            lr_dir = video_dir

        # lr5
        video_lane = 'always'
        lr_lane = renderer_jnode.get('LrLane', video_lane)
        if lr_lane not in RendererArgs._LR:
            lr_lane = video_lane

        # ra1
        radar_position = 'never'
        ra_position = renderer_jnode.get('RaPosition', radar_position)
        if ra_position not in RendererArgs._RA:
            ra_position = radar_position

        # ra2
        radar_speed = 'if-plate'
        ra_speed = renderer_jnode.get('RaSpeed', radar_speed)
        if ra_speed not in RendererArgs._RA:
            ra_speed = radar_speed

        # ra3
        radar_dir = 'if-plate'
        ra_dir = renderer_jnode.get('RaDir', radar_dir)
        if ra_dir not in RendererArgs._RA:
            ra_dir = radar_dir

        # ra4
        radar_length = 'never'
        ra_length = renderer_jnode.get('RaLength', radar_length)
        if ra_length not in RendererArgs._RA:
            ra_length = radar_length

        #
        tz = 'Etc/UTC'
        image_time_tz = renderer_jnode.get('ImageTimeTz', tz)

        # define LrStoreTout
        def_lr_store_tout = 1
        lr_store_tout = renderer_jnode.get('LrStoreTout', def_lr_store_tout)
        if not (1 <= lr_store_tout <= 60):

            _eventlog.log_warning('The param LrStoreTout has invalid value %d. The default value %d is used.'
                                  % (lr_store_tout, def_lr_store_tout))
            lr_store_tout = def_lr_store_tout

        renderer = RendererArgs(image_time, speed_limit, overspeed_limit, lr_plate, lr_plate_trace, lr_plate_upscale,
                                lr_dir, lr_lane, ra_position, ra_speed, ra_dir, ra_length, image_time_tz, lr_store_tout)
        return renderer


class LaneDetectorArgs:
    def __init__(self, id_to_lane_list, img_width, img_height):
        """
        :type id_to_lane_list: list of (int, list of (int, int))
        :type img_width: int
        :type img_height: int
        """
        self.id_to_lane_list = id_to_lane_list
        self.img_width = img_width
        self.img_height = img_height

    @staticmethod
    def deserialize(road_mask_jnode):
        """
        :type road_mask_jnode: dict
        :rtype: LaneDetectorArgs
        """
        id_to_lane_list = map(lambda lane_jnode: (lane_jnode['id'], lane_jnode['mask']),
                              road_mask_jnode.get('lanes', []))

        resolution_jnode = road_mask_jnode.get('resolution', {})
        img_w = resolution_jnode.get('w', 0)
        img_h = resolution_jnode.get('h', 0)

        return LaneDetectorArgs(id_to_lane_list, img_w, img_h)


class KernelArgs:
    def __init__(self, add_scene_react, scene_event, image_store, renderer, lane_detector_args):
        """
        :type add_scene_react: AddSceneReact
        :type scene_event: SceneEventArgs
        :type image_store: ImageStoreArgs
        :type renderer: RendererArgs
        :type lane_detector_args: LaneDetectorArgs
        """
        self.lane_detector_args = lane_detector_args
        self.add_scene_react = add_scene_react
        self.scene_event = scene_event
        self.image_store = image_store
        self.renderer = renderer

    @staticmethod
    def deserialize(main_conf_jnode, log_level):
        """
        :type main_conf_jnode: dict
        :type log_level: str
        :rtype: KernelArgs
        """
        lane_detector_args = LaneDetectorArgs.deserialize(main_conf_jnode.get('road-mask', {}))

        image_store = ImageStoreArgs.deserialize(main_conf_jnode['image-store'])

        add_scene_react = AddSceneReact.deserialize(main_conf_jnode['Source'])
        add_scene_react.log_level = log_level
        add_scene_react.image_wh = image_store.image_max_wh

        scene_event = SceneEventArgs.deserialize(main_conf_jnode['scene-event'])
        scene_event.log_level = log_level

        renderer_jnode = main_conf_jnode['Renderer']

        base_conf_filepath = main_conf_jnode['Base']
        with open(base_conf_filepath) as base_conf_jnode:
            base_conf_jnode = ujson.load(base_conf_jnode)

        timezone = base_conf_jnode['rw']['tz']
        renderer_jnode['ImageTimeTz'] = timezone

        renderer = RendererArgs.deserialize(renderer_jnode)

        kernel_args = KernelArgs(add_scene_react, scene_event,  image_store, renderer, lane_detector_args)
        return kernel_args
