
class SceneImage:
    def __init__(self, image_ts, image_width, image_height, image_codec, image_url):
        """
        :type image_width: int
        :type image_height: int
        :type image_url: str
        :type image_codec: str
        """
        self.image_ts = image_ts
        self.w = image_width
        self.h = image_height
        self.url = image_url
        self.codec = image_codec

    @staticmethod
    def deserialize(image_jnode):
        """
        :type image_jnode: dict
        :rtype: SceneImage
        """
        image_ts = int(image_jnode['Ts'])
        image_width, image_height = image_jnode['Wh']
        image_url = image_jnode['Url']
        image_codec = image_jnode['Codec']

        scene_image = SceneImage(image_ts, image_width, image_height, image_codec, image_url)
        return scene_image


class LrUpscale:
    def __init__(self, left, top, right, bottom):
        """
        :type left: float
        :type top: float
        :type right: float
        :type bottom: float
        """
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom

    @staticmethod
    def deserialize(lr_upscale_jnode):
        """
        :type lr_upscale_jnode: dict
        :rtype: LrUpscale
        """
        left, top, right, bottom = lr_upscale_jnode
        lr_upscale = LrUpscale(left, top, right, bottom)
        return lr_upscale


class LrSceneItem:
    def __init__(self, lr_oid, plate_text, plate_ltrb, direction, history):
        """
        :type lr_oid: int
        :type plate_text: str
        :type plate_ltrb: list of double
        :type direction: int
        :type history: list
        """
        self.id = lr_oid
        self.plate_text = plate_text
        self.plate_ltrb = plate_ltrb
        self.direction = direction
        self.history = history

    def get_ref_point_xy(self, image_width, image_height):
        """
        :type image_width: int
        :type image_height: int
        :rtype: (int, int)
        """
        x = (self.plate_ltrb[0] + self.plate_ltrb[2]) * image_width / 2
        y = self.plate_ltrb[3] * image_height
        return int(x), int(y)

    @staticmethod
    def deserialize(lr_scene_item_jnode):
        """
        :type lr_scene_item_jnode: dict
        :rtype: LrSceneItem
        """
        best_variant_jnode = lr_scene_item_jnode['Plate']

        lr_oid = lr_scene_item_jnode['Id']
        direction = lr_scene_item_jnode['Dir']
        plate_text = best_variant_jnode['P']

        plate_rect_jnode = best_variant_jnode['R']
        plate_ltrb = [float(plate_rect_jnode[0]), float(plate_rect_jnode[1]),
                      float(plate_rect_jnode[2]), float(plate_rect_jnode[3])]

        history = lr_scene_item_jnode['PlateTrace']

        lr_scene_item = LrSceneItem(lr_oid, plate_text, plate_ltrb, direction, history)
        return lr_scene_item


class RaSceneItem:
    def __init__(self, ra_oid, img_x, img_y, speed, length, direction):
        """
        :type ra_oid: int
        :type img_x: float
        :type img_y: float
        :type speed: int
        :type length: int
        :type direction: int
        """
        self.id = ra_oid
        self.img_xy = [img_x, img_y]
        self.speed = speed
        self.length = length
        self.direction = direction

    def get_ref_point_xy(self, image_width, image_height):
        """
        :type image_width: int
        :type image_height: int
        :rtype: (int, int)
        """
        x = self.img_xy[0] * image_width
        y = self.img_xy[1] * image_height
        return int(x), int(y)

    @staticmethod
    def deserialize(ra_scene_item_jnode):
        """
        :type ra_scene_item_jnode: dict
        :rtype: RaSceneItem
        """
        ra_id = ra_scene_item_jnode['oid']
        speed = int(ra_scene_item_jnode['speed'])

        img_x = float(ra_scene_item_jnode['xy'][0])
        img_y = float(ra_scene_item_jnode['xy'][1])
        length = int(ra_scene_item_jnode['len'])
        direction = int(ra_scene_item_jnode['dir'])

        ra_scene_item = RaSceneItem(ra_id, img_x, img_y, speed, length, direction)
        return ra_scene_item


class LrRaSceneItem:
    def __init__(self, lr_scene_item, ra_scene_item):
        """
        :type lr_scene_item: LrSceneItem
        :type ra_scene_item: RaSceneItem
        """
        self.lr_data = lr_scene_item
        self.ra_data = ra_scene_item

    @staticmethod
    def deserialize(lr_ra_jnode):
        """
        :type lr_ra_jnode: dict
        :rtype: LrRaSceneItem
        """
        lr_scene_item = LrSceneItem.deserialize(lr_ra_jnode['Lr'])
        ra_scene_item = RaSceneItem.deserialize(lr_ra_jnode['Ra'])

        lr_ra_scene_item = LrRaSceneItem(lr_scene_item, ra_scene_item)
        return lr_ra_scene_item


class TrScene:
    def __init__(self, scene_image, lr_upscale, lr_ra_items, lr_items, ra_items):
        """
        :type scene_image: SceneImage
        :type lr_upscale: LrUpscale
        :type lr_ra_items: list of LrRaSceneItem
        :type lr_items: list of LrSceneItem
        :type ra_items: list of RaSceneItem
        """
        self.img_data = scene_image
        self.lr_upscale = lr_upscale

        self.tr_data = lr_ra_items
        self.lr_data = lr_items
        self.ra_data = ra_items

    @staticmethod
    def deserialize(tr_scene_jnode):
        """
        :type tr_scene_jnode: dict
        :rtype: TrScene
        """
        tr_scene_image = SceneImage.deserialize(tr_scene_jnode['Image'])

        lr_upscale = LrUpscale.deserialize(tr_scene_jnode['LrUpscaleLtrb'])

        lr_ra_items = map(lambda lr_ra_item_jnode: LrRaSceneItem.deserialize(lr_ra_item_jnode),
                          tr_scene_jnode['LrRa'])

        lr_items = map(lambda lr_scene_item_jnode: LrSceneItem.deserialize(lr_scene_item_jnode),
                       tr_scene_jnode['Lr'])

        ra_items = map(lambda ra_scene_item_jnode: RaSceneItem.deserialize(ra_scene_item_jnode),
                       tr_scene_jnode['Ra'])

        tr_scene = TrScene(tr_scene_image, lr_upscale, lr_ra_items, lr_items, ra_items)
        return tr_scene
