import os
import datetime
import pytz

from PIL import ImageDraw, ImageFont
from edge.lane_detector import LaneDetector
from _kernel_param_set import RendererArgs

import _models

_FONT_PT_SIZE = 20

_RA_BOX_DIR = 1
_RA_BOX_SPEED = 2
_RA_BOX_LENGTH = 3
_RA_BOX_LANE = 4

_LR_BOX_DIR = 5
_LR_BOX_PLATE_TEXT = 6
_LR_BOX_LANE = 7

_RA_BOX_SPEED_WITHOUT_LENGTH = 8
_RA_BOX_DIR_WITHOUT_LENGTH = 9

_SYM_W = 12
_SYM_H = 17
_SYM_MARGIN_W = 6
_SYM_MARGIN_H = 3

_MODE_ALWAYS = 'always'
_MODE_NEVER = 'never'
_MODE_IF_PLATE = 'if-plate'

_LANE_UNDEFINED = 0


def _get_ra_board_box_ltrb(ref_point_xy, box_id):
    """
    :type ref_point_xy: (int, int)
    :type box_id: int
    :rtype: (int, int, int, int)
    """
    top = ref_point_xy[1] + _SYM_MARGIN_H
    bottom = top + _SYM_H + _SYM_MARGIN_H

    if box_id == _RA_BOX_DIR:
        right = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W - _SYM_MARGIN_W
        left = right - _SYM_MARGIN_W - 2 * _SYM_MARGIN_W
        return left, top, right, bottom

    if box_id == _RA_BOX_SPEED:
        right = ref_point_xy[0] - (_SYM_MARGIN_W / 2)
        left = right - 4 * _SYM_W - _SYM_MARGIN_W / 2
        return left, top, right, bottom

    if box_id == _RA_BOX_LENGTH:
        left = ref_point_xy[0] + (_SYM_MARGIN_W / 2)
        right = left + 4 * _SYM_W + _SYM_MARGIN_W
        return left, top, right, bottom

    if box_id == _RA_BOX_LANE:
        left = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W
        right = left + 2 * _SYM_MARGIN_W + _SYM_W
        return left, top, right, bottom

    if box_id == _RA_BOX_SPEED_WITHOUT_LENGTH:
        right = ref_point_xy[0] + 2 * _SYM_W + _SYM_MARGIN_W
        left = ref_point_xy[0] - 2 * _SYM_W - _SYM_MARGIN_W / 4
        return left, top, right, bottom

    if box_id == _RA_BOX_DIR_WITHOUT_LENGTH:
        right = ref_point_xy[0] - 2 * _SYM_W - _SYM_MARGIN_W
        left = (ref_point_xy[0] + (_SYM_MARGIN_W / 2)) - 4 * _SYM_W - _SYM_MARGIN_W
        return left, top, right, bottom

    raise NotImplemented


def _get_lr_board_box_ltrb(ref_point_xy, box_id):
    """
    :type ref_point_xy: (int, int)
    :type box_id: int
    :rtype: (int, int, int, int)
    """
    top = ref_point_xy[1] + _SYM_MARGIN_H
    bottom = top + _SYM_H + _SYM_MARGIN_H

    if box_id == _LR_BOX_DIR:
        right = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W - _SYM_MARGIN_W
        left = right - _SYM_MARGIN_W - 2 * _SYM_MARGIN_W
        return left, top, right, bottom

    if box_id == _LR_BOX_PLATE_TEXT:
        right = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W
        left = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W
        return left, top, right, bottom

    if box_id == _LR_BOX_LANE:
        left = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W + _SYM_MARGIN_W
        right = left + _SYM_MARGIN_W + _SYM_W
        return left, top, right, bottom

    raise NotImplemented


def _render_speed(rgb_image_draw, ltrb, speed, yellow_speed_threshold, red_speed_threshold):
    """
    :type rgb_image_draw: PIL.ImageDraw.ImageDraw
    :type ltrb: (int, int, int, int)
    :type speed: int
    :type yellow_speed_threshold: int | None
    :type red_speed_threshold: int | None
    """
    _FG_COLOR = (0, 0, 0)

    _GREEN_COLOR = (145, 186, 108)
    _YELLOW_COLOR = (255, 214, 0)
    _RED_COLOR = (208, 100, 89)

    if red_speed_threshold and speed > red_speed_threshold:
        _bg_color = _RED_COLOR

    elif yellow_speed_threshold and speed > yellow_speed_threshold:
        _bg_color = _YELLOW_COLOR

    else:
        _bg_color = _GREEN_COLOR

    rgb_image_draw.rectangle(ltrb, fill=_bg_color)

    speed_text = str(speed)
    text_width, text_height = rgb_image_draw.getfont().getsize(speed_text)

    left = (ltrb[2] + ltrb[0]) / 2 - text_width / 2
    top = (ltrb[3] + ltrb[1]) / 2 - _FONT_PT_SIZE / 2

    rgb_image_draw.text((left, top), speed_text, fill=_FG_COLOR)


def _render_length(rgb_image_draw, ltrb, length):
    """
    :type rgb_image_draw: PIL.ImageDraw.ImageDraw
    :type ltrb: (int, int, int, int)
    :type length: int
    """
    _WHITE_COLOR = (255, 255, 255)
    _BLACK_COLOR = (0, 0, 0)

    rgb_image_draw.rectangle(ltrb, fill=_WHITE_COLOR)

    speed_text = str(length)
    text_width, text_height = rgb_image_draw.getfont().getsize(speed_text)

    left = (ltrb[2] + ltrb[0]) / 2 - text_width / 2
    top = (ltrb[3] + ltrb[1]) / 2 - _FONT_PT_SIZE / 2

    rgb_image_draw.text((left, top), speed_text, fill=_BLACK_COLOR)


def _render_direction(rgb_image_draw, ltrb, direction):
    """
    :type rgb_image_draw: PIL.ImageDraw.ImageDraw
    :type ltrb: (int, int, int, int)
    :type direction: int
    """

    _BLACK_COLOR = (0, 0, 0)
    _DIR_APPROACHING = 11
    _DIR_RECEDING = 12

    _TEXT_UNKNOWN = '?'
    _TEXT_APPROACHING = u'\u2193'
    _TEXT_RECEDING = u'\u2191'

    _BG_COLOR_UNKNOWN = (255, 180, 90)
    _BG_COLOR_APPROACHING = (140, 200, 255)
    _BG_COLOR_RECEDING = (215, 165, 255)

    if direction == _DIR_APPROACHING:
        direction_text = _TEXT_APPROACHING
        _bg_color = _BG_COLOR_APPROACHING

    elif direction == _DIR_RECEDING:
        direction_text = _TEXT_RECEDING
        _bg_color = _BG_COLOR_RECEDING

    else:
        direction_text = _TEXT_UNKNOWN
        _bg_color = _BG_COLOR_UNKNOWN

    rgb_image_draw.rectangle(ltrb, fill=_bg_color)

    text_width, text_height = rgb_image_draw.getfont().getsize(direction_text)

    left = (ltrb[2] + ltrb[0]) / 2 - text_width / 2
    top = (ltrb[3] + ltrb[1]) / 2 - _FONT_PT_SIZE / 2

    rgb_image_draw.text((left, top), direction_text, fill=_BLACK_COLOR)


def _render_lane(rgb_image_draw, ltrb, lane):
    """
    :type rgb_image_draw: PIL.ImageDraw.ImageDraw
    :type ltrb: (int, int, int, int)
    :type lane: int
    """
    _BG_COLOR = (206, 180, 141)
    _BLACK_COLOR = (0, 0, 0)

    rgb_image_draw.rectangle(ltrb, fill=_BG_COLOR)

    lane_text = str(lane)
    text_width, text_height = rgb_image_draw.getfont().getsize(lane_text)

    left = (ltrb[2] + ltrb[0]) / 2 - text_width / 2
    top = (ltrb[3] + ltrb[1]) / 2 - _FONT_PT_SIZE / 2

    rgb_image_draw.text((left, top), lane_text, fill=_BLACK_COLOR)


def _render_plate_text(rgb_image_draw, ltrb, plate_text):
    """
    :type rgb_image_draw: PIL.ImageDraw.ImageDraw
    :type ltrb: (int, int, int, int)
    :type plate_text: str
    """
    _WHITE_COLOR = (255, 255, 255)
    _BLACK_COLOR = (0, 0, 0)

    rgb_image_draw.rectangle(ltrb, fill=_WHITE_COLOR)

    text_width, text_height = rgb_image_draw.getfont().getsize(plate_text)

    left = (ltrb[2] + ltrb[0]) / 2 - text_width / 2
    top = (ltrb[3] + ltrb[1]) / 2 - _FONT_PT_SIZE / 2

    rgb_image_draw.text((left, top), plate_text, fill=_BLACK_COLOR)


def _render_scene_ts(rgb_image_draw, scene_ts_usec, tz):
    """
    :type rgb_image_draw: PIL.ImageDraw.ImageDraw
    :type scene_ts_usec: int
    :type tz: int
    """
    _BG_COLOR = (255, 255, 255)
    _FG_COLOR = (0, 0, 0)

    scene_ts_sec = float(scene_ts_usec) / 1000000
    image_time_tz = pytz.timezone(tz)
    text = str(datetime.datetime.fromtimestamp(scene_ts_sec, tz=image_time_tz).strftime('%d.%m.%Y %H:%M:%S.%f'))
    text_width, text_height = rgb_image_draw.getfont().getsize(text)

    ltrb = (0, 0, text_width, text_height)
    rgb_image_draw.rectangle(ltrb, fill=_BG_COLOR)
    rgb_image_draw.text((0, 0), text, fill=_FG_COLOR)


class Renderer:
    _FONT_ABS_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'ttf', 'Cousine-Bold.ttf')
    _FONT = ImageFont.truetype(_FONT_ABS_PATH, _FONT_PT_SIZE, encoding='utf-8')

    _PLATE_MARGIN_X = int(2 * _FONT_PT_SIZE / 10.0)
    _PLATE_MARGIN_Y = int(1 * _FONT_PT_SIZE / 10.0)
    _TEXT_BLACK_COLOR = (0, 0, 0)

    _RA_ITEM_RADIUS = 4

    _COLOR_BLACK = (0, 0, 0)
    _COLOR_WHITE = (255, 255, 255)
    _COLOR_YELLOW = (255, 255, 0)
    _COLOR_GREEN = (0, 255, 0)
    _COLOR_RED = (255, 0, 0)

    # http://www.stm.dp.ua/web-design/color-html.php
    _COLORS = [(255, 248, 220),  # Cornsilk
               (255, 228, 181),  # Moccasin
               (255, 245, 238),  # Seashell
               (240, 255, 240),  # Honeydew
               (230, 230, 250),  # lavender
               (240, 255, 255),  # Azure
               (255, 228, 225),  # MistyRose
               (152, 251, 152),  # PaleGreen
               (255, 215, 0),  # Gold
               (188, 143, 143),  # RosyBrown
               (245, 245, 220),  # Beige
               (245, 222, 179),  # Wheat
               (193, 255, 193),  # DarkSeaGreen1
               (84, 255, 159),  # SeaGreen1
               ]

    def __init__(self, renderer_args, lr_color_tout_store, lane_detector):
        """
        :type renderer_args: RendererArgs
        :type lr_color_tout_store: _lr_color_tout_store.LrColorToutStore
        :type lane_detector: LaneDetector
        """
        self._render_scene_ts = renderer_args.image_time
        self._speed_limit = renderer_args.speed_limit
        self._overspeed_limit = renderer_args.overspeed_limit
        self._render_lr_plate = renderer_args.lr_plate
        self._render_lr_plate_trace = renderer_args.lr_plate_trace
        self._render_lr_upscale_region = renderer_args.lr_plate_upscale
        self._render_lr_dirs = renderer_args.lr_dir
        self._render_lr_lane = renderer_args.lr_lane
        self._render_ra_position = renderer_args.ra_position
        self._render_ra_speed = renderer_args.ra_speed
        self._render_ra_dirs = renderer_args.ra_dir
        self._render_ra_length = renderer_args.ra_length
        self.image_time_tz = renderer_args.image_time_tz
        self._lr_color_tout_store = lr_color_tout_store  # +
        self._lane_detector = lane_detector

    def render_tr_scene(self, rgb_image, tr_scene):
        """
        :type rgb_image: PIL.Image.Image
        :type tr_scene: _models.TrScene
        :rtype: PIL.Image.Image
        """
        image_width, image_height = rgb_image.size
        rgb_image_draw = ImageDraw.Draw(rgb_image)
        rgb_image_draw.setfont(font=Renderer._FONT)

        # if self._render_ra_position == 'always':
        map(lambda ra_item: self._render_ra_board(
            rgb_image_draw, image_width, image_height, ra_item),
            tr_scene.ra_data)

        map(lambda lr_item: self._render_lr_board(
            rgb_image_draw, image_width, image_height, lr_item, tr_scene.lr_upscale),
            tr_scene.lr_data)

        map(lambda lr_ra_item: self._render_lr_ra_board(
            rgb_image_draw, image_width, image_height, tr_scene.lr_upscale, lr_ra_item),
            tr_scene.tr_data)

        if self._render_scene_ts:
            _render_scene_ts(rgb_image_draw, tr_scene.img_data.image_ts, self.image_time_tz)

    def _render_ra_board(self, rgb_image_draw, image_width, image_height, ra_item, ref_point_xy=None):
        """
        :type rgb_image_draw: PIL.ImageDraw.ImageDraw
        :type ra_item: _models.RaSceneItem
        """
        if not ref_point_xy:
            lpr_plate_match_with_ra_point = False
            ref_point_xy = ra_item.get_ref_point_xy(image_width, image_height)
        else:
            lpr_plate_match_with_ra_point = True

        if self._render_ra_position == _MODE_ALWAYS or \
                (lpr_plate_match_with_ra_point and self._render_ra_position == _MODE_IF_PLATE):
            ra_point_color = Renderer._COLOR_RED
            x = int(ra_item.img_xy[0] * image_width)
            y = int(ra_item.img_xy[1] * image_height)

            ra_rect_ltrb = (x - self._RA_ITEM_RADIUS, y - self._RA_ITEM_RADIUS,
                            x + self._RA_ITEM_RADIUS, y + self._RA_ITEM_RADIUS)

            rgb_image_draw.rectangle(ra_rect_ltrb, fill=ra_point_color)

        # ltrb = _get_ra_board_box_ltrb((x, y), _RA_BOX_LANE)
        # rgb_image_draw.rectangle(ltrb, outline=Renderer._COLOR_RED)

        if self._render_ra_length == _MODE_ALWAYS or \
                (lpr_plate_match_with_ra_point and self._render_ra_length == _MODE_IF_PLATE):
            ltrb = _get_ra_board_box_ltrb(ref_point_xy, _RA_BOX_LENGTH)
            _render_length(rgb_image_draw, ltrb, ra_item.length)

            if self._render_ra_speed == _MODE_ALWAYS or \
                    (lpr_plate_match_with_ra_point and self._render_ra_speed == _MODE_IF_PLATE):
                ltrb = _get_ra_board_box_ltrb(ref_point_xy, _RA_BOX_SPEED)
                _render_speed(rgb_image_draw, ltrb, ra_item.speed,
                              self._speed_limit, self._overspeed_limit)

            if self._render_ra_dirs == _MODE_ALWAYS or \
                    (lpr_plate_match_with_ra_point and self._render_ra_dirs == _MODE_IF_PLATE):
                ltrb = _get_ra_board_box_ltrb(ref_point_xy, _RA_BOX_DIR)
                _render_direction(rgb_image_draw, ltrb, ra_item.direction)

        else:
            if self._render_ra_speed == _MODE_ALWAYS or \
                    (lpr_plate_match_with_ra_point and self._render_ra_speed == _MODE_IF_PLATE):
                ltrb = _get_ra_board_box_ltrb(ref_point_xy, _RA_BOX_SPEED_WITHOUT_LENGTH)
                _render_speed(rgb_image_draw, ltrb, ra_item.speed,
                              self._speed_limit, self._overspeed_limit)

            if self._render_ra_dirs == _MODE_ALWAYS or \
                    (lpr_plate_match_with_ra_point and self._render_ra_dirs == _MODE_IF_PLATE):
                ltrb = _get_ra_board_box_ltrb(ref_point_xy, _RA_BOX_DIR_WITHOUT_LENGTH)
                _render_direction(rgb_image_draw, ltrb, ra_item.direction)

    def _render_lr_board(self, rgb_image_draw, image_width, image_height, lr_item, lr_upscale, ref_point_xy=None):
        """
        :type rgb_image_draw: PIL.ImageDraw.ImageDraw
        :type image_width: int
        :type image_height: int
        :type lr_item: _models.LrSceneItem
        :type lr_upscale: _models.LrUpscale
        """
        plate_rect_left = lr_item.plate_ltrb[0] * image_width - self._PLATE_MARGIN_X
        plate_rect_top = lr_item.plate_ltrb[1] * image_height - self._PLATE_MARGIN_Y
        plate_rect_right = lr_item.plate_ltrb[2] * image_width + 2 * self._PLATE_MARGIN_X
        plate_rect_bottom = lr_item.plate_ltrb[3] * image_height + 2 * self._PLATE_MARGIN_Y

        if self._render_lr_plate_trace == _MODE_ALWAYS:
            for list_item in lr_item.history:
                try:
                    image_x = list_item[0] * image_width
                    image_y = list_item[1] * image_height
                    lr_rect_history_ltrb = (image_x - 1, image_y - 1, image_x + 1, image_y + 1)

                    rgb_image_draw.rectangle(lr_rect_history_ltrb,
                                             fill=self._lr_color_tout_store.get_lr_oid_color(lr_item.id))
                except Exception:
                    pass

        if not ref_point_xy:
            ref_point_xy = lr_item.get_ref_point_xy(image_width, image_height)

        if self._render_lr_upscale_region == _MODE_ALWAYS:
            width = plate_rect_right - plate_rect_left
            height = plate_rect_bottom - plate_rect_top

            new_upper_left_x = ref_point_xy[0] - (width * lr_upscale.left)
            new_lower_right_x = ref_point_xy[0] + (width * lr_upscale.right)

            new_upper_left_y = ref_point_xy[1] - (height * lr_upscale.top)
            new_lower_right_y = ref_point_xy[1] + (height * lr_upscale.bottom)
            new_plate_rect_ltrb = (new_upper_left_x, new_upper_left_y, new_lower_right_x, new_lower_right_y)
            rgb_image_draw.rectangle(new_plate_rect_ltrb, outline=Renderer._COLOR_YELLOW)

        if self._render_lr_dirs == _MODE_ALWAYS:
            ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_DIR)
            _render_direction(rgb_image_draw, ltrb, lr_item.direction)

        if self._render_lr_lane == _MODE_ALWAYS:
            ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_LANE)
            lane_set = self._lane_detector.get_lane2(lr_item.plate_ltrb[0], lr_item.plate_ltrb[1],
                                                     lr_item.plate_ltrb[2], lr_item.plate_ltrb[3])
            lane = lane_set.pop() if len(lane_set) == 1 else _LANE_UNDEFINED
            _render_lane(rgb_image_draw, ltrb, lane)

        if self._render_lr_plate == _MODE_ALWAYS:
            plate_rect_ltrb = (plate_rect_left, plate_rect_top, plate_rect_right, plate_rect_bottom)
            rgb_image_draw.rectangle(plate_rect_ltrb, outline=Renderer._COLOR_YELLOW)
            ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_PLATE_TEXT)
            _render_plate_text(rgb_image_draw, ltrb, lr_item.plate_text)

    def _render_lr_ra_board(self, rgb_image_draw, image_width, image_height, lr_upscale, lr_ra_item):
        """
        :type rgb_image_draw: PIL.ImageDraw.ImageDraw
        :type image_width: int
        :type image_height: int
        :type lr_upscale: _models.LrUpscale
        :type lr_ra_item: _models.LrRaSceneItem
        """

        lr_item = lr_ra_item.lr_data
        ref_point_x, ref_point_y = lr_item.get_ref_point_xy(image_width, image_height)
        self._render_lr_board(rgb_image_draw, image_width, image_height,
                              lr_item, lr_upscale, (ref_point_x, ref_point_y + 2 * _SYM_MARGIN_H))

        ra_item = lr_ra_item.ra_data
        self._render_ra_board(rgb_image_draw, image_width, image_height,
                              ra_item, (ref_point_x, ref_point_y + _SYM_H + 5 * _SYM_MARGIN_H))
