#!/usr/bin/python
import json
import sys

import edge.admin.ui_types as ui_types
import edge.eventlog as eventlog
from edge.admin.reply_helpers import *
import edge.pipeline_helpers as pipeline_helpers
import edge.db as db

import pipeline.mv_pipeline_creator as mv_pipeline_creator
import pipeline.rtsp_pipeline_creator as rtsp_pipeline_creator
import pipeline.ui_pipeline_set as ui_pipeline_set
from license.lpr._helpers import get_used_license_count, get_total_license_count
from pipeline.ui_pipeline import UiPipeline, get_pipeline_ext_jnode


class PipelineError(RuntimeError):
    pass


def _get_new_pid(pipeline_jnode_set):
    """
    :type pipeline_jnode_set: list of dict
    :rtype: str | None
    """
    used_ui_pid_set = set(map(lambda pipeline_jnode: int(pipeline_jnode['id']), pipeline_jnode_set))

    ui_pid = None
    for ui_pid_candidate in xrange(0, 32):
        if ui_pid_candidate not in used_ui_pid_set:
            ui_pid = str(ui_pid_candidate)
            eventlog.log_debug("new-pipeline-id: %s" % ui_pid)
            break
    return ui_pid


def _run(query_args):
    """
    :type query_args: edge.admin.ui_types.UiQueryArgs
    :raises: PipelineError, db.DbError
    """
    # handle ../pipeline/*
    ui_pid = query_args.get_ui_pid()
    if ui_pid:
        if ui_pipeline_set.validate_ui_pid(ui_pid) or query_args.path_info:
            ui_handler = UiPipeline.deserialize(query_args)
            reply_jnode = ui_handler.handle()
            return reply_jnode

        else:
            eventlog.log_debug("fail: pipeline (%s) not-found" % ui_pid)
            return reply_fail(REPLY_NOT_FOUND_404)

    # handle ../pipeline
    if query_args.query_string:
        eventlog.log_debug("fail: bad-request")
        return reply_fail(REPLY_BAD_REQUEST_400)

    if query_args.http_method == 'GET':
        if query_args.raw_body:
            eventlog.log_debug("fail: bad-request")
            return reply_fail(REPLY_BAD_REQUEST_400)

        pipeline_set_jnode = pipeline_helpers.get_pipeline_jnode_set()
        if pipeline_set_jnode is None:
            eventlog.log_debug("fail: internal-server-error (invalid pipeline-set-jnode)")
            raise PipelineError

        reply_jnode = map(get_pipeline_ext_jnode, pipeline_set_jnode)

        return reply_done(reply_jnode)

    elif query_args.http_method == 'POST':
        # noinspection PyBroadException
        try:
            total_thread_count = get_total_license_count()
            used_thread_count = get_used_license_count()
            eventlog.log_debug('license total_thread_count: %s, used_thread_count: %s' %
                               (total_thread_count, used_thread_count))

        except RuntimeError:
            eventlog.log_debug('fail: get license')
            raise PipelineError

        if used_thread_count >= total_thread_count:
            eventlog.log_debug('fail: free lpr licence is not available')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code='lpr-licence-absent',
                               error_text='free lpr licence is not available')

        else:
            eventlog.log_debug('License OK')

        # define ui_pid of new channel
        pipeline_jnode_set = pipeline_helpers.get_pipeline_jnode_set2()
        if pipeline_jnode_set is None:
            eventlog.log_debug("fail: internal-server-error")
            raise PipelineError

        ui_pid = _get_new_pid(pipeline_jnode_set)
        if ui_pid is None:
            eventlog.log_debug("fail: service-unavailable")
            return reply_fail(REPLY_SERVICE_UNAVAILABLE_503)

        # create db_post and db_channel
        db_config_jnode = pipeline_helpers.load_db_json()
        db_args = db.DbArgs.deserialize(db_config_jnode['db'])
        post_args = db.DbPostArgs.deserialize(db_config_jnode['post'])
        channel_args = db.DbChannelArgs(ident=int(ui_pid))

        def _handler(db_link):
            with db_link:
                db.edit_post(db_link, post_args)
                db.edit_channel(db_link, post_args, channel_args)

        db.with_db(db_args, _handler)

        if query_args.raw_body:
            reply_jnode = mv_pipeline_creator.create_mv_pipeline(query_args.raw_body, ui_pid)
            return reply_jnode

        else:
            reply_jnode = rtsp_pipeline_creator.create_rtsp_pipeline(ui_pid)
            return reply_jnode

    else:
        eventlog.log_debug("fail: method-not-allowed")
        return reply_fail(REPLY_METHOD_NOT_ALLOWED_405)


# noinspection PyBroadException
def _main():
    eventlog.open_log('cgi-pipeline', 'debug')
    exit_code = 1
    try:
        reply_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
        try:
            query_args = ui_types.UiQueryArgs.deserialize()
            reply_jnode = _run(query_args)
            exit_code = 0
        except (PipelineError, db.DbError):
            pass

        except:
            eventlog.log_debug_last_exc()

        print json.dumps(reply_jnode, separators=(',', ':'))
        if exit_code != 0:
            eventlog.log_debug('fail: internal-server-error')

    except:
        eventlog.log_debug_last_exc()

    finally:
        eventlog.close_log()

    sys.exit(exit_code)

if __name__ == '__main__':
    _main()
