import json
import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *

import edge.pipeline_helpers as helpers


def mc_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    # speed
    if ui_entry_point in [Pipeline.Speed.Ui.MEDIASTORE_0, Pipeline.Speed.Ui.VIDEOSOURCE_0,
                          Pipeline.Speed.Ui.RTSP_STREAMING_0, Pipeline.Speed.Ui.HTTP_STREAMING_0]:

        used_ui_modules.update([Pipeline.Speed.Ui.MEDIASTORE_0, Pipeline.Speed.Ui.VIDEOSOURCE_0,
                                Pipeline.Speed.Ui.RTSP_STREAMING_0, Pipeline.Speed.Ui.HTTP_STREAMING_0])
        used_edge_modules.add(Pipeline.Speed.Edge.MEDIA_CODEC_0)
        creator_ctor = McCreator.create

    elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Speed.Ui.MEDIASTORE_0, Pipeline.Speed.Ui.VIDEOSOURCE_0,
                                Pipeline.Speed.Ui.RTSP_STREAMING_0, Pipeline.Speed.Ui.HTTP_STREAMING_0])
        creator_ctor = McCreator.create

    return creator_ctor


class McCreator(BaseCreator):
    def __init__(self, id_main_host_setupstore):
        """
        :type id_main_host_setupstore: str, dict, dict, str
        """
        BaseCreator.__init__(self)
        self._id_main_host_setupstore = id_main_host_setupstore

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: MsCreator
        """
        # speed
        if ui_entry_point in [Pipeline.Speed.Ui.PIPELINE, Pipeline.Speed.Ui.MEDIASTORE_0,
                              Pipeline.Speed.Ui.VIDEOSOURCE_0, Pipeline.Speed.Ui.RTSP_STREAMING_0,
                              Pipeline.Speed.Ui.HTTP_STREAMING_0]:
            id_main_host_setupstore = McCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = McCreator(id_main_host_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: str, dict, dict, str
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point.ptype == PTYPE_SPEED:
            ui_media_store_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.MEDIASTORE_0]
            edge_identity = Pipeline.Speed.Edge.MEDIA_CODEC_0

        else:
            raise NotImplemented

        edge_gid = edge_identity.gid
        spool_dir = get_spool_dir(ui_pid)

        mjpeg_quality = ui_media_store_jnode['rw']['short-term']['videostream']['codec-mjpeg-quality']

        # create main.conf
        main_conf_jnode = {
            "FileApi": {
                "Store": join_path(spool_dir, 'mc%s-mc%s-store' % (edge_gid, edge_gid)),
                "FileCountLow": 10,
                "FileCountHigh": 20,
                "AddSceneReact": {
                    "Dir": join_path(spool_dir, 'mg%s-mc%s-scene' % (edge_gid, edge_gid))
                },
                "SceneEvent": [
                    join_path(spool_dir, 'mc%s-ms%s-scene-short' % (edge_gid, edge_gid))
                ]
            },
            "codec": {
                "Scale": {
                    "MaxWidth": 0,
                    "MaxHeight": 0,
                    "Interpolation": 1
                },
                "Encode": {
                    "UsedThreadMax": 1,
                    "Quality": mjpeg_quality
                }
            }
        }

        if ui_media_store_jnode['rw']['long-term']['enabled']:
            main_conf_jnode['FileApi']['SceneEvent'].append(
                join_path(spool_dir, 'mc%s-ms%s-scene-long' % (edge_gid, edge_gid)))

        # create setupstore.conf
        edge_id = helpers.get_edge_id(edge_group, edge_identity)
        setupstore_conf_text = get_setupstore_conf_text(edge_id)

        # create host.conf
        host_conf_jnode = {
            "StoreDir": "/opt/edge/libexec/mediacodec/",
            "Core": {
                "Name": "edge-core-mediacodec",
                "ParamSet": {
                    "Load": [
                        "edge-codec-ipp"
                    ]
                }
            }
        }

        return edge_id, main_conf_jnode, host_conf_jnode, setupstore_conf_text

    def serialize(self):
        edge_id, main_conf_jnode, host_conf_jnode, setupstore_conf_text = self._id_main_host_setupstore

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_MEDIACODEC, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_abspath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_abspath, 'w') as json_file:
            json.dump(main_conf_jnode, json_file, indent=2)

        # create setupstore.conf
        setupstore_conf_filepath = join_path(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)

        host_conf_filepath = join_path(edge_jnode_dir, 'host.conf')
        with open(host_conf_filepath, 'w') as json_file:
            json.dump(host_conf_jnode, json_file, indent=2)


class McDeleter:
    def __init__(self, edge_mc_dirpath):
        """
        :type edge_mc_dirpath: str
        """
        self._edge_mc_dirpath = edge_mc_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: McDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            return None

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.MEDIA_CODEC_0)
            edge_type = Pipeline.Speed.Edge.MEDIA_CODEC_0.type
            deleter = McDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point.ptype == PTYPE_LPR_RTSP:
            return None

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_mc_dirpath)
