import json
import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *

import edge.pipeline_helpers as helpers

DEFAULT_MS_CHECK_PERIOD = 20


def ms_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    # speed
    if ui_entry_point in [Pipeline.Speed.Ui.MEDIASTORE_0]:
        used_ui_modules.add(Pipeline.Speed.Ui.MEDIASTORE_0)
        used_edge_modules.update([Pipeline.Speed.Edge.MEDIASTORE_0, Pipeline.Speed.Edge.WEBSERVER_0])
        creator_ctor = MsCreator.create

    elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        used_ui_modules.add(Pipeline.Speed.Ui.MEDIASTORE_0)
        creator_ctor = MsCreator.create

    return creator_ctor


class MsCreator(BaseCreator):
    def __init__(self, id_main_setupstore):
        """
        :type id_main_setupstore: str, dict, str
        """
        BaseCreator.__init__(self)
        self._id_main_setupstore = id_main_setupstore

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: MsCreator
        """

        if ui_entry_point in [Pipeline.Speed.Ui.PIPELINE, Pipeline.Speed.Ui.MEDIASTORE_0]:
            id_main_setupstore = MsCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = MsCreator(id_main_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @ staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: str, dict, str
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point.ptype == PTYPE_SPEED:
            ui_media_store_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.MEDIASTORE_0]
            edge_identity = Pipeline.Speed.Edge.MEDIASTORE_0

        else:
            raise NotImplemented

        edge_gid = edge_identity.gid

        spool_dir = get_spool_dir(ui_pid)
        # create main.conf
        main_conf_jnode = {
            "SocketApi": {
                "ParamSet": {
                    "EndPoint": join_path(spool_dir, 'ms%s-ms%s-uds' % (edge_gid, edge_gid)),
                    "MaxClients": 2
                }
            },
            "FileApi": {
                "SaveFrameEvent": [
                ],
                "SaveFrameReact": [
                    {
                        "SkipRate": ui_media_store_jnode['rw']['short-term']['videostream']['skip-rate'] - 1,           # TODO - fix this workaround
                        "Dir": join_path(spool_dir, 'mc%s-ms%s-scene-short' % (edge_gid, edge_gid)),
                        "Channel": "1"
                    }
                ]
            },
            "Store": {
                "ParamSet": {
                    "Channels": [
                        {
                            "Name": "1",
                            "Url": join_path(spool_dir, 'ms%s-ms%s-store' % (edge_gid, edge_gid)),
                            "CleanupMode": 2,
                            "Cleanup": {
                                "CheckPeriod": 20,
                                "DiskLoUsage": 0.70,
                                "DiskHiUsage": 0.80,
                                "InodeLoMaxUsage": 0.70,
                                "InodeHiMaxUsage": 0.80
                            }
                        }
                    ]
                }
            }
        }
        # configure long-term store (only fom "mv" cameras)
        if ui_entry_point.ptype == PTYPE_SPEED:
            if ui_media_store_jnode['rw']['long-term']['enabled']:

                # noinspection PyUnresolvedReferences
                main_conf_jnode['Store']['ParamSet']['Channels'].append(
                    {
                        "Name": "0",
                        "Url": "/var/opt/edge/pipeline-%s/ms%s-ms%s-store/" % (ui_pid, edge_gid, edge_gid),
                        "CleanupMode": 2,
                        "Cleanup": {
                            "DiskLoUsage": 0.79,
                            "DiskHiUsage": 0.80,
                            "CheckPeriod": 20,
                            "InodeLoMaxUsage": 0.70,
                            "InodeHiMaxUsage": 0.80
                        }
                    })
                # noinspection PyUnresolvedReferences
                main_conf_jnode['FileApi']['SaveFrameReact'].append(
                    {
                        "Channel": "0",
                        "Dir": join_path(spool_dir, 'mc%s-ms%s-scene-long' % (edge_gid, edge_gid)),
                        "SkipRate": ui_media_store_jnode['rw']['long-term']['videostream']['skip-rate']
                    })

        # create setupstore.conf
        edge_id = helpers.get_edge_id(edge_group, edge_identity)
        setupstore_conf_text = get_setupstore_conf_text(edge_id)
        return edge_id, main_conf_jnode, setupstore_conf_text

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text = self._id_main_setupstore
        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_MEDIASTORE, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_filepath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_filepath, 'w') as json_file:
            json.dump(main_conf_jnode, json_file, indent=2)

        setupstore_conf_filepath = join_path(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)


class MsDeleter:
    def __init__(self, edge_ms_dirpath):
        """
        :type edge_ms_dirpath: str
        """
        self._edge_ms_dirpath = edge_ms_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: MsDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            return None

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.MEDIASTORE_0)
            edge_type = Pipeline.Speed.Edge.MEDIASTORE_0.type
            deleter = MsDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point.ptype == PTYPE_LPR_RTSP:
            return None

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_ms_dirpath)
