import json
from twisted.internet import reactor
from twisted.internet.protocol import ReconnectingClientFactory, connectionDone
from twisted.protocols.basic import LineOnlyReceiver

import edge.eventlog as eventlog
from edge.admin.ui_types import *
from edge.path_helpers import *

import _file_lock
import edge.pipeline_helpers as helpers


# noinspection PyUnusedLocal
def _get_ra_uds_endpoint(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """

    edge_identity = Pipeline.Speed.Edge.RADAR_0
    edge_gid = edge_identity.gid
    spool_dir = get_spool_dir(ui_pid)
    return join_path(spool_dir, "ra%s-ra%s-uds" % (edge_gid, edge_gid))


def _get_reactor():
    """
    :rtype: IReactorCore
    """
    return reactor


_QUERY_DONE = 1
_QUERY_INVALID_STATE = 2
_QUERY_INVALID_ARG = 3


# noinspection PyMethodMayBeStatic
class IRadar:
    def __init__(self):
        pass

    def query_set_alignment(self, mode, speed, query_arg):
        """
        :type mode: int
        :type speed: int
        :type query_arg: object | None
        """
        raise NotImplemented

    def query_set_location(self, forward, right, top, azimuth, elevation, pitch, matrix, query_arg):
        """
        :type forward: int
        :type right: int
        :type top: int
        :type azimuth: int
        :param elevation: int
        :type pitch: int
        :type matrix: list
        :type query_arg: object | None
        """
        raise NotImplemented


# noinspection PyMethodMayBeStatic
class IRadarHandler:
    def __init__(self):
        pass

    def connection_made(self, radar):
        """
        :type radar: IRadar
        """
        raise NotImplemented

    def connection_lost(self, query_arg):
        """
        :type query_arg: object | None
        """
        raise NotImplemented

    def handle_set_location(self, status, query_arg):
        """
        :type status: int
        :type query_arg: object
        """
        raise NotImplemented

    def handle_set_alignment(self, status, query_arg):
        """
        :type status: int
        :type query_arg: object
        """
        raise NotImplemented


class _UdsTransport(LineOnlyReceiver):
    _UTF8_ENCODING = 'utf-8'
    delimiter = b'\0'

    def __init__(self):
        pass

    def connectionMade(self):
        raise NotImplementedError

    # Called when an established connection is lost.
    def connectionLost(self, reason=connectionDone):
        raise NotImplementedError

    def lineReceived(self, raw_response):
        unicode_reply = raw_response.decode(_UdsTransport._UTF8_ENCODING)
        eventlog.log_debug("raw-reply: %s" % unicode_reply)

        try:
            reply_jnode = json.loads(unicode_reply)
            self.rx_reply(reply_jnode)
        except ValueError:
            eventlog.log_debug_last_exc()

    def rx_reply(self, reply_jnode):
        raise NotImplementedError

    def tx_query(self, query_jnode):
        """
        :type query_jnode: dict
        """
        if not isinstance(query_jnode, dict):
            eventlog.log_debug("fail: invalid query")
            raise ValueError

        unicode_query = json.dumps(query_jnode, separators=(',', ':'))
        eventlog.log_debug("raw-query: %s" % unicode_query)

        utf8_query = unicode_query.encode(_UdsTransport._UTF8_ENCODING)
        self.sendLine(utf8_query)
        return _QUERY_DONE


class _Radar(_UdsTransport, IRadar):
    _QUERY_SET_LOCATION = 1,
    _QUERY_SET_ALIGNMENT = 2

    def __init__(self, radar_handler):
        _UdsTransport.__init__(self)
        IRadar.__init__(self)

        if not isinstance(radar_handler, IRadarHandler):
            eventlog.log_debug("fail: invalid value of radar-handler")
            raise ValueError

        self._radar_handler = radar_handler
        self._query_action = None
        self._query_arg = None

    def connectionMade(self):
        eventlog.log_debug("done: open ra-socket")
        self._query_arg = None
        self._radar_handler.connection_made(self)

    def connectionLost(self, reason=connectionDone):
        eventlog.log_debug("done: close ra-socket % s" % reason)

        query_arg = self._query_arg
        self._query_arg = None

        self._radar_handler.connection_lost(query_arg)

    def rx_reply(self, reply_jnode):
        """
        :type reply_jnode: dict
        """
        if self._query_action == _Radar._QUERY_SET_LOCATION:
            status = reply_jnode['status']

            query_arg = self._query_arg
            self._query_arg = None

            self._radar_handler.handle_set_location(status, query_arg)

        elif self._query_action == _Radar._QUERY_SET_ALIGNMENT:
            status = reply_jnode['status']

            query_arg = self._query_arg
            self._query_arg = None

            self._radar_handler.handle_set_alignment(status, query_arg)

        else:
            eventlog.log_debug("fail: invalid query-action")

    def query_set_alignment(self, mode, speed, query_arg):
        """
        :type mode: int
        :type speed: int
        :type query_arg: object
        """
        if self._query_arg:
            eventlog.log_debug("fail: another query in progress")
            return _QUERY_INVALID_STATE

        if not query_arg:
            eventlog.log_debug("fail: invalid value of query-arg")
            return _QUERY_INVALID_ARG

        self._query_action = _Radar._QUERY_SET_ALIGNMENT
        self._query_arg = query_arg

        query_jnode = {
            'method': 'set',
            "action": "alignment",
            "args": {
                "sensor": {
                    "mode": mode,
                    "mode-alignment": {
                        "speed-filter": speed
                    }
                }
            }
        }

        self.tx_query(query_jnode)
        return _QUERY_DONE

    def query_set_location(self, forward, left, up, azimuth, elevation, pitch, matrix, query_arg):
        """
        :type forward: int
        :type left: int
        :type up: int
        :type azimuth: int
        :param elevation: int
        :type pitch: int
        :type matrix: list
        :type query_arg: object
        """
        if self._query_arg:
            eventlog.log_debug("fail: another query in progress")
            return _QUERY_INVALID_STATE

        if not query_arg:
            eventlog.log_debug("fail: invalid value of query-arg")
            return _QUERY_INVALID_ARG

        self._query_action = _Radar._QUERY_SET_LOCATION
        self._query_arg = query_arg

        query_jnode = {
            'method': 'set',
            'action': 'location',
            'args': {
                'sensor': {
                    'position': {
                        'forward': forward,
                        'left': left,
                        'up': up
                    },
                    "rotation": {
                        "elevation": elevation,
                        "azimuth": azimuth,
                        "pitch": pitch
                    }
                },
                "image-mapper": {
                    "rhc-to-image-map": matrix
                }
            }
        }

        self.tx_query(query_jnode)
        return _QUERY_DONE


class _RadarClientFactory(ReconnectingClientFactory):
    initialDelay = 1
    maxDelay = 15
    factor = 1.6180339887498948

    def __init__(self, radar_handler):
        self._radar_action_handler = radar_handler

    def buildProtocol(self, addr):
        self.resetDelay()
        protocol = _Radar(self._radar_action_handler)
        protocol.factory = self
        return protocol

    def clientConnectionFailed(self, connector, reason):
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, unused_reason):
        ReconnectingClientFactory.clientConnectionLost(self, connector, unused_reason)


def ra_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    if ui_entry_point in [Pipeline.Speed.Ui.LOCATION_0]:
        used_ui_modules.update([Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.VIDEOSOURCE_0])
        used_edge_modules.add(Pipeline.Speed.Edge.RADAR_0)
        creator_ctor = RaCreator.create

    elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.VIDEOSOURCE_0])
        creator_ctor = RaCreator.create

    return creator_ctor


class RaLocationEditor(IRadarHandler):
    def __init__(self, ra_lock_dirpath, forward, left, up, azimuth, elevation, roll, matrix):
        """
        :type forward: int
        :type left: int
        :type up: int
        :type azimuth: int
        :type elevation: int
        :type roll: int
        :type matrix: list
        :type ra_lock_dirpath: str
        """
        IRadarHandler.__init__(self)

        self._radar_lock_dirpath = ra_lock_dirpath
        self._forward = forward
        self._left = left
        self._up = up
        self._azimuth = azimuth
        self._elevation = elevation
        self._roll = roll
        self._matrix = matrix

        self._uds_reply_stat = 2

    def connection_made(self, radar):
        """
        :type radar: IRadar
        """
        radar.query_set_location(self._forward, self._left, self._up, self._azimuth,
                                 self._elevation, self._roll, self._matrix, "str")
        eventlog.log_debug('')

    def connection_lost(self, query_arg):
        """
        :type query_arg: object | None
        """
        eventlog.log_debug('')

    def handle_set_location(self, status, query_arg):
        """
        :type status: int
        :type query_arg: object
        """
        eventlog.log_debug("status: %s" % status)
        self._uds_reply_stat = status
        twisted_reactor = _get_reactor()
        twisted_reactor.stop()

    def handle_set_alignment(self, status, query_arg):
        """
        :type status: int
        :type query_arg: object | None
        """
        eventlog.log_debug("status: %s" % status)
        self._uds_reply_stat = status
        twisted_reactor = _get_reactor()
        twisted_reactor.stop()

    @staticmethod
    def deserialize(new_ui_jnode, base_lock_dirpath):

        eventlog.log_debug("new ui jnode %s " % new_ui_jnode)
        # deserialize sensor-position
        unit_position_jnode = new_ui_jnode['rw']['unit-position']
        forward = int(unit_position_jnode['forward'] * 100)
        left = int(unit_position_jnode['left'] * 100)
        up = int(unit_position_jnode['up'] * 100)

        # deserialize sensor-rotation
        unit_rotation_manual_jnode = new_ui_jnode['rw']['unit-rotation-manual']
        raw_azimuth = unit_rotation_manual_jnode['azimuth']
        if raw_azimuth < 0:
            azimuth = int((raw_azimuth + 360) * 100)

        else:
            azimuth = int((raw_azimuth + 0) * 100)

        raw_elevation = unit_rotation_manual_jnode['elevation']
        if raw_elevation < 0:
            elevation = int((raw_elevation + 360) * 100)

        else:
            elevation = int((raw_elevation + 0) * 100)

        raw_roll = unit_rotation_manual_jnode['roll']
        if raw_roll < 0:
            roll = int((raw_roll + 360) * 100)

        else:
            roll = int((raw_roll + 0) * 100)

        rhcc_matrix = new_ui_jnode['ro'].get('rhcc-to-ndc') or new_ui_jnode['rt']['rhcc-to-ndc']
        radar_lock_dirpath = base_lock_dirpath + 'radar/'
        editor = RaLocationEditor(radar_lock_dirpath, forward, left, up, azimuth, elevation, roll, rhcc_matrix)
        return editor

    def serialize(self, ui_pid):
        """
        :type ui_pid: str
        """
        lock = _file_lock.try_lock(self._radar_lock_dirpath, ui_pid)
        try:
            uds_endpoint = _get_ra_uds_endpoint(ui_pid)
            client_factory = _RadarClientFactory(radar_handler=self)
            twisted_reactor = _get_reactor()
            twisted_reactor.connectUNIX(uds_endpoint, client_factory)

            twisted_reactor.callLater(15, twisted_reactor.stop)
            twisted_reactor.addSystemEventTrigger('before', 'shutdown', eventlog.log_debug, 'Shutdown...')

            twisted_reactor.run()

            if self._uds_reply_stat != 1:
                raise RuntimeError("fail: invalid uds-reply-stat (%s)" % self._uds_reply_stat)

        finally:
            _file_lock.unlock(lock)


class RaCreator(BaseCreator):
    def __init__(self, id_main_setupstore):
        """
        :type id_main_setupstore: str, dict, str
        """
        self._id_main_setupstore = id_main_setupstore
        BaseCreator.__init__(self)

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: RaCreator
        """
        if ui_entry_point in [Pipeline.Speed.Ui.LOCATION_0, Pipeline.Speed.Ui.PIPELINE]:
            id_main_setupstore = RaCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = RaCreator(id_main_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict | None
        :rtype: str, dict, str
        """

        ui_location_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.LOCATION_0]
        edge_identity = Pipeline.Speed.Edge.RADAR_0

        edge_gid = edge_identity.gid
        spool_dir = get_spool_dir(ui_pid)

        edge_group = helpers.get_edge_group(ui_pid)
        edge_id = helpers.get_edge_id(edge_group, edge_identity)

        # send sms to module radar with new config
        if ui_entry_point == Pipeline.Speed.Ui.LOCATION_0:
            radar_lock_dirpath = '/etc/edge/radar.%s/lock' % ui_pid
            editor = RaLocationEditor.deserialize(ui_location_jnode, radar_lock_dirpath)
            editor.serialize(ui_pid)

        # create main.conf
        main_conf_jnode = {
            "api-server": {
                "endpoint": join_path(spool_dir, "ra%s-ra%s-uds" % (edge_gid, edge_gid))
            },
            "device": {
                "vendor": "smartmicro",
                "endpoint": ui_location_jnode['ro'].get('uri', 'ummr0a:///dev/ttyS0'),
                "runtime": "/etc/edge/radar.%s/runtime/" % edge_gid
            },
            "strobe-react": {
                "endpoint": join_path(spool_dir, "mg%s-ra%s-pipe" % (edge_gid, edge_gid))
            },
            "scene-event": {
                "endpoints": [
                    join_path(spool_dir, "ra%s-tr%s-scene" % (edge_gid, edge_gid))
                ]
            }
        }

        # create setupstore.conf
        setupstore_conf_text = get_setupstore_conf_text(edge_id)

        # create haltaction.conf
        halt_list = ui_location_jnode['ro'].get('haltaction', None)
        if halt_list:
            haltaction_conf = '\n'.join(halt_list)

        else:
            haltaction_conf = ''

        return edge_id, main_conf_jnode, setupstore_conf_text, haltaction_conf

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text, haltaction_conf_text = self._id_main_setupstore

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_RADAR, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_filepath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_filepath, 'w') as json_file:
            json.dump(main_conf_jnode, json_file, indent=2)

        # create setupstore.conf
        setupstore_conf_filepath = join_path(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)

        halt_action_path = join_path(edge_jnode_dir, 'haltaction.conf')
        with open(halt_action_path, 'w') as text_file:
            text_file.write(haltaction_conf_text)


class RaDeleter:
    def __init__(self, edge_ra_dirpath):
        """
        :type edge_ra_dirpath: str
        """
        self._edge_ra_dirpath = edge_ra_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: RaDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.RADAR_0)
            edge_type = Pipeline.Speed.Edge.RADAR_0.type
            deleter = RaDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            return None

        elif ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            return None

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_ra_dirpath)
