import json
import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *

import edge.pipeline_helpers as helpers


def rg_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    if ui_entry_point in [Pipeline.Rtsp.Ui.VIDEOSOURCE_0]:
        used_ui_modules.add(Pipeline.Rtsp.Ui.VIDEOSOURCE_0)
        used_edge_modules.add(Pipeline.Rtsp.Edge.RTSPGRABBER_0)
        creator_ctor = RgCreator.create

    elif ui_entry_point in [Pipeline.Rtsp.Ui.PIPELINE]:
        used_ui_modules.add(Pipeline.Rtsp.Ui.VIDEOSOURCE_0)
        creator_ctor = RgCreator.create

    return creator_ctor


class RgCreator(BaseCreator):
    def __init__(self, id_main_setupstore_halt):
        """
        :type id_main_setupstore_halt: str, dict, str, str
        """
        BaseCreator.__init__(self)

        self._id_main_setupstore_halt = id_main_setupstore_halt

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: RgCreator
        """

        if ui_entry_point in [Pipeline.Rtsp.Ui.PIPELINE, Pipeline.Rtsp.Ui.VIDEOSOURCE_0]:
            id_main_setupstore_halt = RgCreator._serialize(ui_pid, ui_identity_to_jnode_map)
            creator = RgCreator(id_main_setupstore_halt)

        else:
            raise NotImplemented('entry_point not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: str, dict, str
        """
        ui_videosource_jnode = ui_identity_to_jnode_map[Pipeline.Rtsp.Ui.VIDEOSOURCE_0]

        if ui_videosource_jnode['rw']['streaming']['transport'] == 'rtp-over-tcp':
            use_udp = False
        elif ui_videosource_jnode['rw']['streaming']['transport'] == 'rtp-over-ucast-udp-or-tcp':
            use_udp = True
        else:
            raise NotImplemented

        spool_dir = get_spool_dir(ui_pid)
        edge_group = helpers.get_edge_group(ui_pid)
        edge_gid = Pipeline.Rtsp.Edge.RTSPGRABBER_0.gid
        # create main.conf
        main_conf_jnode = {
            "InputStreamType": ui_videosource_jnode['rw']['videostream']['codec'],
            "UseUDP": use_udp,
            "TargetHeight": 0,
            "Url": ui_videosource_jnode['rw']['streaming']['url'],
            "TargetWidth": 0,
            "UseGpu": ui_videosource_jnode['rw']['videostream']['hw-decode'],
            "Store": {
                "Directory": join_path(spool_dir, 'rg%s-rg%s-store' % (edge_gid, edge_gid)),
                "FileCountLow": 26,
                "FileCountHigh": 30
            },
            "FileApi": {
                "SceneEventMono8t": [],
                "SceneEventRaw": [],
                "HaltEvent": {
                    "SignalAbsentTout": 5000,
                    "DeviceAbsentTout": 600000
                },
            },
            "Monitoring": {
                "StatTout": 5,
                "StatPath": join_path(spool_dir, 'zabbix', 'vs%s' % edge_gid)
            }
        }

        if ui_videosource_jnode['rw']['videostream']['color-mode'] == 'raw':
            main_conf_jnode['FileApi']['SceneEventRaw'].append(
                join_path(spool_dir, 'rg%s-lr%s-scene' % (edge_gid, edge_gid)))

            main_conf_jnode['FileApi']['SceneEventRaw'].append(
                join_path(spool_dir, 'rg%s-tr%s-scene' % (edge_gid, edge_gid)))

        elif ui_videosource_jnode['rw']['videostream']['color-mode'] == 'grayscale':
            main_conf_jnode['FileApi']['SceneEventMono8t'].append(
                join_path(spool_dir, 'rg%s-lr%s-scene' % (edge_gid, edge_gid)))

            main_conf_jnode['FileApi']['SceneEventMono8t'].append(
                join_path(spool_dir, 'rg%s-tr%s-scene' % (edge_gid, edge_gid)))

        else:
            raise NotImplemented

        # create setupstore.conf
        edge_id = helpers.get_edge_id(edge_group, Pipeline.Rtsp.Edge.RTSPGRABBER_0)
        setupstore_conf_text = get_setupstore_conf_text(edge_id)

        # create haltaction.conf
        halt_list = ui_videosource_jnode['ro'].get('haltaction', None)
        if halt_list:
            haltaction_conf = '\n'.join(halt_list)

        else:
            haltaction_conf = ''

        return edge_id, main_conf_jnode, setupstore_conf_text, haltaction_conf

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text, haltaction_conf_text = \
            self._id_main_setupstore_halt

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_RTSP_GRABBER, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_abspath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_abspath, 'w') as main_conf:
            json.dump(main_conf_jnode, main_conf, indent=2)

        setupstore_conf_path = join_path(edge_jnode_dir, 'setupstore.conf')

        with open(setupstore_conf_path, 'w') as text_file:
            text_file.write(setupstore_conf_text)

        halt_action_path = join_path(edge_jnode_dir, 'haltaction.conf')
        with open(halt_action_path, 'w') as text_file:
            text_file.write(haltaction_conf_text)


class RgDeleter:
    def __init__(self, edge_rg_dirpath):
        """
        :type edge_rg_dirpath: str
        """
        self._edge_rg_dirpath = edge_rg_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: RgDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Rtsp.Edge.RTSPGRABBER_0)
            edge_type = Pipeline.Rtsp.Edge.RTSPGRABBER_0.type
            deleter = RgDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            return None

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            return None

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_rg_dirpath)
