import os

from edge.admin.ui_types import BaseCreator
from edge.path_helpers import *

import edge.pipeline_helpers as helpers
import edge.db as edge_db

from copy import deepcopy


def sc_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules):
    """
    :type ui_entry_point: UiIdentity
    :type used_ui_modules: set of UiIdentity
    :type used_edge_modules: set of EdgeIdentity
    :rtype: None | (str, UiIdentity, dict of UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    if ui_entry_point in [Pipeline.Speed.Ui.SPEED_ENFORCEMENT_0, Pipeline.Speed.Ui.LANE_ENFORCEMENT_0,
                          Pipeline.Speed.Ui.FACT_RULES_0, Pipeline.Speed.Ui.GEOLOCATION_0,
                          Pipeline.Speed.Ui.EVENTSTORE_0]:
        used_ui_modules.update([Pipeline.Speed.Ui.SPEED_ENFORCEMENT_0,
                                Pipeline.Speed.Ui.LANE_ENFORCEMENT_0,
                                Pipeline.Speed.Ui.EVENTSTORE_0,
                                Pipeline.Speed.Ui.FACT_RULES_0,
                                Pipeline.Speed.Ui.GEOLOCATION_0])
        used_edge_modules.add(Pipeline.Speed.Edge.SPEEDCONTROL_0)
        creator_ctor = ScCreator.create

    elif ui_entry_point in [Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0, Pipeline.Rtsp.Ui.FACT_RULES_0,
                            Pipeline.Rtsp.Ui.GEOLOCATION_0, Pipeline.Rtsp.Ui.EVENTSTORE_0]:
        used_ui_modules.update([Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0, Pipeline.Rtsp.Ui.FACT_RULES_0,
                                Pipeline.Rtsp.Ui.GEOLOCATION_0])
        used_edge_modules.add(Pipeline.Rtsp.Edge.SPEEDCONTROL_0)
        creator_ctor = ScCreator.create

    elif ui_entry_point in [Pipeline.Mv.Ui.LANE_ENFORCEMENT_0, Pipeline.Mv.Ui.FACT_RULES_0,
                            Pipeline.Mv.Ui.GEOLOCATION_0, Pipeline.Mv.Ui.EVENTSTORE_0]:
        used_ui_modules.update([Pipeline.Mv.Ui.LANE_ENFORCEMENT_0, Pipeline.Mv.Ui.FACT_RULES_0,
                                Pipeline.Mv.Ui.GEOLOCATION_0])
        used_edge_modules.add(Pipeline.Mv.Edge.SPEEDCONTROL_0)
        creator_ctor = ScCreator.create

    elif ui_entry_point in [Pipeline.Speed.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Speed.Ui.SPEED_ENFORCEMENT_0,
                                Pipeline.Speed.Ui.LANE_ENFORCEMENT_0,
                                Pipeline.Speed.Ui.EVENTSTORE_0,
                                Pipeline.Speed.Ui.FACT_RULES_0,
                                Pipeline.Speed.Ui.GEOLOCATION_0])
        creator_ctor = ScCreator.create

    elif ui_entry_point in [Pipeline.Rtsp.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0, Pipeline.Rtsp.Ui.FACT_RULES_0,
                                Pipeline.Rtsp.Ui.GEOLOCATION_0])
        creator_ctor = ScCreator.create

    elif ui_entry_point in [Pipeline.Mv.Ui.PIPELINE]:
        used_ui_modules.update([Pipeline.Mv.Ui.LANE_ENFORCEMENT_0, Pipeline.Mv.Ui.FACT_RULES_0,
                                Pipeline.Mv.Ui.GEOLOCATION_0])
        creator_ctor = ScCreator.create

    return creator_ctor


class _Db:
    def __init__(self, db_args):
        """
        :type db_args: edge_db.DbArgs 
        """
        self.db_args = db_args

    def add_speedviolation(self, db_post, db_channel, speed_limit, overspeed_limit):
        """
        :type db_post: str 
        :type db_channel: int
        :type speed_limit: int 
        :type overspeed_limit: int 
        :raises edge_db.DbError
        """
        def _handler(db_link):
            with db_link:
                self._set_platear_options(db_link)
                self._add_update_speedlimit(db_link, db_post, db_channel, speed_limit)
                self._add_update_max_overspeed(db_link, db_post, db_channel, overspeed_limit)

        try:
            edge_db.with_db(self.db_args, _handler)
        except edge_db.DbError as x:
            eventlog.log_debug('fail: add speedviolation')
            raise x

    def remove_speedviolation(self, db_post, db_channel):
        """
        :type db_post: str 
        :type db_channel: int
        :raises edge_db.DbError
        """
        def _handler(db_link):
            with db_link:
                self._set_platear_options(db_link)
                self._remove_speedlimit(db_link, db_post, db_channel)
                self._remove_max_overspeed(db_link, db_post, db_channel)

        try:
            edge_db.with_db(self.db_args, _handler)
        except edge_db.DbError as x:
            eventlog.log_debug('fail: remove speedviolation')
            raise x

    @staticmethod
    def _add_update_speedlimit(db_link, db_post, db_channel, speed_limit):
        """
        :type db_link: Any
        :type db_post: str
        :type db_channel: int
        :type speed_limit: int 
        :raises: edge_db.DbError
        """
        with db_link.cursor() as db_cursor:
            db_cursor.execute(''.join([
                "UPDATE estimates ",
                "SET \"value\"=%s ",
                "WHERE \"name\"=%s AND (pattern->'post'->>1)=%s AND (pattern->'channel'->>1)::int = %s;"
            ]), [str(speed_limit), 'speedlimit', db_post, db_channel])

            if db_cursor.rowcount == -1 or db_cursor.rowcount > 1:
                eventlog.log_debug('fail: invalid number of estimates updated (%s)' % db_cursor.rowcount)
                raise edge_db.DbError

            if db_cursor.rowcount != 1:
                pattern_jnode = {
                    'post': ['eq', db_post],
                    'channel': ['eq', db_channel]
                }

                db_cursor.execute(''.join([
                    "INSERT INTO estimates(\"name\", pattern, \"value\") ",
                    "VALUES (%s, %s, %s)"
                ]), ['speedlimit', edge_db.DbJsonArg(pattern_jnode), str(speed_limit)])

    @staticmethod
    def _add_update_max_overspeed(db_link, db_post, db_channel, overspeed_limit):
        """
        :type db_link: Any
        :type db_post: str
        :type db_channel: int
        :type overspeed_limit: int
        :raises: edge_db.DbError
        """
        with db_link.cursor() as db_cursor:
            db_cursor.execute(''.join([
                "UPDATE estimates ",
                "SET \"value\"=%s ",
                "WHERE \"name\"=%s AND (pattern->'post'->>1)=%s AND (pattern->'channel'->>1)::int = %s;"
            ]), [str(overspeed_limit), 'speedmargin', db_post, db_channel])

            if db_cursor.rowcount == -1 or db_cursor.rowcount > 1:
                eventlog.log_debug('fail: invalid number of estimates updated (%s)' % db_cursor.rowcount)
                raise edge_db.DbError

            if db_cursor.rowcount != 1:
                pattern_jnode = {
                    'post': ['eq', db_post],
                    'channel': ['eq', db_channel]
                }

                db_cursor.execute(''.join([
                    "INSERT INTO estimates(\"name\", pattern, \"value\") ",
                    "VALUES (%s, %s, %s)"
                ]), ['speedmargin', edge_db.DbJsonArg(pattern_jnode), str(overspeed_limit)])

    @staticmethod
    def _set_platear_options(db_link):
        """
        :type db_link: Any
        :raises edge_db.DbError
        """
        with db_link.cursor() as db_cursor:
            db_cursor.execute("select core.setopt('public', 'platear_check_speedlimit', 'true')")
            string_changed = db_cursor.fetchone()[0]
            if string_changed != 1:
                eventlog.log_debug('fail: set platear_check_speedlimit')
                raise edge_db.DbError

            db_cursor.execute("select core.setopt('public', 'platear_common_speedlimit', '0')")
            string_changed = db_cursor.fetchone()[0]
            if string_changed != 1:
                eventlog.log_debug('fail: set platear_common_speedlimit')
                raise edge_db.DbError

    @staticmethod
    def _remove_speedlimit(db_link, db_post, db_channel):
        """
        :type db_link: Any 
        :type db_post: str
        :type db_channel: int
        :raises edge_db.DbError
        """
        with db_link.cursor() as db_cursor:
            db_cursor.execute(''.join([
                "DELETE FROM estimates ",
                "WHERE \"name\"=%s AND (pattern->'post'->>1)=%s AND (pattern->'channel'->>1)::int = %s;"
            ]), ['speedlimit', db_post, db_channel])

            if db_cursor.rowcount > 1:
                eventlog.log_debug('fail: invalid number of estimates deleted (%s)' % db_cursor.rowcount)
                raise edge_db.DbError

    @staticmethod
    def _remove_max_overspeed(db_link, db_post, db_channel):
        """
        :type db_link: Any
        :type db_post: str
        :type db_channel: int
        :raises edge_db.DbError
        """
        with db_link.cursor() as db_cursor:
            db_cursor.execute(''.join([
                "DELETE FROM estimates ",
                "WHERE \"name\"=%s AND (pattern->'post'->>1)=%s AND (pattern->'channel'->>1)::int = %s;"
            ]), ['speedmargin', db_post, db_channel])

            if db_cursor.rowcount > 1:
                eventlog.log_debug('fail: invalid number of estimates deleted (%s)' % db_cursor.rowcount)
                raise edge_db.DbError


class ScCreator(BaseCreator):
    def __init__(self,  id_main_setupstore):
        """
        :type id_main_setupstore: str, dict, str
        """
        BaseCreator.__init__(self)
        self._id_main_setupstore = id_main_setupstore

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: ScCreator
        """
        if ui_entry_point in [Pipeline.Speed.Ui.SPEED_ENFORCEMENT_0, Pipeline.Speed.Ui.PIPELINE,
                              Pipeline.Speed.Ui.LANE_ENFORCEMENT_0, Pipeline.Mv.Ui.LANE_ENFORCEMENT_0,
                              Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0, Pipeline.Rtsp.Ui.PIPELINE,
                              Pipeline.Mv.Ui.PIPELINE, Pipeline.Speed.Ui.FACT_RULES_0,
                              Pipeline.Rtsp.Ui.FACT_RULES_0, Pipeline.Mv.Ui.FACT_RULES_0,
                              Pipeline.Speed.Ui.GEOLOCATION_0, Pipeline.Mv.Ui.GEOLOCATION_0,
                              Pipeline.Rtsp.Ui.GEOLOCATION_0, Pipeline.Speed.Ui.EVENTSTORE_0,
                              Pipeline.Rtsp.Ui.EVENTSTORE_0, Pipeline.Mv.Ui.EVENTSTORE_0]:
            id_main_setupstore = ScCreator._serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map)
            creator = ScCreator(id_main_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: str, dict, str
        """
        main_conf_jnode = {}
        if ui_entry_point in [Pipeline.Speed.Ui.SPEED_ENFORCEMENT_0, Pipeline.Speed.Ui.PIPELINE,
                              Pipeline.Speed.Ui.LANE_ENFORCEMENT_0, Pipeline.Speed.Ui.FACT_RULES_0,
                              Pipeline.Speed.Ui.GEOLOCATION_0, Pipeline.Speed.Ui.EVENTSTORE_0]:
            ui_speed_enforcement_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.SPEED_ENFORCEMENT_0]
            ui_lane_enforcement_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.LANE_ENFORCEMENT_0]
            ui_eventstore_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.EVENTSTORE_0]
            ui_fact_rules_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.FACT_RULES_0]
            ui_geolocation_jnode = ui_identity_to_jnode_map[Pipeline.Speed.Ui.GEOLOCATION_0]
            db_channel = ui_eventstore_jnode['rw']['channel']
            edge_identity = Pipeline.Speed.Edge.SPEEDCONTROL_0
            # copy initial state of main.conf from ui_jnode['ro']
            main_conf_jnode = deepcopy(ui_speed_enforcement_jnode['ro']['main'])

        elif ui_entry_point in [Pipeline.Rtsp.Ui.PIPELINE, Pipeline.Rtsp.Ui.FACT_RULES_0,
                                Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0, Pipeline.Rtsp.Ui.GEOLOCATION_0,
                                Pipeline.Rtsp.Ui.EVENTSTORE_0]:
            ui_speed_enforcement_jnode = None
            ui_fact_rules_jnode = ui_identity_to_jnode_map[Pipeline.Rtsp.Ui.FACT_RULES_0]
            ui_lane_enforcement_jnode = ui_identity_to_jnode_map[Pipeline.Rtsp.Ui.LANE_ENFORCEMENT_0]
            ui_geolocation_jnode = ui_identity_to_jnode_map[Pipeline.Rtsp.Ui.GEOLOCATION_0]
            edge_identity = Pipeline.Rtsp.Edge.SPEEDCONTROL_0

        elif ui_entry_point in [Pipeline.Mv.Ui.PIPELINE, Pipeline.Mv.Ui.FACT_RULES_0,
                                Pipeline.Mv.Ui.LANE_ENFORCEMENT_0, Pipeline.Mv.Ui.GEOLOCATION_0,
                                Pipeline.Mv.Ui.EVENTSTORE_0]:
            ui_speed_enforcement_jnode = None
            ui_fact_rules_jnode = ui_identity_to_jnode_map[Pipeline.Mv.Ui.FACT_RULES_0]
            ui_lane_enforcement_jnode = ui_identity_to_jnode_map[Pipeline.Mv.Ui.LANE_ENFORCEMENT_0]
            ui_geolocation_jnode = ui_identity_to_jnode_map[Pipeline.Mv.Ui.GEOLOCATION_0]
            edge_identity = Pipeline.Mv.Edge.SPEEDCONTROL_0

        else:
            raise NotImplemented

        edge_gid = edge_identity.gid
        spool_dir = get_spool_dir(ui_pid)

        main_conf_jnode["base"] = "/etc/edge/setupstore/admin/actions/time.json"

        main_conf_jnode['Monitoring'] = {
            'StatTout': 60,
            'StatPath': join_path(spool_dir, 'zabbix', 'sc%s' % edge_gid)
        }
        main_conf_jnode['FileApi'] = {
            'AddTrackReact': join_path(spool_dir, 'tr%s-sc%s-track' % (edge_gid, edge_gid))
        }

        fact_event_jnode = main_conf_jnode.get('fact-event', {})
        fact_event_jnode.update({
            'endpoints': [join_path(spool_dir, 'sc%s-es%s-fact' % (edge_gid, edge_gid))],
            'images': ui_fact_rules_jnode['rw']['args']['images']
        })
        main_conf_jnode['fact-event'] = fact_event_jnode

        # set geolocation static args
        if ui_geolocation_jnode['rw']['mode'] == 'manual':
            main_conf_jnode['geolocation'] = [
                "static",
                ui_geolocation_jnode['rw']['latitude'],
                ui_geolocation_jnode['rw']['longitude']
            ]

        # take mediastore or speedcontrol store for jpeg
        if ui_entry_point.ptype in [PTYPE_LPR_RTSP, PTYPE_LPR_MV, PTYPE_SPEED]:
            main_conf_jnode['store'] = {
                'endpoint': join_path(spool_dir, 'sc%s-sc%s-store' % (edge_gid, edge_gid))
            }

        if ui_entry_point.ptype in [PTYPE_SPEED]:
            main_conf_jnode['mediastore'] = [join_path(spool_dir, 'ms%s-ms%s-uds' % (edge_gid, edge_gid)), "1"]

        if ui_entry_point.ptype == PTYPE_SPEED:
            # set speed-limits in DB
            db_jnode = helpers.load_db_json()

            db_post = db_jnode['post']['uuid']
            db_args = edge_db.DbArgs.deserialize(db_jnode['db'])
            db_writer = _Db(db_args)

            if ui_speed_enforcement_jnode['rw']['enabled']:
                speed_limit = ui_speed_enforcement_jnode['rw']['speed-limit']
                max_overspeed = ui_speed_enforcement_jnode['rw']['max-overspeed']
                main_conf_jnode['rules'] = {
                    'speed-violation': [
                        {
                            'speed-limit': speed_limit,
                            'max-overspeed': max_overspeed
                        }
                    ]
                }

                # noinspection PyUnboundLocalVariable
                db_writer.add_speedviolation(db_post, db_channel, speed_limit, max_overspeed)
            else:
                # noinspection PyUnboundLocalVariable
                db_writer.remove_speedviolation(db_post, db_channel)

        # workaround solutions (deleting speed args in DB from previous speed pipeline with same id)
        # if ui_entry_point in [Pipeline.Rtsp.Ui.PIPELINE, Pipeline.Mv.Ui.PIPELINE]:
        #     # optimistic delete speed violations from DB on creation of rtsp or mv channel
        #     db_jnode = helpers.load_db_json()
        #
        #     db_post = db_jnode['post']['uuid']
        #     db_args = edge_db.DbArgs.deserialize(db_jnode['db'])
        #     db_writer = _Db(db_args)
        #     db_writer.remove_speedviolation(db_post, db_channel)

        if len(ui_lane_enforcement_jnode['rw']['road']['lanes']) > 0:
            main_conf_jnode['road-mask'] = {
                'lanes': ui_lane_enforcement_jnode['rw']['road']['lanes'],
                'resolution': {
                    'w': ui_lane_enforcement_jnode['rw']['road']['resolution']['w'],
                    'h': ui_lane_enforcement_jnode['rw']['road']['resolution']['h']
                }
            }

        # create setupstore.conf
        edge_group = helpers.get_edge_group(ui_pid)
        edge_id = helpers.get_edge_id(edge_group, edge_identity)
        setupstore_conf_text = get_setupstore_conf_text(edge_id)

        return edge_id, main_conf_jnode, setupstore_conf_text

    def serialize(self):
        edge_id, main_conf_jnode, setupstore_conf_text = self._id_main_setupstore

        edge_jnode_dir = get_edge_jnode_dirpath(EDGE_TYPE_SPEEDCONTROL, edge_id)
        if not os.path.exists(edge_jnode_dir):
            helpers.silent_mkdir_r(edge_jnode_dir)

        main_conf_filepath = join_path(edge_jnode_dir, 'main.conf')
        with open(main_conf_filepath, 'w') as json_file:
            json.dump(main_conf_jnode, json_file, indent=2)

        # create setupstore.conf
        setupstore_conf_filepath = join_path(edge_jnode_dir, 'setupstore.conf')
        with open(setupstore_conf_filepath, 'w') as text_file:
            text_file.write(setupstore_conf_text)


class ScDeleter:
    def __init__(self, edge_sc_dirpath):
        """
        :type edge_sc_dirpath: str
        """
        self._edge_sc_dirpath = edge_sc_dirpath

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: UiIdentity
        :rtype: ScDeleter
        """
        edge_group = helpers.get_edge_group(ui_pid)

        if ui_entry_point == Pipeline.Rtsp.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Rtsp.Edge.SPEEDCONTROL_0)
            edge_type = Pipeline.Rtsp.Edge.SPEEDCONTROL_0.type
            deleter = ScDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Mv.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Mv.Edge.SPEEDCONTROL_0)
            edge_type = Pipeline.Mv.Edge.SPEEDCONTROL_0.type
            deleter = ScDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        elif ui_entry_point == Pipeline.Speed.Ui.PIPELINE:
            edge_id = helpers.get_edge_id(edge_group, Pipeline.Speed.Edge.SPEEDCONTROL_0)
            edge_type = Pipeline.Speed.Edge.SPEEDCONTROL_0.type
            deleter = ScDeleter(get_edge_jnode_dirpath(edge_type, edge_id))
            return deleter

        else:
            raise NotImplemented

    def run(self):
        helpers.silent_rm_r(self._edge_sc_dirpath)
