import jsonschema

from edge.admin.reply_helpers import *
from edge.admin.ui_types import *

from edge_creator import *
import edge.pipeline_helpers as helpers
import edge.db as db
from ui_pipeline_set import get_pipeline_type


class UiEventStore:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def _check_postname_changed(old_post_name, new_post_name):
        return old_post_name != new_post_name

    @staticmethod
    def _check_channel_changed(old_ui_jnode, new_ui_jnode):
        return old_ui_jnode['channel'] != new_ui_jnode['channel']

    @staticmethod
    def _handle_get(handler_args):
        ui_jnode_filepath = handler_args
        ui_jnode = helpers.load_json_file(ui_jnode_filepath)
        post_guid, post_name = helpers.get_post_guid_name()
        ui_jnode['rw']['post-name'] = post_name
        ui_jnode['rw']['post-guid'] = post_guid

        return reply_done(ui_jnode['rw'])

    @staticmethod
    def _handle_put(handler_args):
        """
        :type handler_args: (srt, UiIdentity, dict)
        :rtype: dict
        """
        ui_pid, entry_point, new_ui_jnode = handler_args

        # load old-ui-jnode
        ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        raw_ui_jnode = helpers.load_json_file(ui_jnode_filepath)
        old_ui_jnode = raw_ui_jnode['rw']
        _, old_post_name = helpers.get_post_guid_name()

        if UiEventStore._check_channel_changed(old_ui_jnode, new_ui_jnode):
            eventlog.log_debug('fail: bad request (channel is read-only)')
            return reply_fail(REPLY_BAD_REQUEST_400)

        if not UiEventStore._check_postname_changed(old_post_name, new_ui_jnode['post-name']):
            return reply_done(new_ui_jnode)

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            helpers.set_post_name(new_ui_jnode['post-name'])

            # change db_post name
            db_config_jnode = helpers.load_db_json()
            db_args = db.DbArgs.deserialize(db_config_jnode['db'])
            post_args = db.DbPostArgs.deserialize(db_config_jnode['post'])

            def _handler(db_link):
                with db_link:
                    db.edit_post(db_link, post_args)

            try:
                db.with_db(db_args, _handler)
            except db.DbError:
                return False

            return True

        try:
            generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            return reply_done(new_ui_jnode)

        except db.DbError:
            eventlog.log_debug('fail: internal-server-error')
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

        except LockError:
            return reply_fail(REPLY_CONFLICT_409)

        except Exception as x:
            eventlog.log_debug("fail: %s" % x.message)
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_debug('fail: bad-request')
            return UiEventStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiEventStore(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)
        return UiEventStore(UiEventStore._handle_get, ui_jnode_filepath)

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiEventStore(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)

        if ui_ptype is None:
            return UiEventStore(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_debug('fail: bad-request')
            return UiEventStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load new-ui-jnode
        new_ui_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_debug('fail: bad-request')
            return UiEventStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load ui-schena-jnode
        try:
            ui_schema_filepath = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = json.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_debug("fail: load ui-schema-jnode (%s)" % x.message)
            return UiEventStore(handler=reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_debug("fail: validate new-ui-jnode (%s)" % x.message)
            return UiEventStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # all arguments are valid
        return UiEventStore(UiEventStore._handle_put, (ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiEventStore
        """
        if query_args.http_method == 'GET':
            return UiEventStore._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiEventStore._deserialize_put(query_args)

        else:
            eventlog.log_debug("fail: method-not-allowed")
            return UiEventStore(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
