import jsonschema

from edge.admin.reply_helpers import *
from edge.admin.ui_types import *
from edge.types import FormatTemplate

from edge_creator import *
import edge.pipeline_helpers as helpers
from ui_pipeline_set import get_pipeline_type


def _validate_format(jnode):
    """
    :type jnode: list
    :rtype: bool
    """
    key_to_true_map = {
        'speed': True,
        'datetime': True,
        'post-name': True,
        'geolocation': True,
        'plate': True}

    for row in jnode:
        for col in row['cols']:
            try:
                FormatTemplate(template=col['format']).substitute(key_to_true_map)
            except KeyError as x:
                eventlog.log_debug('fail: invalid key (%s)' % x.message)
                return False

    return True


class UiFactRules:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def _handle_get(handler_args):
        im_w, im_h, ui_jnode_filepath = handler_args
        ui_jnode = helpers.load_json_file(ui_jnode_filepath)
        ui_jnode['rw']['limits'] = {
            'images': {
                'scene.measurement': {
                    'max-width': im_w
                },
                'scene.measurement.alt':  {
                    'max-width': im_w
                },
                'vehicle.body':  {
                    'max-width': 800
                }
            }
        }
        return reply_done(ui_jnode['rw'])

    @staticmethod
    def _handle_put(handler_args):
        """
        :type handler_args: (srt, UiIdentity, dict)
        :rtype: dict
        """
        ui_pid, entry_point, new_ui_jnode = handler_args

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
                return False

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            return reply_done(new_ui_jnode)

        except LockError:
            return reply_fail(REPLY_CONFLICT_409)

        except Exception as x:
            eventlog.log_debug("fail: %s" % x.message)
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_debug('fail: bad-request')
            return UiFactRules(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiFactRules(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)

        image_wh = helpers.get_videosource_image_wh(ui_pid, ui_id)
        if image_wh is None:
            return UiFactRules(handler=reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        im_w, im_h = image_wh

        return UiFactRules(UiFactRules._handle_get, (im_w, im_h, ui_jnode_filepath))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiFactRules(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)

        if ui_ptype is None:
            return UiFactRules(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_debug('fail: bad-request')
            return UiFactRules(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load new-ui-jnode
        new_ui_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_debug('fail: bad-request')
            return UiFactRules(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load ui-schena-jnode
        try:
            ui_schema_filepath = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = json.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_debug("fail: load ui-schema-jnode (%s)" % x.message)
            return UiFactRules(handler=reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_debug("fail: validate new-ui-jnode (%s)" % x.message)
            return UiFactRules(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # validate "format" jnodes from json
        images_jnode = new_ui_jnode['args']['images']
        vehicle_body_footer_rows = images_jnode['vehicle.body']['footer']['rows']
        scene_alt_footer_rows = images_jnode['scene.measurement.alt']['footer']['rows']

        if _validate_format(vehicle_body_footer_rows) is False or _validate_format(scene_alt_footer_rows) is False:
            return UiFactRules(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # all arguments are valid
        return UiFactRules(UiFactRules._handle_put, (ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiFactRules
        """
        if query_args.http_method == 'GET':
            return UiFactRules._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiFactRules._deserialize_put(query_args)

        else:
            eventlog.log_debug("fail: method-not-allowed")
            return UiFactRules(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
