#!/usr/bin/python
import jsonschema

import edge.eventlog as eventlog
import gps
from edge.admin.reply_helpers import *
from edge.admin.ui_types import *
from edge.path_helpers import *

from edge_creator import *
import edge.pipeline_helpers as helpers
from ui_pipeline_set import get_pipeline_type


def _check_changed(old_jnode, new_jnode):
    if old_jnode['mode'] != new_jnode['mode']:
        return True

    if old_jnode['mode'] == 'manual':
        if old_jnode['latitude'] != new_jnode['latitude']:
            return True

        if old_jnode['longitude'] != new_jnode['longitude']:
            return True
    else:
        pass

    return False


def _handle_get(handler_args):
    """
    :type handler_args: (str, UiIdentity)
    :rtype: dict
    """
    ui_pid, entry_point = handler_args
    ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
    ui_jnode = helpers.load_json_file(ui_jnode_filepath)

    return reply_done(ui_jnode['rw'])


def _handle_put(handler_args):
    """
    :type handler_args: (str, UiIdentity, dict)
    :rtype:
    """
    ui_pid, entry_point, new_ui_jnode = handler_args

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: bool
        """
        if not _check_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            return False

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
        return True

    try:
        generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        return reply_done(new_ui_jnode)

    except LockError:
        return reply_fail(REPLY_CONFLICT_409)

    except Exception, x:
        eventlog.log_debug("fail: %s" % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _get_gps_geodata(arg_list):
    localhost, port = arg_list
    session = gps.gps(localhost, port)
    session.stream(gps.WATCH_ENABLE | gps.WATCH_NEWSTYLE)

    result_jnode = None
    start_time = int(time.time())
    while True:
        try:
            report = session.next()
            if report['class'] == 'DEVICES' and len(report['devices']) == 0:
                return reply_fail(REPLY_GATEWAY_TIMEOUT)

            if report['class'] == 'TPV':
                result_jnode = {"latitude": report['lat'], "longitude": report['lon']}
                break
            end_time = int(time.time())
            diff_time = end_time - start_time
            if diff_time > 2:  # 2 seconds
                return reply_fail(REPLY_GATEWAY_TIMEOUT)

        except KeyError:
            pass
        except StopIteration:
            break

    if not result_jnode:
        return reply_fail(REPLY_NOT_FOUND_404)
    return reply_done(result_jnode)


class UiGeoLocation:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def _validate_module_action(raw_action):
        """
        :type raw_action: str
        :rtype: (str | None, str | None)
        """
        return raw_action == 'define'

    @staticmethod
    def _deserialize_geolocation(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiGeoLocation
        """
        http_method = query_args.http_method
        url_path_info = query_args.path_info
        url_query_string = query_args.query_string
        body = json.loads(query_args.raw_body) if query_args.raw_body else None

        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiGeoLocation(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiGeoLocation(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        if http_method == 'GET':
            if url_path_info and not url_query_string:
                return UiGeoLocation(_handle_get, (ui_pid, entry_point))

        elif http_method == 'PUT':
            schema_abs_path = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            request_schema_jnode = helpers.load_json_file(schema_abs_path)
            if body:
                try:
                    jsonschema.validate(body, request_schema_jnode)
                except (ValueError, jsonschema.ValidationError):
                    return UiGeoLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

                return UiGeoLocation(_handle_put, (ui_pid, entry_point, body))

        else:
            return UiGeoLocation(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

        return UiGeoLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

    @staticmethod
    def _deserialize_geolocation_action(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiGeoLocation
        """
        http_method = query_args.http_method
        url_path_info = query_args.path_info
        url_query_string = query_args.query_string

        try:
            action = query_args.get_path_info_item(UI_PIPELINE_ACTION_INDEX)
        except IndexError:
            return UiGeoLocation(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        success = UiGeoLocation._validate_module_action(action)
        if not success:
            return UiGeoLocation(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        if http_method == 'GET':
            if url_path_info and action and not url_query_string:
                return UiGeoLocation(_get_gps_geodata, ["localhost", "2947"])

        else:
            return UiGeoLocation(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

        return UiGeoLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiGeoLocation
        """
        #
        url_path_info = query_args.path_info

        if len(url_path_info) == 3:
            return UiGeoLocation._deserialize_geolocation(query_args)

        elif len(url_path_info) == 4:
            return UiGeoLocation._deserialize_geolocation_action(query_args)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
