import jsonschema

from edge.admin.reply_helpers import *
from edge.admin.ui_types import *

from edge_creator import *
import edge.pipeline_helpers as helpers
from ui_pipeline_set import get_pipeline_type


def _handle_get(handler_args):
    """
    :type handler_args: (str, UiIdentity)
    :rtype: dict
    """
    ui_pid, entry_point = handler_args

    videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
    if videsource_status is None:
        eventlog.log_debug('fail: get videosource status')
        return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                           error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)
    elif videsource_status != 0:
        eventlog.log_debug('fail: videosource status error')
        return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                           error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

    ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
    ui_jnode = helpers.load_json_file(ui_jnode_filepath)

    return reply_done(ui_jnode['rw'])


def _handle_put(handler_args):
    """
    :type handler_args: (str, UiIdentity, dict)
    :rtype: dict
    """
    ui_pid, entry_point, new_ui_jnode = handler_args

    # check status of videosource
    videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
    if videsource_status is None:
        eventlog.log_debug('fail: get videosource status')
        return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                           error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

    elif videsource_status != 0:
        eventlog.log_debug('fail: videosource status error')
        return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                           error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

    restart_setupstore = []

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: bool
        """
        if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
            return False

        if ui_identity_to_jnode_map[entry_point]['rw']['enabled'] != new_ui_jnode['enabled']:
            restart_setupstore.append(True)

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
        return True

    try:
        generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        if restart_setupstore:
            helpers.sighup_setupstore()
        return reply_done(new_ui_jnode)

    except LockError:
        return reply_fail(REPLY_CONFLICT_409)


class UiHttpStreaming:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """

        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiHttpStreaming
        """
        #
        http_method = query_args.http_method
        url_query_string = query_args.query_string
        new_ui_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None

        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiHttpStreaming(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiHttpStreaming(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            return UiHttpStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        #
        if http_method == 'GET':
            if not new_ui_jnode:
                return UiHttpStreaming(_handle_get, (ui_pid, entry_point))

        elif http_method == 'PUT':
            schema_jnode_filepath = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            request_schema_jnode = helpers.load_json_file(schema_jnode_filepath)
            if new_ui_jnode:
                try:
                    jsonschema.validate(new_ui_jnode, request_schema_jnode)
                    return UiHttpStreaming(_handle_put, (ui_pid, entry_point, new_ui_jnode))

                except (ValueError, jsonschema.ValidationError):
                    return UiHttpStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)
                except LockError:
                    return UiHttpStreaming(handler=reply_fail, handler_args=REPLY_CONFLICT_409)

        else:
            eventlog.log_debug('fail: unsupported http-method')
            return UiHttpStreaming(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

        eventlog.log_debug('fail: deserialize query')
        return UiHttpStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
