import jsonschema
from shapely.geometry import polygon as polygon

import edge.eventlog as eventlog
from edge.admin.reply_helpers import *
from edge.admin.ui_types import *
from edge.path_helpers import *

from edge_creator import *
import edge.pipeline_helpers as helpers
from ui_pipeline_set import get_pipeline_type


class UiLaneEnforcement:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """

        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def _handle_get(handler_args):
        """
        :type handler_args: (str, UiIdentity)
        :rtype: dict
        """
        ui_pid, entry_point = handler_args

        videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
        if videsource_status is None:
            eventlog.log_debug('fail: get videosource status')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)
        elif videsource_status != 0:
            eventlog.log_debug('fail: videosource status error')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

        ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        old_jnode = helpers.load_json_file(ui_jnode_filepath)

        lanes_jnode = old_jnode['rw']['road']['lanes']
        lanes_jnode.sort(key=lambda lane: lane['id'])
        old_jnode['rw']['road']['lanes'] = lanes_jnode

        return reply_done(old_jnode['rw'])

    @staticmethod
    def _handle_put(handler_args):
        """
        :type handler_args: (str, UiIdentity, dict)
        :rtype: dict
        """
        ui_pid, entry_point, new_ui_jnode = handler_args

        videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
        if videsource_status is None:
            eventlog.log_debug('fail: get videosource status')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)
        elif videsource_status != 0:
            eventlog.log_debug('fail: videosource status error')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
                return False

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            generate(ui_pid, entry_point, _edit_ui_jnode_handler)

            return reply_done(new_ui_jnode)

        except LockError:
            return reply_fail(REPLY_CONFLICT_409)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLaneEnforcement
        """
        #
        url_query_string = query_args.query_string
        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        return UiLaneEnforcement(UiLaneEnforcement._handle_get, (ui_pid, entry_point))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLaneEnforcement
        """
        url_query_string = query_args.query_string
        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load new-ui-jnode
        new_ui_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_debug('fail: bad-request')
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load ui-schena-jnode
        try:
            ui_schema_filepath = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = json.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_debug("fail: load ui-schema-jnode (%s)" % x.message)
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_debug("fail: validate new-ui-jnode (%s)" % x.message)
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        try:
            # validate self-intersection of lanes
            lanes_jnode = new_ui_jnode['road']['lanes']
            for lane_jnode in lanes_jnode:
                polymask = polygon.Polygon(lane_jnode["mask"])
                if not polymask.is_valid:
                    eventlog.log_debug("fail: bad-request (invalid road-mask)")
                    return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)
        except(ValueError, IndexError):
            eventlog.log_debug("fail: bad-request (road-mask contains less then 3 vertexes)")
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # all arguments are valid
        return UiLaneEnforcement(UiLaneEnforcement._handle_put, (ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLaneEnforcement
        """
        if query_args.http_method == 'GET':
            return UiLaneEnforcement._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiLaneEnforcement._deserialize_put(query_args)

        else:
            eventlog.log_debug("fail: method-not-allowed")
            return UiLaneEnforcement(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
