import jsonschema
import subprocess

import edge.eventlog as eventlog
from edge.admin.reply_helpers import *
from edge.admin.ui_types import *
from edge.path_helpers import *

from edge_creator import *
import edge.pipeline_helpers as helpers


class UiLocation:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLocation
        """
        if query_args.get_ui_action():
            return UiLocation._deserialize_location_define_rotation(query_args)

        else:
            return UiLocation._deserialize_location(query_args)

    @staticmethod
    def _deserialize_location(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLocation
        """
        if query_args.http_method == 'GET':
            return UiLocation._deserialize_get_location(query_args)

        elif query_args.http_method == 'PUT':
            return UiLocation._deserialize_put_location(query_args)

        else:
            eventlog.log_debug("fail: method-not-allowed")
            return UiLocation(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

    @staticmethod
    def _deserialize_get_location(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLocation
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_debug('fail: bad-request')
            return UiLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        ui_pid = query_args.get_ui_pid()
        ui_type = query_args.get_ui_type()
        ui_id = query_args.get_ui_id()

        ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)
        return UiLocation(UiLocation._handle_get_location, ui_jnode_filepath)

    @staticmethod
    def _deserialize_put_location(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLocation
        """
        if query_args.query_string:
            eventlog.log_debug('fail: bad-request')
            return UiLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load new-ui-jnode
        new_ui_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_debug('fail: bad-request')
            return UiLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        ui_pid = query_args.get_ui_pid()
        ui_type = query_args.get_ui_type()
        ui_id = query_args.get_ui_id()

        # load ui-schena-jnode
        try:
            ui_schema_filepath = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = json.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_debug("fail: load ui-schema-jnode (%s)" % x.message)
            return UiLocation(handler=reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_debug("fail: validate new-ui-jnode (%s)" % x.message)
            return UiLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # all arguments are valid
        return UiLocation(UiLocation._handle_put_location, (ui_pid, ui_type, ui_id, new_ui_jnode))

    @staticmethod
    def _deserialize_location_define_rotation(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        """
        if query_args.get_ui_action() != 'define-rotation':
            eventlog.log_debug('fail: not-found')
            return UiLocation(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        if query_args.raw_body or not query_args.query_string:
            eventlog.log_debug('fail: bad-request')
            return UiLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        if query_args.http_method != 'GET':
            eventlog.log_debug("fail: method-not-allowed")
            return UiLocation(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

        ui_pid = query_args.get_ui_pid()
        ui_id = query_args.get_ui_id()
        videosource_jnode_filepath = helpers.get_ui_pipeline_jnode_filepath(ui_pid, UI_TYPE_VIDEOSOURCE, ui_id)
        videosource_jnode = helpers.load_json_file2(videosource_jnode_filepath)
        fov_h, fov_v = videosource_jnode['ro']['fov-hv']

        rlandmark_command_line = UiLocation._get_rlandmark_command_line(query_args.query_string, fov_h, fov_v)
        if not rlandmark_command_line:
            eventlog.log_debug('fail: bad-request')
            return UiLocation(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        return UiLocation(UiLocation._handle_get_location_define_rotation, rlandmark_command_line)

    @staticmethod
    def _handle_get_location(ui_jnode_filepath):
        """
        :type ui_jnode_filepath: str
        """
        ui_jnode = helpers.load_json_file(ui_jnode_filepath)
        return reply_done(ui_jnode['rw'])

    @staticmethod
    def _check_ui_jnode_changed(old_ui_jnode, new_ui_jnode):
        return old_ui_jnode != new_ui_jnode

    @staticmethod
    def _handle_put_location(handler_args):
        """
        :type handler_args: (str, str, str, dict)
        """
        ui_pid, ui_type, ui_id, new_ui_jnode = handler_args

        entry_point = UiIdentity(PTYPE_SPEED, ui_type, ui_id)

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
                return False

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode

            custom_matrix = ui_identity_to_jnode_map[entry_point]['ro'].get('rhcc-to-ndc')

            if custom_matrix is None:
                videosource_ui_jnode = ui_identity_to_jnode_map[UiIdentity(PTYPE_SPEED, UI_TYPE_VIDEOSOURCE, ui_id)]
                fov_h, fov_v = videosource_ui_jnode['ro']['fov-hv']

                landmark_command_line = UiLocation._get_landmark_command_line(new_ui_jnode, fov_h, fov_v)
                rlandmark_out = UiLocation._handle_get_location_define_rotation(landmark_command_line)
                ui_identity_to_jnode_map[entry_point]['rt'] = {'rhcc-to-ndc': rlandmark_out['result']['rhcc-to-ndc']}

            return True

        try:
            generate(ui_pid, entry_point, _edit_ui_jnode_handler)

            return reply_done(new_ui_jnode)

        except LockError:
            return reply_fail(REPLY_CONFLICT_409)

    @staticmethod
    def _get_landmark_command_line(new_ui_jnode, fov_h, fov_v):
        # deserialize sensor-position
        unit_position_jnode = new_ui_jnode['unit-position']
        forward = unit_position_jnode['forward']
        left = unit_position_jnode['left']
        up = unit_position_jnode['up']

        # deserialize sensor-rotation
        unit_rotation_manual_jnode = new_ui_jnode['unit-rotation-manual']
        raw_azimuth = unit_rotation_manual_jnode['azimuth']
        raw_elevation = unit_rotation_manual_jnode['elevation']
        roll = unit_rotation_manual_jnode.get('roll', 0)

        landmark_args = {
            '--fov-vertical': fov_v,
            '--fov-horizontal': fov_h,
            # '--matrix-path': '/etc/edge/radar.0/mapping.conf',

            '--position-forward': forward,
            '--position-left': left,
            '--position-top': up,

            '--azimuth': raw_azimuth,
            '--elevation': raw_elevation,
            '--roll': roll
        }

        command_line = '/opt/edge/bin/landmark_v2 --mode-landmark'
        for valid_arg_name, valid_arg_value in landmark_args.iteritems():
            command_line += ' %s=%s' % (valid_arg_name, valid_arg_value)

        return command_line

    @staticmethod
    def _get_rlandmark_command_line(query_string, fov_h, fov_v):
        """
        :type query_string: str
        :rtype: str | None
        """
        if not (query_string and query_string.startswith('?')):
            eventlog.log_debug('fail: invalid arg')
            return None

        landmark_args = {
            '--fov-vertical': fov_v,
            '--fov-horizontal': fov_h,

            '--position-forward': None,
            '--position-left': None,
            '--position-top': None,

            '--rotation-road-near-edge-left': None,
            '--rotation-road-far-edge-left': None,

            '--rotation-landmark-image': None,
            '--rotation-landmark-forward': None,
            '--rotation-landmark-left': None
        }

        for param_value in query_string[1:].split('&'):
            param, value = param_value.split('=', 1)

            if param == 'position-forward':
                landmark_args['--position-forward'] = float(value)

            elif param == 'position-left':
                landmark_args['--position-left'] = float(value)

            elif param == 'position-top':
                landmark_args['--position-top'] = float(value)

            elif param == 'rotation-road-near-edge-left':
                landmark_args['--rotation-road-near-edge-left'] = float(value)

            elif param == 'rotation-road-far-edge-left':
                landmark_args['--rotation-road-far-edge-left'] = float(value)

            elif param == 'rotation-landmark-image':
                if value in ['bottom-left', 'bottom-right']:
                    landmark_args['--rotation-landmark-image'] = value

            elif param == 'rotation-landmark-forward':
                landmark_args['--rotation-landmark-forward'] = float(value)

            elif param == 'rotation-landmark-left':
                landmark_args['--rotation-landmark-left'] = float(value)

        invalid_args = filter(lambda (arg_name, arg_value): arg_value is None, landmark_args.iteritems())
        if invalid_args:
            for invalid_arg_name, _ in invalid_args:
                eventlog.log_debug('fail: param "%s" is absent' % invalid_arg_name)

            return None

        command_line = '/opt/edge/bin/landmark_v2 --mode-rlandmark'
        for valid_arg_name, valid_arg_value in landmark_args.iteritems():
            command_line += ' %s=%s' % (valid_arg_name, valid_arg_value)

        return command_line

    @staticmethod
    def _handle_get_location_define_rotation(landmark_command_line):
        """
        :type landmark_command_line: str
        """
        eventlog.log_debug('landmark-cmd: %s' % landmark_command_line)

        landmark = subprocess.Popen([landmark_command_line], stdout=subprocess.PIPE, shell=True)
        rlandmar_stdout = landmark.stdout.read()
        eventlog.log_debug('landmark-out: ' + rlandmar_stdout)

        rlandmark_jnode = json.loads(rlandmar_stdout)
        return reply_done(rlandmark_jnode)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
