import jsonschema

import edge.eventlog as eventlog
from edge.admin.reply_helpers import *
from edge.admin.ui_types import *
from edge.path_helpers import *

from edge_creator import *
import edge.pipeline_helpers as helpers
from ui_pipeline_set import get_pipeline_type


def _check_changed(old_jnode, new_jnode):
    return old_jnode != new_jnode


class UiMediaStore:
    _TYPE_SHORT_LONG_TERM = 1
    _TYPE_SHORT_TERM = 2

    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def _handle_get(handler_args):
        """
        :type handler_args: (str, UiIdentity)
        :rtype: dict
        """
        ui_pid, entry_point = handler_args

        videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
        if videsource_status is None:
            eventlog.log_debug('fail: get videosource status')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)
        elif videsource_status != 0:
            eventlog.log_debug('fail: videosource status error')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

        old_jnode_file = get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        old_jnode = helpers.load_json_file(old_jnode_file)

        return reply_done(old_jnode['rw'])

    @staticmethod
    def _handle_put(handler_args):
        """
        :type handler_args: (str, UiIdentity, dict)
        :rtype: dict
        """
        ui_pid, entry_point, new_ui_jnode = handler_args

        videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
        if videsource_status is None:
            eventlog.log_debug('fail: get videosource status')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)
        elif videsource_status != 0:
            eventlog.log_debug('fail: videosource status error')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if not _check_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
                return False

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            generate(ui_pid, entry_point, _edit_ui_jnode_handler)

            return reply_done(new_ui_jnode)

        except LockError:
            return reply_fail(REPLY_CONFLICT_409)

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiMediaStore
        """
        new_ui_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None
        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        schema_jnode_abs_path = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
        ui_schema_jnode = helpers.load_json_file(schema_jnode_abs_path)
        if not ui_schema_jnode:
            eventlog.log_debug("fail: load ui-schema-jnode")
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        if not new_ui_jnode:
            eventlog.log_debug("fail: load ui-jnode")
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        old_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)
        old_jnode = helpers.load_json_file(old_jnode_filepath)
        old_jnode_type = old_jnode['ro']['type']

        if old_jnode_type == UiMediaStore._TYPE_SHORT_LONG_TERM:
            try:
                quality = new_ui_jnode['short-term']['videostream']['codec-mjpeg-quality']
                long_term_enabled = new_ui_jnode['long-term']['enabled']
                if long_term_enabled:
                    new_ui_jnode['long-term']['videostream']['codec'] = 'mjpeg'
                    new_ui_jnode['long-term']['videostream']['codec-mjpeg-quality'] = quality

            except KeyError as x:
                eventlog.log_debug("fail: bad-request (%s)" % x.message)
                return UiMediaStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

            long_term_jnode = new_ui_jnode.get('long-term')
            if not long_term_jnode:
                eventlog.log_debug("fail: bad-request (#/long-term is absent)")
                return UiMediaStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)
            return UiMediaStore(UiMediaStore._handle_put, (ui_pid, entry_point, new_ui_jnode))

        except (ValueError, jsonschema.ValidationError):
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiMediaStore
        """
        #
        url_query_string = query_args.query_string

        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        #
        if query_args.http_method == 'GET':
            if not query_args.raw_body:
                return UiMediaStore(UiMediaStore._handle_get, (ui_pid, entry_point))

        elif query_args.http_method == 'PUT':
            return UiMediaStore._deserialize_put(query_args)

        else:
            eventlog.log_debug('fail: method-not-allowed')
            return UiMediaStore(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

        eventlog.log_debug('fail: bad-request')
        return UiMediaStore(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
