import commands
import jsonschema
import glob

import edge.eventlog as eventlog
from edge.admin.reply_helpers import *
from edge.admin.ui_types import *

from edge_creator import *
import edge.pipeline_helpers as helpers
from ui_pipeline_set import get_pipeline_type

UI_NETWORK_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/network.json'
MANAGEMENT_MAC_FILEPATH = '/etc/network/management-mac'


class UiRtspStreaming:
    TYPE = 'rtsp-streaming'

    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def _make_rtsp_url(ui_rw_jnode):
        """
        :type ui_rw_jnode: dict
        :rtype: str
        """

        try:
            with open(UI_NETWORK_JNODE_FILEPATH) as json_file:
                ui_network_jnode = json.load(json_file)
            mac_address = ui_network_jnode['ro']['mac']

        except IOError as x:
            if x.errno == 2:
                if os.path.exists(MANAGEMENT_MAC_FILEPATH):
                    with open(MANAGEMENT_MAC_FILEPATH, 'r') as mgmt_mac_file:
                        mac_address = mgmt_mac_file.readlines()[0].strip()
                else:
                    raise

            else:
                eventlog.log_debug('fail: read network settings')
                raise IOError

        if_list = glob.glob('/sys/class/net/*/address')
        mac_to_name_map = {}

        for iface_filepath in if_list:
            with open(iface_filepath, 'r') as iface_file:
                mac_to_name_map[iface_file.readlines()[0].strip()] = iface_filepath.split('/')[-2]

        ifname = mac_to_name_map[mac_address]
        try:
            ip_addr = commands.getoutput("ip addr show %s | grep -v '127.0.0.1' | grep -w inet | awk '{print $2}'"
                                         % ifname).split('/')[0]
        except:
            eventlog.log_debug('fail: discover ip-address')
            raise RuntimeError

        tcp_port = str(ui_rw_jnode['streaming']['tcp-port'])
        ui_rtsp_codec = ui_rw_jnode['live']['videostream']['format']
        if ui_rtsp_codec == 'mjpeg':
            rtsp_codec = 'mjpeg'
        elif ui_rtsp_codec == 'h264-cqp' or 'h264-cbr':
            rtsp_codec = 'h264'
        else:
            eventlog.log_debug('unknown rtsp-codec %s' % ui_rtsp_codec)
            raise ValueError

        rtsp_url = 'rtsp://' + ip_addr + ':' + tcp_port + '/live-' + rtsp_codec

        return rtsp_url

    @staticmethod
    def _handle_get(handler_args):
        """
        :type: handler_args: (str, str, str)
        :rtype: dict
        """
        ui_jnode_filepath, ui_pid, ui_id = handler_args

        videsource_status = helpers.get_videsource_status(ui_pid, ui_id)
        if videsource_status is None:
            eventlog.log_debug('fail: get videosource status')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)
        elif videsource_status != 0:
            eventlog.log_debug('fail: videosource status error')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

        ui_jnode = helpers.load_json_file(ui_jnode_filepath)
        if not ui_jnode:
            eventlog.log_debug('fail: load ui-jnode')
            return reply_fail(REPLY_NOT_FOUND_404)

        try:
            rtsp_url = UiRtspStreaming._make_rtsp_url(ui_jnode['rw'])
        except (ValueError, RuntimeError, IOError):
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

        ui_jnode['rw']['streaming']['url'] = rtsp_url

        return reply_done(ui_jnode['rw'])

    @staticmethod
    def _handle_put(handler_args):
        """
        :type handler_args: (str, UiIdentity, dict)
        :rtype: dict
        """
        ui_pid, entry_point, new_ui_jnode = handler_args

        videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
        if videsource_status is None:
            eventlog.log_debug('fail: get videosource status')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)
        elif videsource_status != 0:
            eventlog.log_debug('fail: videosource status error')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code=FAIL_CODE_VIDEOSOURCE_ERROR,
                               error_text=FAIL_TEXT_VIDEOSOURCE_ERROR)

        # load old-ui-jnode
        ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
        try:
            with open(ui_jnode_filepath) as json_file:
                raw_ui_jnode = json.load(json_file)
                old_ui_jnode = raw_ui_jnode['rw']

        except (IOError, ValueError) as x:
            eventlog.log_debug("fail: load old-ui-jnode (%s)" % x.message)
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

        old_rtsp_url = UiRtspStreaming._make_rtsp_url(old_ui_jnode)
        new_rtsp_url = new_ui_jnode['streaming']['url']

        if old_rtsp_url != new_rtsp_url:
            eventlog.log_debug('fail: not allowed to edit rtsp-url')
            return reply_fail(REPLY_BAD_REQUEST_400)

        if old_ui_jnode == new_ui_jnode:
            return reply_done(new_ui_jnode)

        restart_setupstore = []

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            if ui_identity_to_jnode_map[entry_point]['rw']['enabled'] != new_ui_jnode['enabled']:
                restart_setupstore.append(True)

            ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
            return True

        try:
            generate(ui_pid, entry_point, _edit_ui_jnode_handler)
            if restart_setupstore:
                helpers.sighup_setupstore()
            return reply_done(new_ui_jnode)

        except LockError:
            return reply_fail(REPLY_CONFLICT_409)

    @staticmethod
    def _deserialize_get(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        if query_args.query_string or query_args.raw_body:
            eventlog.log_debug('fail: bad-request')
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)
        return UiRtspStreaming(UiRtspStreaming._handle_get, (ui_jnode_filepath, ui_pid, ui_id))

    @staticmethod
    def _deserialize_put(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        if query_args.query_string:
            eventlog.log_debug('fail: unsupported query-string')
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load new-ui-jnode
        new_ui_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_debug('fail: no new-ui-jnode')
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # load ui-schema-jnode
        try:
            ui_schema_filepath = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = json.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_debug("fail: load ui-schema-jnode (%s)" % x.message)
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_INTERNAL_SERVER_ERROR_500)

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_debug("fail: validate new-ui-jnode (%s)" % x.message)
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        ui_streaming_jnode = new_ui_jnode['streaming']
        if ui_streaming_jnode['transport'] == 'rtp-over-udp-or-tcp':
            if ui_streaming_jnode['min-udp-port'] == 0 and ui_streaming_jnode['max-udp-port'] != 0:
                return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

            if ui_streaming_jnode['max-udp-port'] == 0 and ui_streaming_jnode['min-udp-port'] != 0:
                return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        # all arguments are valid
        return UiRtspStreaming(UiRtspStreaming._handle_put, (ui_pid, entry_point, new_ui_jnode))

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiRtspStreaming
        """
        if query_args.http_method == 'GET':
            return UiRtspStreaming._deserialize_get(query_args)

        elif query_args.http_method == 'PUT':
            return UiRtspStreaming._deserialize_put(query_args)

        else:
            eventlog.log_debug("fail: method-not-allowed")
            return UiRtspStreaming(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
