import jsonschema

import edge.eventlog as eventlog
from edge.admin.reply_helpers import *
from edge.admin.ui_types import *
from edge.path_helpers import *

from edge_creator import *
import edge.pipeline_helpers as helpers
from ui_pipeline_set import get_pipeline_type
from copy import deepcopy


def _check_mv_changed(old_ui_jnode, new_ui_jnode):
    return old_ui_jnode['videostream'] != new_ui_jnode['videostream']


def _check_mv_resolution_changed(old_ui_jnode, new_ui_jnode):
    return old_ui_jnode['videostream']['resolution'] != new_ui_jnode['videostream']['resolution']


def _check_gc_mode_changed(old_ui_jnode, new_ui_jnode):
    old_gc_mode = old_ui_jnode['videostream']['control']['profile'] == 'auto-70-150'

    new_gc_mode = new_ui_jnode['videostream']['control']['profile'] == 'auto-70-150'

    return old_gc_mode != new_gc_mode


def _check_rtsp_link_changed(old_ui_jnode, new_ui_jnode):
    return old_ui_jnode['streaming']['url'] != new_ui_jnode['streaming']['url']


def _check_rtsp_changed(old_ui_jnode, new_ui_jnode):
    if old_ui_jnode['streaming'] != new_ui_jnode['streaming']:
        return True

    if old_ui_jnode['videostream']['hw-decode'] != new_ui_jnode['videostream']['hw-decode']:
        return True

    if old_ui_jnode['videostream']['codec'] != new_ui_jnode['videostream']['codec']:
        return True

    if old_ui_jnode['videostream']['color-mode'] != new_ui_jnode['videostream']['color-mode']:
        return True

    return False


def _check_fps_changed(old_ui_jnode, new_ui_jnode):
    return old_ui_jnode['videostream']['fps'] != new_ui_jnode['videostream']['fps']


def _handle_get(handler_args):
    """
    :type handler_args: (str, UiIdentity)
    """
    ui_pid, entry_point = handler_args
    ui_jnode_filepath = get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)

    eventlog.log_debug("load: %s" % ui_jnode_filepath)
    old_jnode = helpers.load_json_file(ui_jnode_filepath)

    if old_jnode['rw']['type'] == 'rtsp':
        image_wh = helpers.get_videosource_image_wh(ui_pid, entry_point.id, old_jnode)
        if image_wh:
            old_jnode['rw']['videostream']['resolution'] = {
                "w": image_wh[0],
                "h": image_wh[1]
            }

    return reply_done(old_jnode['rw'])


def _handle_put(handler_args):
    """
    :type handler_args: (str, UiIdentity, dict)
    """
    ui_pid, entry_point, new_ui_jnode = handler_args

    if entry_point.ptype == PTYPE_SPEED or entry_point.ptype == PTYPE_LPR_MV:
        reply_jnode = _put_mv(ui_pid, entry_point, new_ui_jnode)

    elif entry_point.ptype == PTYPE_LPR_RTSP:
        reply_jnode = _put_rtsp(ui_pid, entry_point, new_ui_jnode)

    else:
        eventlog.log_debug("fail: unsupported pipeline type")
        reply_jnode = reply_fail(REPLY_BAD_REQUEST_400)

    return reply_jnode


def _put_rtsp(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: UiIdentity
    :type new_ui_jnode: dict
    :rtype: dict
    """
    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :return: dict
        """
        if not _check_rtsp_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            return False

        if _check_rtsp_link_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            for ui_identity in ui_identity_to_jnode_map:
                if ui_identity.type == UI_TYPE_LPR:
                    ui_identity_to_jnode_map[ui_identity]['rw']['plate-recognizer']['image-azone'] = \
                        {'include-mask': [], 'exclude-mask': [], 'resolution': {'w': 65535, 'h': 65355}}

                if ui_identity.type == UI_TYPE_LANE_ENFORCEMENT:
                    ui_identity_to_jnode_map[ui_identity]['rw']['road']['lanes'] = []

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode

        return True

    try:
        generate(ui_pid, entry_point, _edit_ui_jnode_handler)
        return reply_done(new_ui_jnode)

    except LockError:
        return reply_fail(REPLY_CONFLICT_409)


def _put_mv(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: UiIdentity
    :type new_ui_jnode: dict
    :rtype: dict
    """
    restart_setupstore = []
    restart_mediastore = []
    if new_ui_jnode['videostream']['control']['profile'] in ['auto-127', 'auto-70-150']:
        min_gain = new_ui_jnode['videostream']['control']['gain-auto-min']
        max_gain = new_ui_jnode['videostream']['control']['gain-auto-max']

        min_exposure = new_ui_jnode['videostream']['control']['exposure-auto-min']
        max_exposure = new_ui_jnode['videostream']['control']['exposure-auto-max']

        # GrabControl works bad when min_value == max_value
        if min_gain >= max_gain or min_exposure >= max_exposure:
            eventlog.log_debug('fail: min value bigger than max (gain or exposure)')
            return reply_fail(REPLY_BAD_REQUEST_400)

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: bool
        """
        if not _check_mv_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            return False

        if _check_fps_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            restart_mediastore.append(True)

        # if vs changed :
        if _check_mv_resolution_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            for ui_identity in ui_identity_to_jnode_map:
                if ui_identity.type == UI_TYPE_LPR:
                    ui_identity_to_jnode_map[ui_identity]['rw']['plate-recognizer']['image-azone'] = \
                        {'include-mask': [], 'exclude-mask': [], 'resolution': {'w': 65535, 'h': 65355}}
                if ui_identity.type == UI_TYPE_LANE_ENFORCEMENT:
                    ui_identity_to_jnode_map[ui_identity]['rw']['road']['lanes'] = []

        if _check_gc_mode_changed(ui_identity_to_jnode_map[entry_point]['rw'], new_ui_jnode):
            restart_setupstore.append(True)

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
        return True

    try:
        generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        if restart_setupstore:
            helpers.sighup_setupstore()

        if restart_mediastore:
            edge_group = helpers.get_edge_group(ui_pid)
            edge_identity = EdgeIdentity(EDGE_TYPE_MEDIASTORE, '0')
            edge_id = helpers.get_edge_id(edge_group, edge_identity)
            helpers.sighup_edge_module(EDGE_TYPE_MEDIASTORE, edge_id)

        return reply_done(new_ui_jnode)

    except LockError:
        return reply_fail(REPLY_CONFLICT_409)


class UiVideosource:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """

        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiVideosource
        """
        #
        http_method = query_args.http_method
        url_query_string = query_args.query_string
        body_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None

        try:
            ui_pid = query_args.get_path_info_item(UI_PIPELINE_PID_INDEX)
            ui_type = query_args.get_path_info_item(UI_PIPELINE_TYPE_INDEX)
            ui_id = query_args.get_path_info_item(UI_PIPELINE_ID_INDEX)
        except IndexError:
            return UiVideosource(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiVideosource(handler=reply_fail, handler_args=REPLY_NOT_FOUND_404)

        entry_point = UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            eventlog.log_debug('fail: wrong query-string')
            return UiVideosource(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        #
        if http_method == 'GET':
            if not body_jnode:
                return UiVideosource(_handle_get, (ui_pid, entry_point))

        elif http_method == 'PUT':
            schema_abs_path = get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            request_schema_jnode = helpers.load_json_file(schema_abs_path)
            if body_jnode:
                try:
                    jsonschema.validate(body_jnode, request_schema_jnode)
                    # ROI checks - (width_offset + ROI_width) < max_width
                    if entry_point.ptype == PTYPE_SPEED or entry_point.ptype == PTYPE_LPR_MV:
                        roi_check = deepcopy(body_jnode)
                        roi_check['videostream']['resolution']['w'] += roi_check['videostream']['resolution']['x']
                        roi_check['videostream']['resolution']['h'] += roi_check['videostream']['resolution']['y']
                        jsonschema.validate(roi_check, request_schema_jnode)

                    return UiVideosource(_handle_put, (ui_pid, entry_point, body_jnode))

                except (ValueError, jsonschema.ValidationError) as x:
                    eventlog.log_debug('fail: %s' % x.message)
                    return UiVideosource(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

        else:
            eventlog.log_debug('fail: method not allowed')
            return UiVideosource(handler=reply_fail, handler_args=REPLY_METHOD_NOT_ALLOWED_405)

        eventlog.log_debug('fail: deserialize')
        return UiVideosource(handler=reply_fail, handler_args=REPLY_BAD_REQUEST_400)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
