#!/usr/bin/python
import datetime
import json
import os
import shutil
import subprocess
import sys
import time

import edge.admin.ui_types as ui_types
import edge.eventlog as eventlog
import jsonschema
from edge.admin.reply_helpers import *
from edge.fs_helpers import create_schema_symlink, create_jnode_from_default, try_lock_pid_file
from edge.pipeline_helpers import load_db_json
import edge.db as db


UI_TIME_LOCK_FILEPATH = '/var/run/edge/time'
UI_TIME_SCHEMA_FILEPATH = '/etc/edge/setupstore/schemas/time.json'
UI_TIME_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/time.json'
UI_TIME_DEFJNODE_FILEPATH = '/opt/edge/share/setupstore/ui/time-jnode.json'

UI_TIME_TZ_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/time-tz.json'

UI_PIPELINE_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/pipeline.json'

NTP_CONF_PATH = '/etc/ntp.conf'
NTP_DATE_CONF_PATH = '/etc/default/ntpdate'

ACTION_NONE = 0
ACTION_RESTART_OR = 1

BACKUP_DIRPATH = '/etc/edge/setupstore/backup/time.' + \
                 str(round(time.time() * 1000000)) + '/'


def try_create_backup():
    try:
        os.makedirs(BACKUP_DIRPATH)
        shutil.copy(NTP_CONF_PATH, BACKUP_DIRPATH)
        shutil.copy(UI_TIME_JNODE_FILEPATH, BACKUP_DIRPATH)
        return True

    except Exception:
        pass

        return False


def try_restore_backup():
    try:
        shutil.copy(BACKUP_DIRPATH + 'ntp.conf', NTP_CONF_PATH)
        shutil.copy(BACKUP_DIRPATH + 'time.json', UI_TIME_JNODE_FILEPATH)
        os.rmdir(BACKUP_DIRPATH)
        return True

    except Exception:
        return False


def remove_backup():
    try:
        shutil.rmtree(BACKUP_DIRPATH)
    except Exception:
        pass


def _check_ntp_changed(old_jnode, new_jnode):
    if old_jnode['mode'] != new_jnode['mode']:
        return True

    if new_jnode['mode'] == 'ntp':
        if old_jnode['servers'] != new_jnode['servers']:
            return True

    if new_jnode['mode'] == 'manual':
        if old_jnode['ts'] != new_jnode['ts']:
            return True

    return False


def _edit_ntp_conf(raw_old_jnode, new_jnode):

    new_ntp_lines = []

    new_mode = new_jnode['mode']
    if new_mode == 'ntp':
        if len(new_ntp_lines) > 0 and not new_ntp_lines[-1].endswith('\n'):
            new_ntp_lines.append('\n')

        for server_jnode in new_jnode['servers']:
            address = server_jnode['address']
            if server_jnode['prefered']:
                new_ntp_lines.append('server %s prefer\n' % address)
            else:
                new_ntp_lines.append('server %s\n' % address)

        for server_jnode in new_jnode['servers']:
            new_ntp_lines.append('restrict %s noquery notrap\n' % server_jnode['address'])

        # append parameters from 'ro'-jnode as is
        ro_lines = '\n'.join(raw_old_jnode['ro'].get('ntp-conf-footer', []))
        new_ntp_lines.append(ro_lines)

    elif new_mode == 'manual':
        pass

    else:
        raise NotImplemented

    with open(NTP_CONF_PATH, 'w') as new_ntp_file:
        new_ntp_file.write(''.join(new_ntp_lines))


def _edit_ntp_date_conf(new_mode):
    with open(NTP_DATE_CONF_PATH, 'r') as old_hosts_file:
        old_hosts_lines = old_hosts_file.readlines()

    new_host_lines = filter(lambda old_line: not old_line.startswith('exit 0'), old_hosts_lines)
    if new_mode == 'ntp':
        pass

    elif new_mode == 'manual':
        new_host_lines.insert(0, 'exit 0\n')

    else:
        raise NotImplemented

    with open(NTP_DATE_CONF_PATH, 'w') as new_file:
        new_file.write(''.join(new_host_lines))


def _edit_ntpd(raw_ui_jnode, new_ui_jnode):
    old_ui_jnode = raw_ui_jnode['rw']
    if not _check_ntp_changed(old_ui_jnode, new_ui_jnode):
        return ACTION_NONE

    old_mode = old_ui_jnode['mode']
    new_mode = new_ui_jnode['mode']
    _edit_ntp_conf(raw_ui_jnode, new_ui_jnode)
    _edit_ntp_date_conf(new_mode)

    if old_mode == 'ntp' and new_mode == 'ntp':
        os.system("service ntp restart")

    elif old_mode == 'manual' and new_mode == 'ntp':
        os.system("update-rc.d ntp enable")
        os.system("service ntp start")

    elif old_mode == 'ntp' and new_mode == 'manual':
        os.system("service ntp stop")
        os.system("update-rc.d ntp disable")

        ts = new_ui_jnode['ts']
        temp = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
        subprocess.Popen(['date --set="%s"' % temp], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # sync BIOS time to OS time
        subprocess.Popen(['hwclock -w'], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    elif old_mode == 'manual' and new_mode == 'manual':
        ts = new_ui_jnode['ts']
        temp = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
        subprocess.Popen(['date --set="%s"' % temp], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # sync BIOS time to OS time
        subprocess.Popen(['hwclock -w'], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # TODO reboot edgeserver if time changed to past

    else:
        raise NotImplemented

    return ACTION_NONE


def _set_db_timezone(iso_tz_name):
    """
    :type iso_tz_name: str
    """
    db_config_jnode = load_db_json()
    db_args = db.DbArgs.deserialize(db_config_jnode['db'])

    def _handler(db_link):
        with db_link:
            with db_link.cursor() as cur:
                cur.execute("select core.setopt('public', 'local_timezone', '%s');" % iso_tz_name)
                changed_row_count = cur.fetchone()
            if changed_row_count[0] != 1:
                raise RuntimeError('setopt %s' % iso_tz_name)

    try:
        db.with_db(db_args, _handler)
    except db.DbError:
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _handle_get(query_args):
    """
    :type query_args: UiQueryArgs
    :rtype: dict
    """
    if query_args.raw_body:
        eventlog.log_debug('fail: bad-request')
        return reply_fail(REPLY_BAD_REQUEST_400)

    # create ui-jnode
    if not os.path.exists(UI_TIME_JNODE_FILEPATH):
        # noinspection PyBroadException
        try:
            create_jnode_from_default(UI_TIME_DEFJNODE_FILEPATH, UI_TIME_JNODE_FILEPATH)

        except:
            eventlog.log_debug_last_exc()
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    # create ui-jnode
    if not os.path.islink(UI_TIME_SCHEMA_FILEPATH) or not os.path.exists(os.readlink(UI_TIME_SCHEMA_FILEPATH)):
        if not os.path.exists(UI_TIME_JNODE_FILEPATH):
            eventlog.log_debug('fail: %s not found' % UI_TIME_JNODE_FILEPATH)
            return reply_fail(REPLY_NOT_FOUND_404)

        create_schema_symlink(UI_TIME_SCHEMA_FILEPATH, UI_TIME_JNODE_FILEPATH)

    try:
        with open(UI_TIME_JNODE_FILEPATH, 'r') as json_file:
            ui_jnode = json.load(json_file)

        return reply_done(ui_jnode['rw'])

    except (IOError, ValueError) as x:
        eventlog.log_debug('fail: load ui-jnode (%s)' % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _handle_put(query_args):
    """
    :type query_args: UiQueryArgs
    :rtype: dict
    """
    # deserialize new-ui-jnode
    if not query_args.raw_body:
        eventlog.log_debug('fail: bad-request')
        return reply_fail(REPLY_BAD_REQUEST_400)

    try:
        new_ui_jnode = json.loads(query_args.raw_body) if query_args.raw_body else None

    except ValueError as x:
        eventlog.log_debug("fail: bad-request (%s)" % x.message)
        return reply_fail(REPLY_BAD_REQUEST_400)

    # load ui-node-schema
    try:
        with open(UI_TIME_SCHEMA_FILEPATH, 'r') as json_file:
            ui_schema_jnode = json.load(json_file)

    except (IOError, ValueError) as x:
        eventlog.log_debug("fail: load ui-jnode-schema (%s)" % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    # validate new-ui-jnode
    try:
        jsonschema.validate(new_ui_jnode, ui_schema_jnode)

    except (ValueError, jsonschema.ValidationError) as x:
        eventlog.log_debug("fail: validate new-ui-jnode (%s)" % x.message)
        return reply_fail(REPLY_BAD_REQUEST_400)

    try:
        with open(UI_TIME_TZ_JNODE_FILEPATH, 'r') as json_file:
            ui_tz_jnode = json.load(json_file)

    except (IOError, ValueError) as x:
        eventlog.log_debug("fail: load ui-tz-jnode (%s)" % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    if new_ui_jnode['tz'] not in ui_tz_jnode:
        eventlog.log_debug("fail: bad-request (param tz has invalid value)")
        return reply_fail(REPLY_BAD_REQUEST_400)

    # load old-ui-jnode
    try:
        with open(UI_TIME_JNODE_FILEPATH, 'r') as json_file:
            raw_ui_jnode = json.load(json_file)
        old_ui_jnode = raw_ui_jnode['rw']

    except (IOError, ValueError) as x:
        eventlog.log_debug("fail: load old-ui-jnode (%s)" % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    if not try_lock_pid_file(UI_TIME_LOCK_FILEPATH):
        eventlog.log_debug("fail: lock %s" % UI_TIME_LOCK_FILEPATH)
        return reply_fail(REPLY_CONFLICT_409)

    _set_db_timezone(new_ui_jnode['tz'])

    if not try_create_backup():
        eventlog.log_debug("fail: create backup")
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    success = False
    # noinspection PyBroadException
    try:
        _edit_ntpd(raw_ui_jnode, new_ui_jnode)

        raw_ui_jnode['rw'] = new_ui_jnode
        with open(UI_TIME_JNODE_FILEPATH, 'w') as json_file:
            json.dump(raw_ui_jnode, json_file, indent=2)

        if old_ui_jnode['tz'] != new_ui_jnode['tz']:
            eventlog.log_debug('restarting renderers')
            subprocess.Popen(["pkill -9 overlayrender"], shell=True)

        success = True

    except:
        eventlog.log_debug_last_exc()
        pass

    if success:
        remove_backup()
        return reply_done(new_ui_jnode)

    if try_restore_backup():
        remove_backup()

    response_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
    return response_jnode


def _run(query_args):
    """
    :type query_args: UiQueryArgs
    :rtype: dict
    """
    if query_args.path_info:
        eventlog.log_debug('fail: not-found')
        return reply_fail(REPLY_NOT_FOUND_404)

    if query_args.query_string:
        eventlog.log_debug('fail: bad-request')
        return reply_fail(REPLY_BAD_REQUEST_400)

    if query_args.http_method == 'GET':
        return _handle_get(query_args)

    if query_args.http_method == 'PUT':
        return _handle_put(query_args)

    return reply_fail(REPLY_METHOD_NOT_ALLOWED_405)


# noinspection PyBroadException
def _main():
    eventlog.open_log("cgi-time", "debug")
    exit_code = 1
    try:
        query_args = ui_types.UiQueryArgs.deserialize()
        reply_jnode = _run(query_args)

        print json.dumps(reply_jnode, separators=(',', ':'))
        exit_code = 0
    except:
        eventlog.log_debug_last_exc()
        print json.dumps(reply_fail(REPLY_INTERNAL_SERVER_ERROR_500), separators=(',', ':'))

    finally:
        eventlog.close_log()

    sys.exit(exit_code)

if __name__ == '__main__':
    _main()
