#!/usr/bin/python
import json
import os
import shutil
import subprocess
import sys
import time
import errno

import edge.admin.ui_types as ui_types
import edge.eventlog as eventlog
import jsonschema
from edge.admin.reply_helpers import *
from edge.fs_helpers import create_schema_symlink, create_jnode_from_default, try_lock_pid_file

UI_ZABBIX_SCHEMA_FILEPATH = '/etc/edge/setupstore/schemas/zabbix.json'
UI_ZABBIX_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/zabbix.json'
UI_ZABBIX_DEFJNODE_FILEPATH = '/opt/edge/share/setupstore/ui/zabbix-jnode.json'

LOCK_FILEPATH = '/var/run/edge/zabbix'

ZABBIX_CONF_PATH = '/etc/zabbix/zabbix_agentd.conf.d/web.conf'
LOCAL_VARUSR = '/opt/varusr/local/etc/zabbix/varusr.conf'

RESTART_NONE = 0
RESTART_ZABBIX = 1

BACKUP_DIRPATH = '/etc/edge/setupstore/backup/zabbix.' + \
                 str(round(time.time() * 1000000)) + '/'


# noinspection PyBroadException
def try_create_backup():
    try:
        os.makedirs(BACKUP_DIRPATH)
        try:
            shutil.copy(ZABBIX_CONF_PATH, BACKUP_DIRPATH)
        except IOError as x:
            if x.errno == errno.ENOENT:
                pass
            else:
                raise
        shutil.copy(UI_ZABBIX_JNODE_FILEPATH, BACKUP_DIRPATH)
        return True

    except Exception:
        pass

    return False


# noinspection PyBroadException
def try_restore_backup():
    try:
        shutil.copy(BACKUP_DIRPATH + 'web.conf', ZABBIX_CONF_PATH)
        shutil.copy(BACKUP_DIRPATH + 'zabbix.json', UI_ZABBIX_JNODE_FILEPATH)
        os.rmdir(BACKUP_DIRPATH)
        return True

    except Exception:
        return False


# noinspection PyBroadException
def remove_backup():
    try:
        shutil.rmtree(BACKUP_DIRPATH)
    except Exception:
        pass


def check_zabbix_changed(old_jnode, new_jnode):
    if old_jnode['enabled'] != new_jnode['enabled']:
        return True
    if new_jnode['enabled']:
        if old_jnode['agent-port'] != new_jnode['agent-port']:
            return True
        if old_jnode['servers'] != new_jnode['servers']:
            return True
    else:
        pass

    return False


def edit_zabbix_conf(old_jnode, new_jnode):
    if os.path.exists(ZABBIX_CONF_PATH):
        with open(ZABBIX_CONF_PATH, 'r') as old_zabbix_file:
            old_zabbix_lines = old_zabbix_file.readlines()
    else:
        old_zabbix_lines = []

    new_zabbix_lines = []
    if len(old_zabbix_lines) > 0:
        for old_line in old_zabbix_lines:

            if not old_line.startswith('#'):
                continue

            new_zabbix_lines.append(old_line)

    else:
        new_zabbix_lines = old_zabbix_lines

    enabled = new_jnode['enabled']
    if enabled:
        new_zabbix_lines.append('\n')
        new_zabbix_lines.append('ListenPort=%s\n' % new_jnode['agent-port'])

        if len(new_jnode['servers']) > 0:
            pass

        else:
            new_jnode['servers'] = old_jnode['servers']

        new_zabbix_lines.append('Server=')
        server_list = map(lambda server: server['host'], new_jnode['servers'])
        new_zabbix_lines.append(', '.join(server_list))

        new_zabbix_lines.append('\nServerActive=')
        server_active_list = map(lambda server: ('%s:%d' % (server['host'], server['port'])), new_jnode['servers'])
        new_zabbix_lines.append(', '.join(server_active_list))

        new_zabbix_lines.append('\nInclude=/etc/edge/zabbix_agentd.d/\n')
        if os.path.exists(LOCAL_VARUSR):
            new_zabbix_lines.append('\nInclude=%s\n' % LOCAL_VARUSR)

    with open(ZABBIX_CONF_PATH, 'w') as new_ntp_file:
        new_ntp_file.write(''.join(new_zabbix_lines))

    if enabled:
        os.system("update-rc.d zabbix-agent enable")
        os.system("service zabbix-agent restart")
    else:
        os.system("service zabbix-agent stop")
        os.system("update-rc.d zabbix-agent disable")


def edit_zabbix(old_jnode, new_jnode):
    if not check_zabbix_changed(old_jnode, new_jnode):
        return RESTART_NONE

    edit_zabbix_conf(old_jnode, new_jnode)
    return RESTART_ZABBIX


def restart_module(module):
    """
    :type module: str
    """
    subprocess.Popen(["kill -9 `ps ax |"
                      "grep \"%s --id.0\" |"
                      "grep -v grep |"
                      "awk '{print $1}'`" % module],
                     shell=True)


def handle_get(query_args):
    """
    :type query_args: UiQueryArgs
    :rtype: dict
    """
    if query_args.path_info:
        eventlog.log_debug('fail: not-found')
        return reply_fail(REPLY_NOT_FOUND_404)

    if query_args.query_string:
        eventlog.log_debug('fail: bad-request')
        return reply_fail(REPLY_BAD_REQUEST_400)

    # create ui-jnode
    if not os.path.exists(UI_ZABBIX_JNODE_FILEPATH):
        # noinspection PyBroadException
        try:
            create_jnode_from_default(UI_ZABBIX_DEFJNODE_FILEPATH, UI_ZABBIX_JNODE_FILEPATH)

        except:
            eventlog.log_debug_last_exc()
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    # create ui-schema
    if not os.path.islink(UI_ZABBIX_SCHEMA_FILEPATH) or not os.path.exists(os.readlink(UI_ZABBIX_SCHEMA_FILEPATH)):
        if not os.path.exists(UI_ZABBIX_JNODE_FILEPATH):
            eventlog.log_debug('fail: %s not found' % UI_ZABBIX_JNODE_FILEPATH)
            return reply_fail(REPLY_NOT_FOUND_404)

        create_schema_symlink(UI_ZABBIX_SCHEMA_FILEPATH, UI_ZABBIX_JNODE_FILEPATH)

    try:
        with open(UI_ZABBIX_JNODE_FILEPATH, 'r') as json_file:
            ui_jnode = json.load(json_file)

        return reply_done(ui_jnode['rw'])

    except (IOError, ValueError) as x:
        eventlog.log_debug('fail: load ui-jnode (%s)' % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def handle_put(raw_ui_jnode, new_ui_jnode):
    """
    :type raw_ui_jnode: dict
    :type new_ui_jnode: dict
    :rtype: dict
    """
    # load ui-jnode-schema
    try:
        with open(UI_ZABBIX_SCHEMA_FILEPATH, 'r') as json_file:
            ui_schema_jnode = json.load(json_file)

    except (IOError, ValueError) as x:
        eventlog.log_debug("fail: load ui-jnode-schema (%s)" % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    # validate new-ui-jnode
    try:
        jsonschema.validate(new_ui_jnode, ui_schema_jnode)

    except (ValueError, jsonschema.ValidationError) as x:
        eventlog.log_debug("fail: validate new-ui-jnode (%s)" % x.message)
        return reply_fail(REPLY_BAD_REQUEST_400)

    # change zabbix conf-file
    if not try_lock_pid_file(LOCK_FILEPATH):
        return reply_fail(REPLY_CONFLICT_409)

    if not try_create_backup():
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    old_ui_jnode = raw_ui_jnode['rw']

    success = False
    try:
        restart_mask = 0
        restart_mask |= edit_zabbix(old_ui_jnode, new_ui_jnode)

        raw_ui_jnode['rw'] = new_ui_jnode
        with open(UI_ZABBIX_JNODE_FILEPATH, 'w') as f:
            f.write((json.dumps(raw_ui_jnode, indent=2)))

        success = True

    except Exception:
        eventlog.log_debug_last_exc()
        pass

    if success:
        remove_backup()
        return reply_done(new_ui_jnode)

    if try_restore_backup():
        remove_backup()

    return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _run(query_args):
    """
    :type query_args: UiQueryArgs
    :rtype: dict
    """
    if query_args.http_method == 'GET':
        return handle_get(query_args)

    with open(UI_ZABBIX_JNODE_FILEPATH, 'r') as json_file:
        raw_ui_jnode = json.load(json_file)

    if query_args.http_method == 'PUT':
        new_ui_jnode = json.loads(query_args.raw_body)
        response_jnode = handle_put(raw_ui_jnode, new_ui_jnode)

    else:
        response_jnode = reply_fail(REPLY_METHOD_NOT_ALLOWED_405)

    return response_jnode


def _main():
    eventlog.open_log("cgi-zabbix", "debug")
    exit_code = 1
    try:
        query_args = ui_types.UiQueryArgs.deserialize()
        reply_jnode = _run(query_args)

        print json.dumps(reply_jnode, separators=(',', ':'))
        exit_code = 0
    except:
        eventlog.log_debug_last_exc()
        print json.dumps(reply_fail(REPLY_INTERNAL_SERVER_ERROR_500), separators=(',', ':'))

    finally:
        eventlog.close_log()

    sys.exit(exit_code)


if __name__ == '__main__':
    _main()
