#!/usr/bin/python

import os
import sys
import datetime
import time

import ujson
import psycopg2
import psycopg2.extras

import edge.eventlog as eventlog

HTTP_GET = 'GET'

DBNAME = 'autocode'
DBUSER = 'autocode'
DBPASS = 'autocode'


def _response_success(result_jnode):
    jnode = {
        'result': result_jnode,
        'status': 200
    }

    return jnode


def _response_fail(http_code):
    jnode = {
        'errors': [],
        'status': http_code
    }

    return jnode


class _ReportArgs:
    def __init__(self, min_ts, max_ts, min_ts_gap, min_total_count):
        """
        :type min_ts: int
        :type max_ts: int
        :type min_ts_gap: int
        :type min_total_count: int
        """
        self.min_ts = min_ts
        self.max_ts = max_ts
        self.min_ts_gap = min_ts_gap
        self.min_total_count = min_total_count

    @staticmethod
    def deserialize(query_string):
        """
        :type query_string: str
        :rtype: _ReportArgs
        """
        min_ts = None
        max_ts = None
        min_ts_gap = None
        min_total_count = None

        if len(query_string) > 1:
            entry_set = query_string[1:].split('&')
            for entry in entry_set:
                key_value = entry.split('=', 1)
                if len(key_value) != 2:
                    eventlog.log_debug("fail: invalid query-arg")
                    return None

                key, value = key_value
                if key == "min-ts":
                    try:
                        raw_min_ts = int(value)
                        if 0 <= raw_min_ts <= 2147476447000000:
                            min_ts = raw_min_ts
                        else:
                            eventlog.log_debug("fail: invalid min-ts")
                            return None
                    except ValueError:
                        eventlog.log_debug("fail: invalid min-ts")
                        return None

                elif key == "max-ts":
                    try:
                        raw_max_ts = int(value)
                        if 0 <= raw_max_ts <= 2147476447000000:
                            max_ts = raw_max_ts
                        else:
                            eventlog.log_debug("fail: invalid max-ts")
                            return None
                    except ValueError:
                        eventlog.log_debug("fail: invalid max-ts")
                        return None

                elif key == "min-ts-gap":
                    try:
                        raw_min_ts_gap = int(value)
                        if 0 <= raw_min_ts_gap <= 65535:
                            eventlog.log_debug("raw_min_ts_gap: %s" % raw_min_ts_gap)
                            min_ts_gap = raw_min_ts_gap
                        else:
                            eventlog.log_debug("fail: invalid min-ts-gap")
                            return None
                    except ValueError:
                        eventlog.log_debug("fail: invalid min-ts-gap")
                        return None

                elif key == "min-total-count":
                    try:
                        raw_min_total_count = int(value)
                        if 1 <= raw_min_total_count <= 65535:
                            min_total_count = raw_min_total_count
                        else:
                            eventlog.log_debug("fail: invalid min-total-count")
                            return None

                    except ValueError:
                        eventlog.log_debug("fail: invalid min-total-count")
                        return None

                else:
                    eventlog.log_debug("fail: invalid query-arg")
                    return None

        # assign default values
        if min_ts is None:
            default_min_dt = datetime.datetime.now() - datetime.timedelta(days=7)  # 7 days ago
            min_ts = int(time.mktime(default_min_dt.timetuple())) * 1000000
        if max_ts is None:
            max_ts = int(time.time()) * 1000000
        if min_ts_gap is None:
            min_ts_gap = 10
        if min_total_count is None:
            min_total_count = 2

        max_dt = datetime.date.fromtimestamp(max_ts / 1000000)
        min_dt = datetime.date.fromtimestamp(min_ts / 1000000)
        if (max_dt - min_dt).days > 7:
            eventlog.log_debug("fail: invalid time period")
            return None

        # eventlog.log_debug("min_ts %s" % min_ts)
        # eventlog.log_debug("max_ts %s" % max_ts)
        # eventlog.log_debug("min_ts_gap %s" % min_ts_gap)
        # eventlog.log_debug("min_total_count %s" % min_total_count)

        # make report
        if min_ts and max_ts and min_ts_gap and min_total_count:
            return _ReportArgs(min_ts, max_ts, min_ts_gap, min_total_count)

        else:
            eventlog.log_debug("fail: invalid number of query-arg")
            return None


def _get_report(report_args):
    """
    :type report_args: _ReportArgs
    :rtype: list of dict | None
    """
    dbapi_jnode = {
        '@header': {
            'vendor': 'VIT',
            'class': 'vitml.edge.api',
            'version': '0.1',
            'object': 'report',
            'action': 'get'
        },
        'name': 'plate.scurry',
        'param': {
            'fromtime': report_args.min_ts,
            'totime': report_args.max_ts,
            'mingap': report_args.min_ts_gap,
            'mindet': report_args.min_total_count
        }
    }
    psycopg2.extras.register_default_jsonb(globally=True)
    try:
        with psycopg2.connect(database=DBNAME, user=DBUSER, password=DBPASS) as conn:
            cursor = conn.cursor()
            cursor.execute("select api_report_get(%s::jsonb)", (psycopg2.extras.Json(dbapi_jnode, dumps=ujson.dumps),))
            raw_report = cursor.fetchone()[0]

        if not raw_report or raw_report['errcode'] != 0:
            eventlog.log_debug('fail: api_report_get (%s)' % raw_report['errmsg'])
            return None

        # eventlog.log_debug('raw_report: %s' % ujson.dumps(raw_report))

        result_jnode = raw_report['result']
        if len(result_jnode) > 0:
            expected_set = frozenset(['plate-text', 'total-count', 'last-guid', 'last-datetime', 'last-ts'])
            actual_set = set(result_jnode[0].keys())
            if len(actual_set.symmetric_difference(expected_set)) > 0:
                eventlog.log_debug("fail: invalid report from db")
                return None

        return result_jnode

    except psycopg2.Error as e:
        eventlog.log_debug('fail: db exception (%s)' % e.message)

    return None


def _run(http_method, query_string):
    """
    :type http_method: str
    :type query_string: str
    :rtype: dict
    """
    # Dispatch http-method.
    if http_method != HTTP_GET:
        eventlog.log_debug("fail: unsupported http-method")
        return _response_fail(405)

    report_args = _ReportArgs.deserialize(query_string)
    if not report_args:
        eventlog.log_debug('fail: deserialize report-args')
        return _response_fail(400)

    report = _get_report(report_args)
    if report is None:
        eventlog.log_debug('fail: get-report')
        return _response_fail(500)

    return _response_success(report)


def main():
    eventlog.open_log('eventstore-fact', 'debug')
    try:
        http_mode = os.environ['REQUEST_METHOD']
        path_info = os.environ['PATH_INFO']
        query_string = os.environ['QUERY_STRING']

        if path_info:
            eventlog.log_debug("fail: invalid path-info")
            report_jnode = _response_fail(404)
        else:
            report_jnode = _run(http_mode, query_string)

        print ujson.dumps(report_jnode)
        sys.exit(0)

    finally:
        eventlog.close_log()


if __name__ == '__main__':
    main()
