import os
import sys
import ujson
import datetime
import iso8601
import urlparse
import pytz

from edge.admin.reply_helpers import *
from edge.pipeline_helpers import load_json_file, get_pipeline_jnode_set, load_db_json
import edge.db as edge_db
import edge.path_helpers as path_helpers
import edge.eventlog as eventlog


_UI_TIME_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/time.json'
_MAX_REPORT_DAYS = 14


class ReportError(RuntimeError):
    pass


def _fail_param(param_key):
    """
    :type param_key: str
    :rtype: str
    """
    return 'fail: param <%s> has invalid value' % param_key


def _fail_param2(param_key, reason):
    """
    :type param_key: str
    :type reason: str
    :rtype: str
    """
    return 'fail: param <%s> has invalid value (%s)' % (param_key, reason)


def _get_edge_tzinfo():
    """
    :rtype: datetime.tzinfo
    :raises: ReportError
    """
    try:
        with open(_UI_TIME_JNODE_FILEPATH, 'r') as json_file:
            ui_jnode = ujson.load(json_file)

        ui_tz = ui_jnode['rw']['tz']
        return pytz.timezone(ui_tz)
    except (IOError, ValueError, KeyError) as x:
        eventlog.log_debug('fail: load ui-jnode (%s)' % x.message)
        raise ReportError


def _normalize_iso_datetime(iso_datetime):
    """
    :type iso_datetime: str
    :rtype: (str, int)
    :raises: ReportError
    """
    # define tz-offset
    aware_dt = iso8601.parse_date(iso_datetime, default_timezone=None)
    if aware_dt.tzinfo is None:
        # in case of 'YYYY-MM-DDThh:mm:ss' use edge-tz-offset
        dt_tzinfo = _get_edge_tzinfo()
    else:
        # in case of 'YYYY-MM-DDThh:mm:ss<tz-offset>' use supplied tz-offset
        dt_tzinfo = aware_dt.tzinfo

    tz_offset = dt_tzinfo.utcoffset(aware_dt)
    if tz_offset is None:
        eventlog.log_debug('fail: define tz-offset')
        raise ReportError

    # remove tz-offset from iso_datetime if present
    naive_iso_datetime = aware_dt.replace(tzinfo=None).isoformat()

    return naive_iso_datetime, int(tz_offset.total_seconds())


def _deserialize_traffic_flow_args(query_string):
    """
    :type query_string: str
    :rtype: dict, dict, int
    :raises: ReportError
    """
    url_query = urlparse.parse_qs(query_string[1:], strict_parsing=True)

    limits_jnode = dict()
    api_query_jnode = {
        'ordinate': 'none',
        'datetime-step': 'day'
    }

    iso_datetime_min = None
    iso_datetime_max = None
    for key, values in url_query.iteritems():
        # deserialize channel
        if key == 'channel':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            try:
                db_channel = int(values.pop())
            except ValueError:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['channel'] = db_channel
            api_query_jnode['channel'] = db_channel
            continue

        # deserialize datetime-max
        if key == 'datetime-max':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            iso_datetime_max = values.pop()
            try:
                iso8601.parse_date(iso_datetime_max)
            except iso8601.ParseError as x:
                eventlog.log_debug(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-max'] = iso_datetime_max
            continue

        # deserialize datetime-min
        if key == 'datetime-min':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            iso_datetime_min = values.pop()
            try:
                iso8601.parse_date(iso_datetime_min)
            except iso8601.ParseError as x:
                eventlog.log_debug(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-min'] = iso_datetime_min
            continue

        # deserialize datetime-step
        if key == 'datetime-step':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            datetime_step = values.pop()
            if datetime_step not in ['hour', 'day']:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['datetime-step'] = datetime_step
            api_query_jnode['datetime-step'] = datetime_step
            continue

        # deserialize plate-stat
        if key == 'plate-stat':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            plate_stat = values.pop()
            if plate_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['plate-stat'] = plate_stat
            api_query_jnode['plate-stat'] = plate_stat
            continue

        # deserialize speed-stat
        if key == 'speed-stat':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            speed_stat = values.pop()
            if speed_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['speed-stat'] = speed_stat
            api_query_jnode['speed-stat'] = speed_stat
            continue

        # deserialize speed-min
        if key == 'speed-min':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            try:
                speed_min = int(values.pop())
            except ValueError:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            if not 0 <= speed_min <= 255:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['speed-min'] = speed_min
            api_query_jnode['speed-min'] = speed_min
            continue

        # deserialize speed-max
        if key == 'speed-max':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            try:
                speed_max = int(values.pop())
            except ValueError:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            if not 0 <= speed_max <= 255:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['speed-max'] = speed_max
            api_query_jnode['speed-max'] = speed_max
            continue

        # deserialize lane-stat
        if key == 'lane-stat':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            lane_stat = values.pop()
            if lane_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['lane-stat'] = lane_stat
            api_query_jnode['lane-stat'] = lane_stat
            continue

        # deserialize lane
        if key == 'lane':
            lane_jnode = []
            for raw_lane in values:
                try:
                    lane = int(raw_lane)
                except ValueError:
                    eventlog.log_debug(_fail_param(key))
                    raise ReportError

                if not 1 <= int(lane) <= 64:
                    eventlog.log_debug(_fail_param(key))
                    raise ReportError

                lane_jnode.append(lane)

            limits_jnode['lane'] = lane_jnode
            api_query_jnode['lane'] = lane_jnode
            continue

        # deserialize alarm
        if key == 'alarm':
            alarm_jnode = []
            for alarm_type in values:
                if not _validate_alarm_type(alarm_type):
                    eventlog.log_debug(_fail_param(key))
                    raise ReportError

                alarm_jnode.append(alarm_type)

            limits_jnode['alarm'] = alarm_jnode
            api_query_jnode['alarm'] = alarm_jnode
            continue

        eventlog.log_debug('fail: param <%s> is not supported' % key)
        raise ReportError

    # define iso-datetime-max
    if iso_datetime_max is None:
        iso_datetime_max = datetime.datetime.now().isoformat()

    naive_iso_dt_max, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_max)
    api_query_jnode['datetime-max'] = naive_iso_dt_max

    # define iso-datetime-min
    if iso_datetime_min is None:
        datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
        datetime_min = datetime_max - datetime.timedelta(days=7)
        iso_datetime_min = datetime_min.isoformat()

    naive_iso_dt_min, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_min)
    api_query_jnode['datetime-min'] = naive_iso_dt_min

    # validate date period
    datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
    datetime_min = iso8601.parse_date(api_query_jnode['datetime-min'])
    if datetime_max - datetime_min > datetime.timedelta(days=_MAX_REPORT_DAYS):
        eventlog.log_debug('fail: report period is too long')
        raise ReportError

    return limits_jnode, api_query_jnode, api_query_tz_offset


def _deserialize_traffic_alarm_args(query_string):
    """
    :type query_string: str
    :rtype: dict, dict, int
    :raises ReportError
    """
    url_query = urlparse.parse_qs(query_string[1:], strict_parsing=True)

    api_query_jnode = {
        'ordinate': 'alarm',
        'datetime-step': 'day'
    }

    limits_jnode = dict()

    iso_datetime_min = None
    iso_datetime_max = None

    for key, values in url_query.iteritems():
        # deserialize channel
        if key == 'channel':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            try:
                channel = int(values.pop())
            except ValueError:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['channel'] = channel
            api_query_jnode['channel'] = channel
            continue

        # deserialize datetime-max
        if key == 'datetime-max':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            iso_datetime_max = values.pop()
            try:
                iso8601.parse_date(iso_datetime_max)
            except iso8601.ParseError as x:
                eventlog.log_debug(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-max'] = iso_datetime_max
            continue

        # deserialize datetime-min
        if key == 'datetime-min':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            iso_datetime_min = values.pop()
            try:
                iso8601.parse_date(iso_datetime_min)
            except iso8601.ParseError as x:
                eventlog.log_debug(_fail_param2(key, str(x)))
                raise ReportError

            limits_jnode['datetime-min'] = iso_datetime_min
            continue

        # deserialize datetime-step
        if key == 'datetime-step':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            datetime_step = values.pop()
            if datetime_step not in ['hour', 'day']:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['datetime-step'] = datetime_step
            api_query_jnode['datetime-step'] = datetime_step
            continue

        # deserialize lane-stat
        if key == 'lane-stat':
            if len(values) != 1:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            lane_stat = values.pop()
            if lane_stat not in ['undefined', 'defined', 'any']:
                eventlog.log_debug(_fail_param(key))
                raise ReportError

            limits_jnode['lane-stat'] = lane_stat
            api_query_jnode['lane-stat'] = lane_stat
            continue

        # deserialize lane
        if key == 'lane':
            lane_jnode = []
            for raw_lane in values:
                try:
                    lane = int(raw_lane)
                except ValueError:
                    eventlog.log_debug(_fail_param(key))
                    raise ReportError

                if not 1 <= int(lane) <= 64:
                    eventlog.log_debug(_fail_param(key))
                    raise ReportError

                lane_jnode.append(lane)

            limits_jnode['lane'] = lane_jnode
            api_query_jnode['lane'] = lane_jnode
            continue

        # deserialize alarm
        if key == 'alarm':
            alarm_jnode = []
            for alarm_type in values:
                if not _validate_alarm_type(alarm_type):
                    eventlog.log_debug(_fail_param(key))
                    raise ReportError

                alarm_jnode.append(alarm_type)

            limits_jnode['alarm'] = alarm_jnode
            api_query_jnode['alarm'] = alarm_jnode
            continue

        eventlog.log_debug('fail: param <%s> is not supported' % key)
        raise ReportError

    # define iso-datetime-max
    if iso_datetime_max is None:
        iso_datetime_max = datetime.datetime.now().isoformat()

    naive_iso_dt_max, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_max)
    api_query_jnode['datetime-max'] = naive_iso_dt_max

    # define iso-datetime-min
    if iso_datetime_min is None:
        datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
        datetime_min = datetime_max - datetime.timedelta(days=7)
        iso_datetime_min = datetime_min.isoformat()

    naive_iso_dt_min, api_query_tz_offset = _normalize_iso_datetime(iso_datetime_min)
    api_query_jnode['datetime-min'] = naive_iso_dt_min

    # validate date period
    datetime_max = iso8601.parse_date(api_query_jnode['datetime-max'])
    datetime_min = iso8601.parse_date(api_query_jnode['datetime-min'])
    if datetime_max - datetime_min > datetime.timedelta(days=_MAX_REPORT_DAYS):
        eventlog.log_debug('fail: report period is too long')
        raise ReportError

    return limits_jnode, api_query_jnode, api_query_tz_offset


def _get_report(api_query_jnode, api_query_tz_offset):
    """
    :type api_query_jnode: dict
    :type api_query_tz_offset: int
    :rtype: dict
    :raises: ReportError
    """
    dbapi_query_jnode = {
        '@header': {
            'vendor': 'VIT',
            'class': 'vitml.edge.api',
            'version': '0.1',
            'object': 'report',
            'action': 'get'
        },
        'name': 'plate.traffic.combined',
        'param': api_query_jnode
    }

    db_jnode = load_db_json()
    db_args = edge_db.DbArgs.deserialize(db_jnode['db'])

    def db_handler(db_link):
        """
        :rtype: dict
        :raises: edge_db.DbError
        :raises: ReportError
        """
        with db_link, db_link.cursor() as db_cursor:
            db_cursor.execute(
                "SET LOCAL TIME ZONE INTERVAL '%sS';select api_report_get(%s)",
                [api_query_tz_offset, edge_db.DbJsonArg(dbapi_query_jnode)])

            api_reply_jnode = db_cursor.fetchone()[0]
            if not api_reply_jnode or api_reply_jnode['errcode'] != 0:
                eventlog.log_debug('fail: api_report_get (%s)' % api_reply_jnode['errmsg'])
                raise ReportError

            report_jnode = api_reply_jnode['result']
            return report_jnode

    try:
        return edge_db.with_db(db_args, db_handler)

    except edge_db.DbError:
        eventlog.log_debug('fail: db')
        raise ReportError


def _get_limits(require_alarm):
    """
    :type require_alarm: bool
    :rtype: dict 
    """
    channels_jnode = {}
    limits_jnode = {
        'channels': channels_jnode
    }

    pipeline_jnode_set = get_pipeline_jnode_set()
    if pipeline_jnode_set:
        for pipeline_jnode in pipeline_jnode_set:
            pid = pipeline_jnode['id']
            es_jnode_filepath = path_helpers.get_ui_pipeline_jnode_filepath(
                pid, path_helpers.UI_TYPE_EVENTSTORE, '0')
            es_jnode = load_json_file(es_jnode_filepath)

            alarm_jnode = _get_alarmtypes(pid)
            if require_alarm and not alarm_jnode:
                continue

            channel = es_jnode['rw']['channel']
            channels_jnode[str(channel)] = {
                'lane': _get_lanes(pid),
                'alarm': alarm_jnode
            }

    return limits_jnode


def _get_lanes(pid):
    """
    :param pid: str 
    :return: list of int
    """
    jnode_filepath = path_helpers.get_ui_pipeline_jnode_filepath(
        pid, path_helpers.UI_TYPE_LANE_ENFORCEMENT, '0')
    ui_jnode = load_json_file(jnode_filepath)
    lanes_jnode = ui_jnode['rw']['road']['lanes']
    lane_list = map(lambda lane: lane['id'], lanes_jnode)
    lane_list.sort()

    return lane_list


def _get_alarmtypes(pid):
    """
    :type pid: str 
    :rtype: list of str
    :raises: ReportError
    """
    # TODO: get alarmtypes from DB
    pipelines_jnode = filter(
        lambda pipeline_jnode: pipeline_jnode['id'] == pid,
        get_pipeline_jnode_set())

    if len(pipelines_jnode) != 1:
        eventlog.log_debug('fail: define pipeline-jnode')
        raise ReportError

    modules_jnode = pipelines_jnode[0]['modules']

    if filter(lambda x: x['type'] == path_helpers.UI_TYPE_SPEED_ENFORCEMENT, modules_jnode):
        alarm_types = ['_any', '_none', 'speedviolation']
    else:
        alarm_types = []

    return alarm_types


def _validate_alarm_type(alarm_type):
    """
    :type alarm_type: str
    :rtype: bool
    """
    return alarm_type in ['_any', '_none', 'speedviolation']


def _run(http_method, query_string, path_info):
    """
    :type http_method: str
    :type query_string: str
    :type path_info: list of str
    :rtype: dict
    """

    # validate url-path
    valid_url_path_list = [
        ['traffic-flow'],
        ['traffic-flow', 'limits'],
        ['traffic-alarm'],
        ['traffic-alarm', 'limits']
    ]

    if path_info not in valid_url_path_list:
        eventlog.log_debug("fail: not-found")
        return reply_fail(REPLY_NOT_FOUND_404)

    # validate http-method
    if http_method != 'GET':
        eventlog.log_debug('fail: method-not-allowed (%s)' % http_method)
        return reply_fail(REPLY_METHOD_NOT_ALLOWED_405)

    # reply report-limits
    if len(path_info) == 2:
        try:
            if path_info[0] == 'traffic-flow':
                reply_jnode = _get_limits(require_alarm=False)

            elif path_info[0] == 'traffic-alarm':
                reply_jnode = _get_limits(require_alarm=True)

            else:
                eventlog.log_debug('fail: report is not supported')
                raise ReportError

            return reply_done(reply_jnode)

        except ReportError:
            eventlog.log_debug('fail: get report-limits')
            raise

    # define report-args
    try:
        if path_info[0] == 'traffic-flow':
            limits_jnode, api_query_jnode, api_query_tz_offset \
                = _deserialize_traffic_flow_args(query_string)

        elif path_info[0] == 'traffic-alarm':
            limits_jnode, api_query_jnode, api_query_tz_offset = \
                _deserialize_traffic_alarm_args(query_string)

        else:
            eventlog.log_debug('fail: report is not supported')
            raise ReportError

    except ReportError:
        eventlog.log_debug('fail: bad-request (deserialize query-string)')
        return reply_fail(REPLY_BAD_REQUEST_400)

    # reply report
    try:
        report_jnode = _get_report(api_query_jnode, api_query_tz_offset)
        report_jnode['limits'] = limits_jnode
        return reply_done(report_jnode)

    except ReportError:
        eventlog.log_debug('fail: get-report')
        raise


def _main():
    exit_code = 1
    eventlog.open_log('eventstore-traffic', 'debug')
    # noinspection PyBroadException
    try:
        http_method = os.environ['REQUEST_METHOD']
        raw_path_info = os.environ['PATH_INFO']
        query_string = os.environ['QUERY_STRING']
        eventlog.log_debug("%s|%s|%s" % (http_method, raw_path_info, query_string))

        path_info = raw_path_info.split('/')[1:]
        reply_jnode = _run(http_method, query_string, path_info)
        print ujson.dumps(reply_jnode)
        exit_code = 0

    except ReportError:
        print ujson.dumps(reply_fail(REPLY_INTERNAL_SERVER_ERROR_500))

    except:
        print ujson.dumps(reply_fail(REPLY_INTERNAL_SERVER_ERROR_500))
        eventlog.log_debug_last_exc()

    finally:
        eventlog.close_log()
        sys.exit(exit_code)

if __name__ == '__main__':
    _main()
