import time
import datetime
from os.path import isabs
import pytz
import ujson

import edge.eventlog as eventlog
import edge.pipeline_helpers
from edge.types import FormatTemplate

from _models import Geodata, Fact


class ArgsError(RuntimeError):
    pass


class GpsdArgs:
    def __init__(self, geodata, endpoint):
        """
        :type geodata: Geodata | None
        :type endpoint: (str, int) | None
        """
        self.geodata = geodata
        self.endpoint = endpoint

    @staticmethod
    def deserialzie(geodata_args):
        """
        :type geodata_args: list | None
        :rtype: GpsdArgs
        :raises: ValueError
        """
        # deserialize geodata
        geodata = None
        endpoint = ('localhost', 2947)

        if geodata_args is None:
            pass

        elif geodata_args[0] == 'static':
            _, latitude, longitude = geodata_args
            if not (type(latitude) in [int, float] and -90.0 <= latitude <= +90.0):
                eventlog.log_error('param #/Geodata/Latitude has invalid value')
                raise ValueError

            if not (type(longitude) in [int, float] and -180.0 <= longitude <= +180.0):
                eventlog.log_error('param #/Geodata/Longitude has invalid value')
                raise ValueError
            # http://catb.org/gpsd/gpsd_json.html
            geodata = Geodata(latitude, longitude, altitude=None, speed=None)

        elif geodata_args[0] == 'gpsd':
            _, gpsd_host, gpsd_port = geodata_args
            if not (type(gpsd_host) in [str, unicode] and len(gpsd_host) <= 32):
                eventlog.log_error('param #/Gpsd/Host has invalid value')
                raise ValueError

            if not (type(gpsd_port) in [int] and 0 < gpsd_port <= 65535):
                eventlog.log_error('param #/Gpsd/Port has invalid value')
                raise ValueError

            endpoint = (gpsd_host, gpsd_port)

        else:
            raise ValueError

        return GpsdArgs(geodata, endpoint)


class RaasArgs:
    def __init__(self, raas_host, raas_port, raas_user, raas_pass):
        """
        :type raas_host: str
        :type raas_port: str
        :type raas_user str
        :type raas_pass: str
        """
        self.raas_host = raas_host
        self.raas_port = raas_port
        self.raas_user = raas_user
        self.raas_pass = raas_pass

    @staticmethod
    def deserialize(raas_jnode):
        """
        :type raas_jnode: dict
        :rtype: RaasArgs
        """
        host_name = str(raas_jnode.get('host'))
        host_port = str(raas_jnode.get('port'))
        user_name = str(raas_jnode.get('user'))
        user_secret = str(raas_jnode.get('pass'))

        if host_name and host_port and user_name and user_secret:
            raas_scene = RaasArgs(host_name, host_port, user_name, user_secret)
            return raas_scene

        raise ValueError


class LaneDetectorArgs:
    def __init__(self, id_to_lane_list, img_width, img_height):
        """
        :type id_to_lane_list: list of (int, list of (int, int))
        :type img_width: int
        :type img_height: int
        """
        self.id_to_lane_list = id_to_lane_list
        self.img_width = img_width
        self.img_height = img_height

    @staticmethod
    def deserialize(road_mask_jnode):
        """
        :type road_mask_jnode: dict
        :rtype: LaneDetectorArgs
        """
        id_to_lane_list = map(lambda lane_jnode: (lane_jnode['id'], lane_jnode['mask']),
                              road_mask_jnode.get('lanes', []))

        resolution_jnode = road_mask_jnode.get('resolution', {})
        img_w = resolution_jnode.get('w', 0)
        img_h = resolution_jnode.get('h', 0)

        return LaneDetectorArgs(id_to_lane_list, img_w, img_h)


class PlateAbsentDbArgs:
    def __init__(self, db_directory, lanes, min_time, max_time):
        """
        :type db_directory: str
        :type min_time: datetime.time
        :type max_time: datetime.time
        :type lanes list of int
        """
        self.db_directory = db_directory
        self.min_time = min_time
        self.max_time = max_time
        self.lanes = lanes

    @staticmethod
    def deserialize(plate_absent_db_jnode):
        if not plate_absent_db_jnode:
            return None

        try:
            db_directory = plate_absent_db_jnode['directory']
            time_min_max = plate_absent_db_jnode['limits'].get('time-min-max', [])
            lanes = plate_absent_db_jnode['limits'].get('lanes', [])

            min_time = datetime.time(hour=0, minute=0, second=0)
            max_time = datetime.time(hour=23, minute=59, second=59)
            if len(time_min_max) == 2:
                raw_min_time, raw_max_time = time_min_max
                min_struct_time = time.strptime(raw_min_time, "%H:%M")
                max_struct_time = time.strptime(raw_max_time, "%H:%M")

                min_time = datetime.time(min_struct_time.tm_hour, min_struct_time.tm_min)
                max_time = datetime.time(max_struct_time.tm_hour, max_struct_time.tm_min)

            return PlateAbsentDbArgs(db_directory, lanes, min_time, max_time)
        except KeyError:
            eventlog.log_debug('fail: plate_absent_db jnode have invalid keys')

        return None


class CountersArgs:
    def __init__(self, stat_filepath, stat_tout_sec):
        """
        :type stat_filepath: str
        :type stat_tout_sec: int
        """
        self.stat_filepath = stat_filepath
        self.stat_tout_sec = stat_tout_sec

    @staticmethod
    def deserialize(monitoring_jnode):
        """
        :type monitoring_jnode: dict
        :rtype: CountersArgs
        """
        stat_filepath = monitoring_jnode.get('StatPath')
        if not (unicode == type(stat_filepath) and len(stat_filepath) < 64 and isabs(stat_filepath)):
            raise ValueError('main.conf#/Monitoring/StatPath')

        raw_stat_tout_sec = monitoring_jnode.get('StatTout')
        if not (int == type(raw_stat_tout_sec) and 0 <= raw_stat_tout_sec <= 255):
            raise ValueError('main.conf#/Monitoring/StatTout')

        args = CountersArgs(stat_filepath, raw_stat_tout_sec)
        return args


class MediaStoreArgs:
    def __init__(self, endpoint, channel):
        """
        :type endpoint: str
        :type channel: str
        """
        self.endpoint = endpoint
        self.channel = channel

    @staticmethod
    def deserialize(mediastore_jnode):
        """
        :type mediastore_jnode: list of str
        :rtype: MediaStoreArgs
        """
        endpoint, channel = mediastore_jnode
        return MediaStoreArgs(endpoint, channel)


class JpegQueueArgs:
    def __init__(self, endpoint, min_file_count, max_file_count):
        """
        :type endpoint: str
        :type min_file_count: int
        :type max_file_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_file_count
        self.max_file_count = max_file_count
        self.log_level = 'info'

    @staticmethod
    def deserialize(store_jnode):
        """
        :type store_jnode: dict
        :rtype: JpegQueueArgs
        """
        endpoint = store_jnode.get('endpoint')
        if not (unicode == type(endpoint) and len(endpoint) < 64 and isabs(endpoint)):
            raise ValueError('main.conf#/store/endpoint')

        max_file_count = store_jnode.get('max-file-count', 50)
        if not (int == type(max_file_count) and 15 <= max_file_count <= 200):
            raise ValueError('main.conf#/store/max-file-count')

        return JpegQueueArgs(endpoint, max_file_count - 5, max_file_count)


class JsonQueueArgs:
    def __init__(self, endpoint, min_file_count, max_file_count):
        """
        :type endpoint: str
        :type min_file_count: int
        :type max_file_count: int
        """
        self.endpoint = endpoint
        self.min_file_count = min_file_count
        self.max_file_count = max_file_count


class EventConsumerArgs:
    def __init__(self, require_plate, json_queue_args_list, jpeg_queue_args):
        """
        :type require_plate: bool
        :type json_queue_args_list: list of JsonQueueArgs
        :type jpeg_queue_args: JpegQueueArgs
        """
        self.require_plate = require_plate
        self.json_queue_args_list = json_queue_args_list
        self.jpeg_queue_args = jpeg_queue_args

    @staticmethod
    def deserialize(fact_event_jnode, store_jnode):
        """
        :type fact_event_jnode: dict
        :type store_jnode: dict
        :rtype: EventConsumerArgs
        """
        max_file_count = fact_event_jnode.get('max-file-count', 15)
        if not (int == type(max_file_count) and 15 <= max_file_count <= 50):
            raise ValueError('main.conf#/event-store/max-file-count')

        json_queue_args_list = map(
            lambda endpoint: JsonQueueArgs(endpoint, max_file_count - 5, max_file_count),
            fact_event_jnode['endpoints'])

        jpeg_queue_args = JpegQueueArgs.deserialize(store_jnode)

        require_plate = fact_event_jnode.get('require-plate', True)
        return EventConsumerArgs(require_plate, json_queue_args_list, jpeg_queue_args)


class KernelArgs:
    def __init__(self, lane_detector_args, plate_absent_db_args, track_react_endpoint, event_consumer,
                 counters, ms, raas, speed_rule_args, gpsd, im_proc):
        """
        :type lane_detector_args: LaneDetectorArgs
        :type plate_absent_db_args: PlateAbsentDbArgs
        :type track_react_endpoint: str
        :type event_consumer: EventConsumerArgs
        :type ms: MediaStoreArgs | None
        :type raas: RaasArgs | None
        :type speed_rule_args: SpeedRulesArgs | None
        :type gpsd: GpsdArgs
        :type im_proc: ImProcArgs
        """
        self.lane_detector = lane_detector_args
        self.plate_absent_db = plate_absent_db_args

        self.track_react_endpoint = track_react_endpoint
        self.event_consumer = event_consumer

        self.counters = counters

        self.ms = ms
        self.raas = raas
        self.gpsd = gpsd

        self.speed_rule_args = speed_rule_args

        self.im_proc = im_proc

    @staticmethod
    def deserialize(main_conf_jnode, log_level):
        """
        :type main_conf_jnode: dict
        :type log_level: str
        :rtype: KernelArgs
        """
        # define stat-counters-args
        monitoring_jnode = main_conf_jnode.get('Monitoring')
        if not monitoring_jnode:
            raise ValueError('main.conf#/Monitoring')

        counters = CountersArgs.deserialize(monitoring_jnode)

        # define def-footer-mapping
        base_conf_filepath = main_conf_jnode['base']
        with open(base_conf_filepath) as base_conf_jnode:
            base_conf_jnode = ujson.load(base_conf_jnode)

        tz_name = base_conf_jnode['rw']['tz']

        post_name = edge.pipeline_helpers.get_post_guid_name()[1]
        def_footer_mapping = {
            'tz-name': tz_name,
            'post-name': post_name,
            'geolocation': '',
            'plate': '',
            'speed': '-',
        }

        # define others
        lane_detector_args = LaneDetectorArgs.deserialize(main_conf_jnode.get('road-mask', {}))
        plate_absent_db_args = PlateAbsentDbArgs.deserialize(main_conf_jnode.get('plate-absent-db'))
        track_react_endpoint = main_conf_jnode['FileApi']['AddTrackReact']

        # define gpsd-args
        gpsd = GpsdArgs.deserialzie(main_conf_jnode.get('geolocation'))

        # define event-consumer-args
        event_consumer = EventConsumerArgs.deserialize(
            main_conf_jnode['fact-event'], main_conf_jnode['store'])

        event_consumer.jpeg_queue_args.log_level = log_level

        images_jnode = main_conf_jnode['fact-event']['images']
        im_proc = ImProcArgs.deserialize(images_jnode, def_footer_mapping)

        # define mediastore-args
        ms_jnode = main_conf_jnode.get('mediastore')
        ms = MediaStoreArgs.deserialize(ms_jnode) if ms_jnode else None

        # define raas-args
        raas_jnode = main_conf_jnode.get('raas')
        raas = RaasArgs.deserialize(raas_jnode) if raas_jnode else None

        # define speed-rules-args
        speed_rules_args = None
        rules_jnode = main_conf_jnode.get('rules')
        if rules_jnode:
            speed_violation_jnode = rules_jnode.get('speed-violation')
            if speed_violation_jnode:
                speed_rules_args = SpeedRulesArgs.deserialize(speed_violation_jnode)

        kernel_args = KernelArgs(
            lane_detector_args, plate_absent_db_args, track_react_endpoint, event_consumer,
            counters, ms, raas, speed_rules_args,
            gpsd, im_proc)

        return kernel_args


class SpeedRulesArgs:
    def __init__(self, lane_to_limits_map):
        """
        :type lane_to_limits_map: dict[int, (int, int)]
        """
        self._lane_to_limits_map = lane_to_limits_map

    @staticmethod
    def deserialize(speed_violation_jnode):
        """
        :type speed_violation_jnode: dict
        :rtype: SpeedRulesArgs
        """
        lane_to_limit_map = {}

        for rule_jnode in speed_violation_jnode:
            speed_limit = rule_jnode['speed-limit']
            if not (0 <= speed_limit <= 250):
                raise ValueError("main.conf#/rules/speed-violation/speed-limit")

            max_overspeed = rule_jnode['max-overspeed']
            if not (0 <= max_overspeed <= 250):
                raise ValueError("main.conf#/rules/speed-violation/max-overspeed")

            lane_list = rule_jnode.get('lane')
            if lane_list is None:
                lane_to_limit_map[0] = (speed_limit, max_overspeed)

            else:
                for lane in lane_list:
                    if not (1 <= lane <= 64):
                        raise ValueError("main.conf#/rules/speed-violation/lane")

                    lane_to_limit_map[lane] = (speed_limit, max_overspeed)

        return SpeedRulesArgs(lane_to_limit_map)

    def get_speed_enforcement(self, lane):
        """
        :type lane: int
        :rtype: (int, int) | None
        """
        speed_overspeed = self._lane_to_limits_map.get(lane)
        if speed_overspeed is None:
            speed_overspeed = self._lane_to_limits_map.get(0)

        return speed_overspeed


class ImFooterColArgs:
    def __init__(self, width, align, template):
        """
        :type width: int
        :type align:  int
        :type template: str
        """
        self.width = width
        self.align = align
        self.template = template

    def eval(self, mapping):
        """
        :type mapping: dict
        :rtype: ImFooterColArgs
        """
        value = FormatTemplate(self.template).substitute(mapping)
        return ImFooterColArgs(self.width, self.align, value)

    @staticmethod
    def deserialize(col_jnode):
        """
        :type col_jnode: dict
        :rtype: ImFooterColArgs
        :raise: ArgsError
        """
        width = col_jnode['width']

        raw_align = col_jnode.get('align')
        if raw_align == 'left':
            align = -1
        elif raw_align == 'center':
            align = 0
        elif raw_align == 'right':
            align = +1
        else:
            eventlog.log_debug('fail: param #/cols/x/align has invalid value')
            raise ArgsError

        template = col_jnode['format']

        return ImFooterColArgs(width, align, template)


class ImFooterRowArgs:
    def __init__(self, enabled, cols):
        """
        :type enabled: bool
        :type cols: list of ImFooterColArgs
        """
        self.enabled = enabled
        self.cols = cols

    def eval(self, mapping):
        """
        :type mapping: dict
        :rtype: ImFooterRowArgs
        """
        cols = map(lambda col: col.eval(mapping), self.cols)
        return ImFooterRowArgs(self.enabled, cols)

    @staticmethod
    def deserialize(row_jnode):
        """
        :type row_jnode: dict
        :rtype: ImFooterRowArgs
        :raise: ArgsError
        """
        enabled = row_jnode['enabled']
        if enabled:
            cols_jnode = row_jnode['cols']
            cols = map(ImFooterColArgs.deserialize, cols_jnode)
        else:
            cols = []

        return ImFooterRowArgs(enabled, cols)


class ImFooterArgs:
    def __init__(self, height, font_size, rows, def_mapping):
        """
        :type height: int
        :param font_size: Font size, pixels
        :type font_size: int
        :type rows: list of ImFooterRowArgs
        :type def_mapping: dict
        """
        self.height = height
        self.font_size = font_size
        self.rows = rows

        self.def_mapping = def_mapping

    def eval(self, ts, fact):
        """
        :type ts: int
        :type fact: Fact
        :rtype: ImFooterArgs
        """
        pholders = []
        for row in self.rows:
            for col in row.cols:
                pholders.extend(FormatTemplate.get_pholders(col.template))

        mapping = {}
        for pholder in pholders:
            if pholder == 'datetime':
                tz = pytz.timezone(self.def_mapping['tz-name'])
                dt = datetime.datetime.fromtimestamp(ts / 1000000.0, tz)
                mapping[pholder] = dt.strftime('%d.%m.%Y %H:%M:%S')

            elif pholder == 'geolocation':
                mapping[pholder] = fact.geodata.serialize2() if fact.geodata else self.def_mapping.get(pholder, '')

            elif pholder == 'post-name':
                mapping[pholder] = self.def_mapping.get(pholder, '')

            elif pholder == 'speed':
                mapping[pholder] = fact.speed or self.def_mapping.get(pholder, '')

            elif pholder == 'plate':
                if fact.plate:
                    mapping[pholder] = fact.plate.text or self.def_mapping.get(pholder, '')
                else:
                    # for facts only by radar
                    mapping[pholder] = self.def_mapping.get(pholder, '')

            else:
                raise NotImplemented

        rows = map(lambda r: r.eval(mapping), self.rows)
        return ImFooterArgs(self.height, self.font_size, rows, self.def_mapping)

    @staticmethod
    def deserialize(footer_jnode, def_mapping):
        """
        :type footer_jnode: dict
        :type def_mapping: dict
        :rtype: ImFooterArgs
        :raise: ArgsError
        """
        font_size = footer_jnode['font-height']

        rows_jnode = footer_jnode.get('rows')
        if not rows_jnode:
            eventlog.log_debug('fail: param #/rows has invalid value')
            raise ArgsError

        rows = filter(lambda row: row.enabled, map(ImFooterRowArgs.deserialize, rows_jnode))

        if len(rows) > 0:
            height = int((1.4 * font_size) * len(rows) + 0.4 * font_size)  # 120%
        else:
            # no active rows in footer
            height = 0

        return ImFooterArgs(height, font_size, filter(lambda row: row.enabled, rows), def_mapping)


class BodyImAnnoArgs:
    def __init__(self, plate=False):
        """
        :type plate: bool
        """
        self.plate = plate


class BodyImArgs:
    def __init__(self, width, footer, anno):
        """
        :type width: int
        :type footer: ImFooterArgs
        :type anno: BodyImAnnoArgs
        """
        self.width = width
        self.wh_ratio = 4.0 / 3.0

        self.footer = footer
        self.anno = anno

    @staticmethod
    def deserialize(profile_jnode, def_footer_mapping):
        """
        :type profile_jnode: dict
        :type def_footer_mapping: dict
        :rtype: BodyImArgs
        :raise: ArgsError
        """

        # deserialize image-width
        image_w = profile_jnode['width']

        # deserialize footer-args
        footer_jnode = profile_jnode.get('footer')
        if not footer_jnode:
            eventlog.log_debug('fail: param #/footer has invalid value')
            raise ArgsError

        footer = ImFooterArgs.deserialize(footer_jnode, def_footer_mapping)

        return BodyImArgs(image_w, footer, BodyImAnnoArgs(plate=True))


class SceneImArgs:
    def __init__(self, width):
        """
        :type width: int
        """
        self.width = width

    @staticmethod
    def deserialize(profile_jnode):
        """
        :type profile_jnode: dict
        :rtype: SceneImArgs
        :raise: ArgsError
        """
        image_w = profile_jnode['width']
        return SceneImArgs(image_w)


class SceneAltImAnnoArgs:
    def __init__(self, track=False, body=False, plate=False):
        """
        :type track: bool
        :type body: bool
        :type plate: bool
        """
        self.track = track
        self.body = body
        self.plate = plate

    @staticmethod
    def deserialize(anno_jnode):
        """
        :type anno_jnode: dict
        :rtype: SceneAltImAnnoArgs
        :raise: ArgsError
        """
        vehicle_jnode = anno_jnode.get('vehicle')
        if not vehicle_jnode:
            return SceneAltImAnnoArgs()

        track = vehicle_jnode.get('track', False)
        body = vehicle_jnode.get('body', False)
        plate = vehicle_jnode.get('plate', False)

        return SceneAltImAnnoArgs(track, body, plate)


class SceneAltImArgs:
    def __init__(self, width, footer, anno):
        """
        :type width: int
        :type footer: ImFooterArgs
        :type anno: SceneAltImAnnoArgs
        """
        self.width = width
        self.footer = footer
        self.anno = anno

    @staticmethod
    def deserialize(profile_jnode, def_footer_mapping):
        """
        :type profile_jnode: dict
        :type def_footer_mapping: dict
        :rtype: SceneAltImArgs
        :raise: ArgsError
        """

        # deserialize image-width
        image_w = profile_jnode['width']

        # deserialize footer-args
        footer_jnode = profile_jnode.get('footer')
        if not footer_jnode:
            eventlog.log_debug('fail: param #/footer has invalid value')
            raise ArgsError

        footer = ImFooterArgs.deserialize(footer_jnode, def_footer_mapping)

        # deserialize annotations
        anno_jnode = profile_jnode.get('annotations')
        anno = SceneAltImAnnoArgs.deserialize(anno_jnode) if anno_jnode else SceneAltImAnnoArgs()

        return SceneAltImArgs(image_w, footer, anno)


class ImProcArgs:
    def __init__(self, body, scene, scene_alt, scene_footer, body_footer, plate_w, plate_wh_ratio, body_thumb_w):
        """
        :type body: BodyImArgs
        :type scene: SceneImArgs
        :type scene_alt: SceneAltImArgs

        :type scene_footer: dict
        :type body_footer: dict
        """
        self.scene_footer = scene_footer

        self.body_footer = body_footer

        self.plate_w = plate_w
        self.plate_wh_ratio = plate_wh_ratio
        self.body = body
        self.body_thumb_w = body_thumb_w
        self.scene = scene
        self.scene_alt = scene_alt

    @staticmethod
    def deserialize(args_jnode, def_footer_mapping):
        """
        :type args_jnode: dict
        :type def_footer_mapping: dict
        :rtype: ImProcArgs
        :raise: ArgsError
        """

        # deserialize profile "scene.measurement"
        scene_jnode = args_jnode.get('scene.measurement')
        if not scene_jnode:
            eventlog.log_debug('fail: param #/scene.measurement has invalid value')
            raise ArgsError()

        scene = SceneImArgs.deserialize(scene_jnode)

        # deserialize profile "scene.measurement.alt"
        scene_alt_jnode = args_jnode.get('scene.measurement.alt')
        if not scene_alt_jnode:
            eventlog.log_debug('fail: param #/scene.measurement.alt has invalid value')
            raise ArgsError()

        scene_alt = SceneAltImArgs.deserialize(scene_alt_jnode, def_footer_mapping)

        scene_footer_jnode = scene_alt_jnode.get('footer')
        if not scene_footer_jnode:
            raise ValueError('#/images/scene.measurement.alt/footer')

        # deserialize profile "vehicle.body"
        body_jnode = args_jnode.get('vehicle.body')
        if not body_jnode:
            eventlog.log_debug('fail: param #/vehicle.body has invalid value')
            raise ArgsError()

        body = BodyImArgs.deserialize(body_jnode, def_footer_mapping)

        body_footer_jnode = body_jnode.get('footer')
        if not body_footer_jnode:
            raise ValueError('#/images/vehicle.body/footer')

        plate_w = 134
        plate_wh_ratio = 67.0 / 15.0
        body_thumb_w = 220

        return ImProcArgs(body, scene, scene_alt,
                          scene_footer=scene_footer_jnode,
                          body_footer=body_footer_jnode,
                          plate_w=plate_w,
                          plate_wh_ratio=plate_wh_ratio,
                          body_thumb_w=body_thumb_w)
