import _models

_FAKE_PLATE_SIZE = 0.01


def _check_rect_dist_from_img_borders(plate_rect):
    """
    :type plate_rect: _models.PlateRect
    :rtype: bool
    """
    plate_width = plate_rect.right - plate_rect.left
    plate_height = plate_rect.bottom - plate_rect.top

    return (plate_rect.top > plate_height and
            plate_rect.bottom < (1.0 - 2 * plate_height) and
            plate_rect.left > plate_width and
            plate_rect.right < (1.0 - plate_width))


def _get_rect_center(plate_rect):
    """
    :type plate_rect: _models.PlateRect
    :rtype: (float, float)
    """
    return (plate_rect.right - plate_rect.left) / 2, (plate_rect.bottom - plate_rect.top / 2)


def _get_rect_dist_to_img_center(rect_center_x, rect_center_y):
    distance = (rect_center_x - 0.5) ** 2 + (rect_center_y - 0.5) ** 2
    return distance


def _check_next_lr_track_item_better(best_item, next_item):
    """
    :type best_item: _models.LrTrackItem
    :type next_item: _models.LrTrackItem
    :rtype: bool
    """
    # check vehicle's plate is fully visible on the image.
    best_visible = best_item.plate_rect.check_visible()
    next_visible = next_item.plate_rect.check_visible()

    if not best_visible and next_visible:
        return True

    if best_visible and not next_visible:
        return False

    # check plate's recognition is better.
    if next_item.plate_weight > best_item.plate_weight:
        return True

    elif next_item.plate_weight < best_item.plate_weight:
        return False

    # Check vehicle's body is more visible.
    best_border_ok = _check_rect_dist_from_img_borders(best_item.plate_rect)
    next_border_ok = _check_rect_dist_from_img_borders(next_item.plate_rect)

    if not best_border_ok and next_border_ok:
        return True

    if best_border_ok and not next_border_ok:
        return False

    # check plate's resolution is better.
    best_plate_height = best_item.plate_rect.bottom - best_item.plate_rect.top
    next_plate_height = next_item.plate_rect.bottom - next_item.plate_rect.top
    if next_plate_height > best_plate_height:
        return True

    return False


def get_best_lr_track_item(track_items, search_image_list):
    """
    :type track_items: list of _models.LrTrackItem
    :type search_image_list: list of int
    :rtype: _models.LrTrackItem | None
    """
    lr_track_item_count = len(track_items)
    if not lr_track_item_count:
        raise ValueError

    best_track_item = None
    for next_track_item in track_items:
        if next_track_item.ts not in search_image_list:
            continue

        if not best_track_item:
            best_track_item = next_track_item
            continue

        if _check_next_lr_track_item_better(best_track_item, next_track_item):
            best_track_item = next_track_item

    if not best_track_item:
        return None

    if best_track_item.plate_rect.left < 0.0:
        best_track_item.plate_rect.left = 0.0

    if best_track_item.plate_rect.top < 0.0:
        best_track_item.plate_rect.top = 0.0

    if best_track_item.plate_rect.right > 1.0:
        best_track_item.plate_rect.right = 1.0

    if best_track_item.plate_rect.bottom > 1.0:
        best_track_item.plate_rect.bottom = 1.0

    return best_track_item


def get_best_ra_ts_and_rect(track_items, search_image_list):
    """
    :type track_items: list of _models.RaTrackItem
    :type search_image_list: list of int
    :rtype: (int, _models.PlateRect)| None
    """
    best_line_y = 0.5
    best_item_dist = 2.0
    best_item_index = None
    for index, item in enumerate(track_items):
        dist = abs(item.y - best_line_y)
        if dist < best_item_dist:
            if item.ts in search_image_list:
                # noinspection PyChainedComparisons
                if item.x < 1.0 - _FAKE_PLATE_SIZE and item.x > _FAKE_PLATE_SIZE \
                        and item.y < 1.0 - _FAKE_PLATE_SIZE and item.y > _FAKE_PLATE_SIZE:
                    best_item_dist = dist
                    best_item_index = index

    if best_item_index is None:
        return None

    radar_point = track_items[best_item_index]

    plate_rect = _models.PlateRect(
        radar_point.x - _FAKE_PLATE_SIZE, radar_point.y - _FAKE_PLATE_SIZE,
        radar_point.x + _FAKE_PLATE_SIZE, radar_point.y + _FAKE_PLATE_SIZE)

    return track_items[best_item_index].ts, plate_rect
