import time
from time import time as get_systime_sec
from os import rename

import edge.eventlog as eventlog

from _args import CountersArgs
from _models import TrTrack
from ujson import dump as ujson_dump


def _get_list_min_max_med(float_list):
    """
    :type float_list: list of float
    :rtype (float, float, float)
    """
    min_value = 0.0
    max_value = 0.0
    med_value = 0.0

    float_list_len = len(float_list)
    if float_list_len > 0:
        float_list.sort()
        min_value = float_list[0]
        max_value = float_list[-1]
        med_value = float_list[float_list_len / 2]

    return min_value, max_value, med_value


class ReactProducerCounters:
    def __init__(self):
        self.lrra_track_count = long(0)
        self.lr_track_count = long(0)
        self.ra_track_count = long(0)
        self.track_count = long(0)

        self.near_plate_w_list = []
        self.near_plate_h_list = []
        self.far_plate_w_list = []
        self.far_plate_h_list = []

    def reset(self):
        self.near_plate_w_list[:] = []
        self.near_plate_h_list[:] = []
        self.far_plate_w_list[:] = []
        self.far_plate_h_list[:] = []

    def inc_plate_rect(self, catch_plate_rect, lost_plate_rect):
        """
        :type catch_plate_rect: _models.PlateRect
        :type lost_plate_rect: _models.PlateRect
        """
        catch_plate_w = catch_plate_rect.right - catch_plate_rect.left
        catch_plate_h = catch_plate_rect.bottom - catch_plate_rect.top

        lost_plate_width = lost_plate_rect.right - lost_plate_rect.left
        lost_plate_height = lost_plate_rect.bottom - lost_plate_rect.top

        if catch_plate_w * catch_plate_h >= lost_plate_width * lost_plate_height:
            self.near_plate_w_list.append(catch_plate_w)
            self.near_plate_h_list.append(catch_plate_h)

            self.far_plate_w_list.append(lost_plate_width)
            self.far_plate_h_list.append(lost_plate_height)

        else:
            self.near_plate_w_list.append(lost_plate_width)
            self.near_plate_h_list.append(lost_plate_height)

            self.far_plate_w_list.append(catch_plate_w)
            self.far_plate_h_list.append(catch_plate_h)

    def inc_track(self, tr_track):
        """
        :type tr_track: TrTrack
        """
        self.track_count += 1

        if bool(tr_track.lr_track) and bool(tr_track.lr_track):
            self.lrra_track_count += 1

        if bool(tr_track.lr_track):
            self.lr_track_count += 1

        if bool(tr_track.lr_track):
            self.ra_track_count += 1


class EventConsumerCounters:
    def __init__(self):
        self.fact_count = long(0)

    def inc_fact(self):
        self.fact_count += 1


class GeodataCounters:
    """
    :type mts: long | None
    :type tpv_ts: long | None
    :type tpv_lat: float | None
    :type tpv_lon: float | None
    :type tpv_alt: float | None
    """
    def __init__(self):
        self.mts = None

        self.tpv_ts = None
        self.tpv_lat = None
        self.tpv_lon = None
        self.tpv_alt = None

    def inc(self, tpv_ts, tpv_lat, tpv_lon, tpv_alt):
        """
        :type tpv_ts: long | None
        :type tpv_lat: float | None
        :type tpv_lon: float | None
        :type tpv_alt: float | None
        """
        self.mts = int(get_systime_sec())

        self.tpv_ts = tpv_ts
        self.tpv_lat = tpv_lat
        self.tpv_lon = tpv_lon
        self.tpv_alt = tpv_alt


class GpsdCounters:
    """
    :type mode: int
    :type mode_mts: int
    :type store_count: long
    """

    MODE_0D_FIX = 0
    MODE_1D_FIX = 1
    MODE_2D_FIX = 2
    MODE_3D_FIX = 3

    def __init__(self):
        self.mode = 0
        self.mode_mts = 0

        self.store_count = long()
        self.geodata = GeodataCounters()

    def inc_mode(self, mode):
        if self.mode != mode:
            self.mode = mode
            self.mode_mts = int(get_systime_sec())

    def inc_store_count(self, count):
        self.store_count += count


class StatCounters:
    def __init__(self, twisted_reactor, args, stat_gpsd):
        """
        :type args: CountersArgs
        :type stat_gpsd: bool
        """
        self.gpsd = GpsdCounters() if stat_gpsd else None

        self.react = ReactProducerCounters()
        self.event = EventConsumerCounters()

        self._stat_abspath = args.stat_filepath
        self._stat_tmp_abspath = args.stat_filepath + '.tmp'
        self._stat_tout_sec = args.stat_tout_sec

        if args.stat_tout_sec > 0:
            twisted_reactor.callLater(args.stat_tout_sec, self._log_counters, twisted_reactor)

    def _log_counters(self, twisted_reactor):
        self._report_counters()
        self._reset()

        twisted_reactor.callLater(self._stat_tout_sec, self._log_counters, twisted_reactor)

    def _reset(self):
        self.react.reset()

    def _report_counters(self):
        sys_utc_time = time.time()

        npw_min, npw_max, npw_med = _get_list_min_max_med(self.react.near_plate_w_list)
        nph_min, nph_max, nph_med = _get_list_min_max_med(self.react.near_plate_h_list)
        fpw_min, fpw_max, fpw_med = _get_list_min_max_med(self.react.far_plate_w_list)
        fph_min, fph_max, fph_med = _get_list_min_max_med(self.react.far_plate_h_list)

        stat = {
            "stat": {
                "mtime": [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                "tout-sec": self._stat_tout_sec
            },
            "react": {
                "count": {
                    "lr": self.react.lr_track_count,
                    "ra": self.react.ra_track_count,
                    "lrra": self.react.lrra_track_count,
                    "total": self.react.track_count
                },
                "plate": {
                    "near-w": [npw_min, npw_med, npw_max],
                    "near-h": [nph_min, nph_med, nph_max],
                    "far-w": [fpw_min, fpw_med, fpw_max],
                    "far-h": [fph_min, fph_med, fph_max]
                }
            },
            "event": {
                "count": self.event.fact_count
            }
        }

        if self.gpsd is not None:
            stat['gpsd'] = {
                "last-geodata-count": self.gpsd.store_count
            }

            if self.gpsd.mode_mts != 0:
                stat['gpsd']['fix'] = {
                    "mode": self.gpsd.mode,
                    "mtime": [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(int(self.gpsd.mode_mts))),
                              int(self.gpsd.mode_mts)]
                }

            if self.gpsd.geodata.mts is not None:
                stat['gpsd']['last-geodata'] = {
                    "mtime": [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(int(self.gpsd.geodata.mts))),
                              int(self.gpsd.geodata.mts)]
                }

                # add time
                if self.gpsd.geodata.tpv_ts is not None:
                    stat['gpsd']['last-geodata']['time'] = [
                        time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(int(self.gpsd.geodata.tpv_ts))),
                        int(self.gpsd.geodata.tpv_ts)
                    ]

                # add latitude/longitude
                if None not in [self.gpsd.geodata.tpv_lat, self.gpsd.geodata.tpv_lon]:
                    # Latitude in degrees: +/- signifies North/South.
                    if self.gpsd.geodata.tpv_lat >= 0.0:
                        lat_long = "N%.6f" % +self.gpsd.geodata.tpv_lat
                    else:
                        lat_long = "S%.6f" % -self.gpsd.geodata.tpv_lat

                    # Longitude in degrees: +/- signifies East/West.
                    if self.gpsd.geodata.tpv_lon >= 0:
                        lat_long += "E%.6f" % +self.gpsd.geodata.tpv_lon
                    else:
                        lat_long += "W%.6f" % -self.gpsd.geodata.tpv_lon

                    stat['gpsd']['last-geodata']['lat-long'] = [
                        self.gpsd.geodata.tpv_lat,
                        self.gpsd.geodata.tpv_lon,
                        lat_long
                    ]

                # add altitude
                if self.gpsd.geodata.tpv_alt is not None:
                    stat['gpsd']['last-geodata']['alt'] = self.gpsd.geodata.tpv_alt

        # noinspection PyBroadException
        try:
            with open(self._stat_tmp_abspath, 'w') as json_file:
                # Use ensure_ascii=false to output UTF-8.
                ujson_dump(stat, json_file, ensure_ascii=False)

            rename(self._stat_tmp_abspath, self._stat_abspath)

        except:
            eventlog.log_debug_last_exc()
