import edge.eventlog as eventlog
from edge.queue import JsonFileQueue, ImageFileQueue

from _args import EventConsumerArgs
from _counters import EventConsumerCounters
from _models import Fact, FactImages


class IEventConsumer:
    def __init__(self):
        pass

    def consume_fact(self, fact, images):
        """
        :type fact: Fact
        :type images: FactImages
        """
        raise NotImplemented


class EventConsumer(IEventConsumer):

    _TARGET_CODEC = 'JPEG'

    def __init__(self, counters, json_q_list, jpeg_q, require_plate):
        """
        :type counters: EventConsumerCounters
        :type json_q_list: list[JsonFileQueue]
        :type jpeg_q: ImageFileQueue
        :type require_plate: bool
        """
        IEventConsumer.__init__(self)

        self._counters = counters

        self._json_q_list = json_q_list
        self._jpeg_q = jpeg_q

        self._require_plate = require_plate

    @staticmethod
    def create(twisted_reactor, counters, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTime
        :type counters: EventConsumerCounters
        :type args: EventConsumerArgs
        """
        json_q_list = map(lambda json_q_args: JsonFileQueue(json_q_args), args.json_queue_args_list)

        jpeg_q = ImageFileQueue(
            twisted_reactor,
            EventConsumer._TARGET_CODEC,
            args.jpeg_queue_args.endpoint,
            args.jpeg_queue_args.min_file_count,
            args.jpeg_queue_args.max_file_count,
            args.jpeg_queue_args.log_level)

        return EventConsumer(counters, json_q_list, jpeg_q, args.require_plate)

    def consume_fact(self, fact, images):
        """
        :type fact: Fact
        :type images: FactImages
        """
        if self._require_plate and not fact.plate:
            eventlog.log_debug('skip fact wout plate')
            return

        # noinspection PyBroadException
        try:
            self._consume_fact(fact, images)
        except:
            eventlog.log_debug_last_exc()

    def _consume_fact(self, fact, images):
        """
        :type fact: Fact
        :type images: FactImages
        """
        def mk_image_jnode(blob, meta, name):
            """
            :type blob: str | None
            :type meta: dict | None
            :type name: str
            :rtype: dict
            """
            image_jnode = {
                'name': name
            }

            if blob:
                image_jnode['blob'] = blob

            if meta:
                image_jnode['meta'] = meta

            return image_jnode

        images_jnode = []
        if images.plate_blob:
            images_jnode.append(
                mk_image_jnode(
                    self._jpeg_q.add_blob_image(images.plate_blob).url,
                    images.plate_meta, 'plate'))

        if images.body_thumb_blob:
            images_jnode.append(
                mk_image_jnode(
                    self._jpeg_q.add_blob_image(images.body_thumb_blob).url,
                    images.body_thumb_meta, 'body.thumb'))

        if images.body_blob:
            images_jnode.append(
                mk_image_jnode(self._jpeg_q.add_blob_image(images.body_blob).url, images.body_meta, 'body'))

        if images.scene_alt_blob:
            images_jnode.append(
                mk_image_jnode(
                    self._jpeg_q.add_blob_image(images.scene_alt_blob).url,
                    images.scene_alt_meta, 'scene.alt'))

        if images.scene_blob:
            images_jnode.append(
                mk_image_jnode(
                    self._jpeg_q.add_blob_image(images.scene_blob).url,
                    images.scene_meta, 'scene'))

        fact_jnode = Fact.serialize(fact, images.ts, images_jnode)
        map(lambda json_q: json_q.add_jnode(fact_jnode), self._json_q_list)
        self._counters.inc_fact()
