import ctypes
import cStringIO
from PIL import Image, ImageDraw, ImageFont

import edge.eventlog as eventlog
import edge_rawutil.rawutil as rawutil

from _args import ImProcArgs, ImFooterArgs
import _models


class ImError (RuntimeError):
    pass


class _RgbaImBlob:
    def __init__(self, ts, w, h, dh):
        """
        :type ts: int
        :type w: int
        :type h: int
        :type dh: int
        """
        self.ts = ts

        self.w = w
        self.h = h
        self.dh = dh
        self.wh = w, h

        self.stride = 4 * w
        self.buffer = ctypes.create_string_buffer(self.stride * (h + dh))

    def set_im_size(self, ts, w, h, dh):
        """
        :type ts: int
        :type w: int
        :type h: int
        :type dh: int
        """
        self.ts = ts

        self.w = w
        self.h = h
        self.dh = dh
        self.wh = w, h

        self.stride = 4 * w

    def check_buffer_size(self, w, h, dh):
        """
        :type w: int
        :type h: int
        :type dh: int
        :rtype: bool
        """
        return 4 * w * (h + dh) <= len(self.buffer)


class _ImTrans:

    #
    # (x', y') - scene-im point; (x'', y'') - recrop-image point.
    #
    # |x''|   | a  b  x |   |x'|
    # |y''| = | c  d  y | * |y'|
    # |1  |   | 0  0  1 |   |1 |
    #
    #
    # |x''|   | Ix Jx Ox |   |x'|
    # |y''| = | Iy Jy Oy | * |y'|
    # |1  |   | 0  0  1  |   |1 |
    #

    def __init__(self, ix=1.0, iy=0.0, jx=0.0, jy=1.0, ox=0.0, oy=0.0):
        """
        :type ix: float
        :type iy: float
        :type jx: float
        :type jy: float
        :type ox: float
        :type oy: float
        """
        self._ix = ix
        self._iy = iy
        self._jx = jx
        self._jy = jy
        self._ox = ox
        self._oy = oy

    def map_rect(self, ltrb):
        """
        :type ltrb: list of float
        :rtype: list of float
        """
        l, t, r, b = ltrb

        l2 = l * self._ix + t * self._jx + self._ox
        t2 = l * self._iy + t * self._jy + self._oy

        r2 = r * self._ix + b * self._jx + self._ox
        b2 = r * self._iy + b * self._jy + self._oy

        return [l2, t2, r2, b2]

    def serialize(self):
        """
        :rtype: dict
        """
        return {
            'a': self._ix,
            'b': self._jx,
            'c': self._iy,
            'd': self._jy,
            'x': self._ox,
            'y': self._oy
        }


class _LtrbWh:
    def __init__(self, l, t, r, b, w, h):
        """
        :type l: float
        :type t: float
        :type r: float
        :type b: float
        :type w: float
        :type h: float
        """
        self.l = l
        self.t = t
        self.r = r
        self.b = b
        self.w = w
        self.h = h

    @staticmethod
    def deserialize(rect, (image_w, image_h)):
        """
        :type rect: _models.PlateRect
        :type image_w: int
        :type image_h: int
        :rtype: _LtrbWh
        :raises: ValueError
        """
        def norm(max_coord, coord):
            """
            :type coord: float
            :type max_coord: int
            :rtype: float
            """
            if coord < 0.0:
                return 0.0

            elif coord > max_coord:
                return float(max_coord)

            else:
                return coord

        l0 = image_w * rect.left
        t0 = image_h * rect.top
        r0 = image_w * rect.right
        b0 = image_h * rect.bottom

        # tweak plate rect for better view
        symbol_height_cm = 8
        symbol_margin_cm = 1.5

        h0 = b0 - t0
        delta = symbol_margin_cm * h0 / symbol_height_cm
        l0 = l0 - 3 * delta
        t0 = t0 - delta
        r0 = r0 + 2 * delta
        b0 = b0 + delta

        # normalize rect to fit image
        l = norm(image_w, l0)
        t = norm(image_h, t0)
        r = norm(image_w, r0)
        b = norm(image_h, b0)

        if l < r and t < b:
            return _LtrbWh(l, t, r, b, w=r - l, h=b - t)

        raise ValueError('l=%s, r=%s, t=%s, b=%s' % (l, t, r, b))


def _mk_jpeg_str(source_image):
    """
    :type source_image: PIL.Image.Image
    :rtype: str
    """
    jpeg_stream = cStringIO.StringIO()
    source_image.save(jpeg_stream, format="JPEG")
    jpeg_bytes = jpeg_stream.getvalue()
    jpeg_stream.close()

    return jpeg_bytes


def _resize_im(scene_blob, (target_w, target_h)):
    """
    :type scene_blob: _RgbaImBlob
    :type target_w: int | float
    :type target_h: int | float
    :rtype: (PIL.Image.Image, _ImTrans)
    """
    #
    # (x', y') - scene-im point; (x'', y'') - recrop-image point.
    #
    # |x''|   | Ix Jx Ox |   |x'|
    # |y''| = | Iy Jy Oy | * |y'|
    # |1  |   | 0  0  1  |   |1 |
    #
    #     scale           move              trans
    # | Ix  0   0 |   | 1   0   0  |   | Ix  0   0  |
    # | 0   Jy  0 | * | 0   1   0  | = | 0   Jy  0  |
    # | 0   0   1 |   | 0   0   1  |   | 0   0   1  |
    #

    im_trans = _ImTrans(
        ix=target_w / float(scene_blob.w),
        jy=target_h / float(scene_blob.h))

    with Image.frombuffer(
            'RGBA', scene_blob.wh, scene_blob.buffer,
            'raw', 'RGBA', scene_blob.stride, 1) as crop_im:

        resize_im = crop_im.resize((int(target_w), int(target_h)), Image.NEAREST)
        return resize_im, im_trans


def _crop_resize_im(source_blob, crop_ltrbwh, (target_w, target_h)):
    """
    :type source_blob: _RgbaImBlob
    :type crop_ltrbwh: _LtrbWh
    :type target_w: int | float
    :type target_h: int | float
    :rtype: (PIL.Image.Image, _ImTrans)
    """

    #
    # (x', y') - scene-im point; (x'', y'') - recrop-image point.
    #
    # |x''|   | Ix Jx Ox |   |x'|
    # |y''| = | Iy Jy Oy | * |y'|
    # |1  |   | 0  0  1  |   |1 |
    #
    #     scale           move              trans
    # | Ix  0   0 |   | 1   0   Ox |   | Ix  0   Ox*Ix |
    # | 0   Jy  0 | * | 0   1   Oy | = | 0   Jy  Oy*Jy |
    # | 0   0   1 |   | 0   0   1  |   | 0   0   1     |
    #

    ix = target_w / float(crop_ltrbwh.w)
    jy = target_h / float(crop_ltrbwh.h)
    im_trans = _ImTrans(ix=ix, jy=jy, ox=-crop_ltrbwh.l * ix, oy=-crop_ltrbwh.t * jy)

    scene_mv = memoryview(source_blob.buffer)
    crop_offset = 4 * (crop_ltrbwh.t * source_blob.w + crop_ltrbwh.l)

    with Image.frombuffer(
            'RGBA', (crop_ltrbwh.w, crop_ltrbwh.h), scene_mv[crop_offset:],
            'raw', 'RGBA', source_blob.stride, 1) as crop_im:

        resize_im = crop_im.resize((int(target_w), int(target_h)), Image.NEAREST)
        return resize_im, im_trans


def _draw_rect(pil_image, rect_ltrbwh, rect_trans=None):
    """
    :type pil_image: PIL.Image.Image
    :type rect_ltrbwh: _LtrbWh
    :type rect_trans: _ImTrans | None
    """
    ltrb = [rect_ltrbwh.l, rect_ltrbwh.t, rect_ltrbwh.r, rect_ltrbwh.b]
    if rect_trans:
        ltrb = rect_trans.map_rect(ltrb)

    ImageDraw.Draw(pil_image).rectangle(ltrb, outline='red')


def _draw_footer(pil_image, footer_args, font, font_size):
    """
    :type pil_image: PIL.Image.Image
    :type footer_args: ImFooterArgs
    :type font: PIL.ImageFont.FreeTypeFont
    """
    draw = ImageDraw.Draw(pil_image)
    # font_size = font.size

    y_delta = 0.4 * font_size  # top-margin
    start_y = int(y_delta / 2)

    leter_width = font.getsize('M')[0]

    for row in footer_args.rows:
        start_x = leter_width / 2  # left-margin = 0.5em
        for column in row.cols:
            column_width = column.width
            column_text = column.template
            column_align = column.align

            if column_align == -1:
                x_delta = 0

            elif column_align == 0:
                text_width, text_height = font.getsize(column_text)
                x_delta = (column_width - text_width) / 2 - leter_width / 2

            elif column_align == 1:
                text_width, text_height = font.getsize(column_text)
                x_delta = column_width - text_width - leter_width

            else:
                raise NotImplemented

            draw.text((start_x + x_delta, start_y), column_text, font=font)
            start_x = start_x + column_width

        start_y = start_y + font_size + y_delta


def _draw_track(pil_image, plate_ltrbwh, track, image_width, image_height):
    """
    :type plate_ltrbwh _LtrbWh
    :type pil_image: PIL.Image.Image
    :type track: _models.LrTrack
    """
    DIR_APROACHING = 11
    plate_trace = track.plate_trace

    for index in range(0, len(plate_trace) - 1):
        left_point = plate_trace[index]
        right_point = plate_trace[index + 1]

        l_image_x = (left_point.plate_rect.left + left_point.plate_rect.right) / 2.0 * image_width
        l_image_y = (left_point.plate_rect.top + left_point.plate_rect.bottom) / 2.0 * image_height

        r_image_x = (right_point.plate_rect.left + right_point.plate_rect.right) / 2.0 * image_width
        r_image_y = (right_point.plate_rect.top + right_point.plate_rect.bottom) / 2.0 * image_height

        left_inside_plate = plate_ltrbwh.t < l_image_y < plate_ltrbwh.b and plate_ltrbwh.l < l_image_x < plate_ltrbwh.r
        right_inside_plate = plate_ltrbwh.t < r_image_y < plate_ltrbwh.b and plate_ltrbwh.l < r_image_x < plate_ltrbwh.r

        both_ok = False
        if left_inside_plate and right_inside_plate:
            continue
        elif left_inside_plate and not right_inside_plate:
            # tweak left point
            k = (r_image_x - l_image_x) / (r_image_y - l_image_y)
            b = (l_image_x * r_image_y - r_image_x * l_image_y) / (r_image_y - l_image_y)
            # x = k * y + b
            if track.direction == DIR_APROACHING:
                # tweak down
                l_image_y = plate_ltrbwh.b + 0.5 * (plate_ltrbwh.b - plate_ltrbwh.t)
            else:
                # tweak up
                l_image_y = plate_ltrbwh.t - 0.5 * (plate_ltrbwh.b - plate_ltrbwh.t)

            l_image_x = k * l_image_y + b

        elif right_inside_plate and not left_inside_plate:
            # tweak right point
            k = (r_image_x - l_image_x) / (r_image_y - l_image_y)
            b = (l_image_x * r_image_y - r_image_x * l_image_y) / (r_image_y - l_image_y)
            # x = k * y + b
            if track.direction == DIR_APROACHING:
                # tweak up
                r_image_y = plate_ltrbwh.t - 0.5 * (plate_ltrbwh.b - plate_ltrbwh.t)
            else:
                # tweak down
                r_image_y = plate_ltrbwh.b + 0.5 * (plate_ltrbwh.b - plate_ltrbwh.t)

            r_image_x = k * r_image_y + b

        else:
            # both points OK
            both_ok = True

        _dir = 1 if track.direction == DIR_APROACHING else -1

        if ((r_image_y - l_image_y) * _dir > 0 and (left_inside_plate or right_inside_plate)) or both_ok:
            ImageDraw.Draw(pil_image).line([(l_image_x, l_image_y), (r_image_x, r_image_y)], fill='white')

            left_point = (l_image_x - 2, l_image_y - 2, l_image_x + 2, l_image_y + 2)
            ImageDraw.Draw(pil_image).rectangle(left_point, fill='red')

            right_point = (r_image_x - 2, r_image_y - 2, r_image_x + 2, r_image_y + 2)
            ImageDraw.Draw(pil_image).rectangle(right_point, fill='red')


class FactImProc:
    """
    :type _body_dh_blob: _RgbaImBlob | None
    :type _scene_dh_blob: _RgbaImBlob | None
    :type _scene_alt_dh_blob: _RgbaImBlob | None
    """

    _FONT_PATH = '/usr/share/fonts/truetype/ttf-dejavu/DejaVuSans.ttf'

    def __init__(self, args):
        """
        :type args: ImProcArgs
        """
        self._body_font = ImageFont.truetype(
            FactImProc._FONT_PATH, args.body.footer.font_size, encoding='unic')
        self._scene_alt_font = ImageFont.truetype(
            FactImProc._FONT_PATH, args.scene_alt.footer.font_size, encoding='unic')

        self._body_dh_blob = None
        self._scene_dh_blob = None
        self._scene_alt_dh_blob = None

        self._args = args

        self._rawutil = rawutil.RawUtil.create()

    def destroy(self):
        self._rawutil.destroy()

    def _mk_scene_dh_blob(self, im_ref, dh):
        """
        :type im_ref: _models.LrImageRef
        :type dh: int
        :rtype: _RgbaImBlob
        """
        if im_ref.codec == 'JPEG':
            img = Image.open(im_ref.url)
            im_ref.width = img.size[0]
            im_ref.height = img.size[1]

        if self._scene_dh_blob and self._scene_dh_blob.check_buffer_size(im_ref.width, im_ref.height, dh):
            self._scene_dh_blob.set_im_size(im_ref.ts, im_ref.width, im_ref.height, dh)

        else:
            self._scene_dh_blob = _RgbaImBlob(im_ref.ts, im_ref.width, im_ref.height, dh)

        if im_ref.codec == 'JPEG':

            with Image.frombuffer(
                    'RGBA', self._scene_dh_blob.wh, self._scene_dh_blob.buffer,
                    'raw', 'RGBA', 0, 1) as rgba_pil_image:

                rgba_pil_image.readonly = 0

                jpeg_image = Image.open(im_ref.url)
                rgba_pil_image.paste(jpeg_image)

        elif im_ref.codec in ['MONO8T', 'I420PLN']:
            success = self._rawutil.handle_image3(
                im_ref.url, im_ref.codec, im_ref.width, im_ref.height,
                self._scene_dh_blob.buffer)

            if not success:
                eventlog.log_debug('fail: conv image %s -> RGBA' % im_ref.codec)
                raise ValueError

        else:
            raise NotImplemented

        return self._scene_dh_blob

    def _mk_scene_alt_dh_blob(self, ts, w, h, dh):
        """
        :type ts: int
        :type w: int
        :type h: int
        :type dh: int
        :rtype: _RgbaImBlob
        """
        if self._scene_alt_dh_blob and self._scene_alt_dh_blob.check_buffer_size(w, h, dh):
            self._scene_alt_dh_blob.set_im_size(ts, w, h, dh)

        else:
            self._scene_alt_dh_blob = _RgbaImBlob(ts, w, h, dh)

        return self._scene_alt_dh_blob

    def _mk_body_dh_blob(self, ts, w, h, dh):
        """
        :type ts: int
        :type w: int
        :type h: int
        :type dh: int
        :rtype: _RgbaImBlob
        """
        if self._body_dh_blob and self._body_dh_blob.check_buffer_size(w, h, dh):
            self._body_dh_blob.set_im_size(ts, w, h, dh)

        else:
            self._body_dh_blob = _RgbaImBlob(ts, w, h, dh)

        return self._body_dh_blob

    @staticmethod
    def _get_body_ltrbwh(plate_ltrbwh, (scene_w, scene_h), wh_ratio=4.0 / 3.0):
        """
        :type scene_w: int
        :type scene_h: int
        :type plate_ltrbwh: _LtrbWh
        :type wh_ratio: float
        :rtype: _LtrbWh
        :raises: ValueError
        """
        def get_body1_wh():
            """
            Vehicle's make and model would be recognizable in case of wh_ratio 4.0/3.0
            :rtype: float, float
            """

            # define the width of a vehicle as the width of the lane (typ. 2.75m)
            if plate_ltrbwh.w / plate_ltrbwh.h > 2.0:
                # single-line plates with w/h-ratio more then 2:1 (typ. 5:1)
                # suppose the average plate's height that was defined by ANPR is around 10cm
                body_w = (2.75 / 0.10) * plate_ltrbwh.h

            else:
                # double-line plates with w/h-ratio 2:1 or less
                # suppose the average plate's height that was defined by ANPR is around 17cm
                body_w = (2.75 / 0.17) * plate_ltrbwh.h

            return body_w, body_w / wh_ratio

        body1_w, body1_h = get_body1_wh()

        def tweak_body1_wh():
            """
            :rtype: float, float
            :raises: ValueError
            """
            if body1_w <= scene_w and body1_h <= scene_h:
                return body1_w, body1_h

            # strip extra h-pixels from body's height (and keep aspect-ratio)
            if body1_h > scene_h:
                next_body_w = scene_h * wh_ratio
                if plate_ltrbwh.w < next_body_w <= scene_w:
                    return next_body_w, scene_h

            # strip extra w-pixels from body's width (and keep aspect-ratio)
            if body1_w > scene_w:
                next_body_h = scene_w / wh_ratio
                if plate_ltrbwh.h < next_body_h <= scene_h:
                    return scene_w, next_body_h

            # it is imposible to define a vehicle's body
            raise ValueError

        body2_w, body2_h = tweak_body1_wh()
        body2_l = plate_ltrbwh.l + (plate_ltrbwh.w - body2_w) / 2.0
        body2_t = plate_ltrbwh.t + (plate_ltrbwh.h - body2_h) / 2.0

        # move body2 to a visible part of the scene
        if body2_l < 0.0:
            body2_l = 0.0

        elif body2_l + body2_w > scene_w:
            body2_l = scene_w - body2_w

        if body2_t < 0.0:
            body2_t = 0.0

        elif body2_t + body2_h > scene_h:
            body2_t = scene_h - body2_h

        # map spatial coordinates to image's pixel indices
        body3_l, body3_t, body3_w, body3_h = int(body2_l), int(body2_t), int(body2_w), int(body2_h)
        if body3_l + body3_w >= scene_w:
            body3_w -= 1
            if body3_l + body3_w >= scene_w:
                raise ValueError

        if body3_t + body3_h >= scene_h:
            body3_h -= 1
            if body3_t + body3_h >= scene_h:
                raise ValueError

        return _LtrbWh(body3_l, body3_t, body3_l + body3_w, body3_t + body3_h, body3_w, body3_h)

    @staticmethod
    def _get_fake_body_ltrbwh(plate_ltrbwh, (scene_w, scene_h), wh_ratio=4.0 / 3.0):
        """
        :type scene_w: int
        :type scene_h: int
        :type plate_ltrbwh: _LtrbWh
        :type wh_ratio: float
        :rtype: _LtrbWh | None
        """

        plate_center_x = (plate_ltrbwh.l + plate_ltrbwh.r) / 2.0
        plate_center_y = (plate_ltrbwh.t + plate_ltrbwh.b) / 2.0

        def norm(max_coord, coord):
            """
            :type coord: float
            :type max_coord: int
            :rtype: int
            """
            if coord < 0.0:
                return 0

            elif coord > max_coord:
                return int(max_coord)

            else:
                return int(coord)

        def get_body1_ltwh():
            """
            Full scene height and wh_ratio 4.0/3.0
            :rtype: (float, float, float, float) | None
            """
            body2_w = scene_h * wh_ratio
            body2_t = 0
            # tweak x coordinate
            body_center_x_min = 0 + body2_w / 2.0
            body_center_x_max = scene_w - body2_w / 2.0

            if plate_center_x < body_center_x_min:
                body2_l = body_center_x_min - body2_w / 2.0
            elif plate_center_x > body_center_x_max:
                body2_l = body_center_x_max - body2_w / 2.0
            else:
                body2_l = plate_center_x - body2_w / 2.0

            return (body2_l, body2_t, body2_w, scene_h) if body2_w < scene_w else None

        def get_body2_ltwh():
            """
            Full scene widht and wh_ratio 4.0/3.0
            :rtype: (float, float, float, float) | None
            """
            body1_h = scene_w / wh_ratio
            body1_l = 0
            # tweak y coordinate
            body_center_y_min = 0 + body1_h / 2.0
            body_center_y_max = scene_h - body1_h / 2.0

            if plate_center_x < body_center_y_min:
                body1_t = body_center_y_min - body1_h / 2.0
            elif plate_center_x > body_center_y_max:
                body1_t = body_center_y_max - body1_h / 2.0
            else:
                body1_t = plate_center_y - body1_h / 2.0

            return (body1_l, body1_t, scene_w, body1_h) if body1_h < scene_h else None

        result = get_body1_ltwh() or get_body2_ltwh()

        if result is None:
            return None

        l, t, w, h = result

        return _LtrbWh(norm(scene_w, l), norm(scene_h, t), norm(scene_w, l + w), norm(scene_h, t + h),
                       norm(scene_w, w), norm(scene_h, h))

    @staticmethod
    def _mk_scene_jpeg(scene_blob, target_w):
        """
        :type scene_blob: _RgbaImBlob
        :type target_w: int
        :rtype: str
        """
        if scene_blob.w != target_w:
            target_wh = target_w, target_w / (scene_blob.w / float(scene_blob.h))
            return _mk_jpeg_str(_resize_im(scene_blob, target_wh)[0])

        else:
            with Image.frombuffer(
                    'RGBA', scene_blob.wh, scene_blob.buffer,
                    'raw', 'RGBA', scene_blob.stride, 1) as scene_im:

                return _mk_jpeg_str(scene_im)

    @staticmethod
    def _mk_body_thumb_jpeg(scene_blob, body_ltrbwh, target_wh):
        """
        :type scene_blob: _RgbaImBlob
        :type body_ltrbwh: _LtrbWh
        :type target_wh: (int | float, int | float)
        :rtype: str
        """
        return _mk_jpeg_str(_crop_resize_im(scene_blob, body_ltrbwh, target_wh)[0])

    @staticmethod
    def _mk_plate_jpeg(scene_blob, plate_ltrbwh, target_wh):
        """
        :type scene_blob: _RgbaImBlob
        :type plate_ltrbwh: _LtrbWh
        :type target_wh: (int | float, int | float)
        :rtype: str
        """
        crop_offset = 4 * (int(plate_ltrbwh.t) * scene_blob.w + int(plate_ltrbwh.l))

        # crop/scale image
        scene_mv = memoryview(scene_blob.buffer)
        with Image.frombuffer(
                'RGBA', (int(plate_ltrbwh.w), int(plate_ltrbwh.h)), scene_mv[crop_offset:],
                'raw', 'RGBA', scene_blob.stride, 1) as crop_im:

            target_im = crop_im.resize((int(target_wh[0]), int(target_wh[1])), Image.BICUBIC)
            return _mk_jpeg_str(target_im)

    def _mk_body_jpeg(self, scene_blob, body_ltrbwh, plate_ltrbwh, fact, target_wh):
        """
        :type scene_blob: _RgbaImBlob
        :type body_ltrbwh: _LtrbWh
        :type plate_ltrbwh: _LtrbWh
        :type fact: _models.Fact
        :type target_wh: (int | float, int | float)
        :rtype: (str, dict, [int, int])
        """
        footer_args = self._args.body.footer.eval(scene_blob.ts, fact)

        body_im, body_trans = _crop_resize_im(scene_blob, body_ltrbwh, target_wh)
        body_dh_blob = self._mk_body_dh_blob(
            scene_blob.ts, body_im.size[0], body_im.size[1], self._args.body.footer.height)

        # render footer
        body_dh_mv = memoryview(body_dh_blob.buffer)
        footer_offset = int(body_dh_blob.stride * body_dh_blob.h)
        with Image.frombuffer(
                'RGBA', (body_dh_blob.w, body_dh_blob.dh), body_dh_mv[footer_offset:],
                'raw', 'RGBA', body_dh_blob.stride, 1) as body_dh_im:

            body_dh_im.readonly = 0
            w, h = body_dh_im.size
            ImageDraw.Draw(body_dh_im).rectangle([0, 0, w, h], fill='black')
            _draw_footer(body_dh_im, footer_args, self._body_font, self._args.body.footer.font_size)

        # render body and annotations
        with Image.frombuffer(
                'RGBA', (body_dh_blob.w, body_dh_blob.h + body_dh_blob.dh), body_dh_mv,
                'raw', 'RGBA', body_dh_blob.stride, 1) as body_dh_im:

            body_dh_im.readonly = 0
            body_size = body_dh_im.size

            body_dh_im.paste(body_im, (0, 0, body_im.size[0], body_im.size[1]))
            if self._args.body.anno.plate:
                _draw_rect(body_dh_im, plate_ltrbwh, body_trans)

            return _mk_jpeg_str(body_dh_im), body_trans.serialize(), body_size

    def _mk_scene_alt_jpeg(self, scene_dh_blob, body_ltrbwh, plate_ltrbwh, fact, track, target_w):
        """
        :type scene_dh_blob: _RgbaImBlob
        :type body_ltrbwh: _LtrbWh | None
        :type plate_ltrbwh: _LtrbWh
        :type fact: _models.Fact
        :type track: _models.TrTrack
        :type target_w: int | float
        :rtype: str
        """
        if scene_dh_blob.w != self._args.scene_alt.width:
            target_wh = target_w, target_w / (scene_dh_blob.w / float(scene_dh_blob.h))
            alt_im, alt_trans = _resize_im(scene_dh_blob, target_wh)

            alt_dh_blob = self._mk_scene_alt_dh_blob(
                scene_dh_blob.ts, alt_im.size[0], alt_im.size[1],
                self._args.scene_alt.footer.height)

            # render scene
            with Image.frombuffer(
                    'RGBA', (alt_dh_blob.w, alt_dh_blob.h + alt_dh_blob.dh), alt_dh_blob.buffer,
                    'raw', 'RGBA', alt_dh_blob.stride, 1) as alt_dh_im:

                alt_dh_im.readonly = 0
                alt_dh_im.paste(alt_im, (0, 0, alt_im.size[0], alt_im.size[1]))

            # render annotations and footer
            return self._mk_scene_alt_jpeg2(
                alt_dh_blob, alt_trans, body_ltrbwh, plate_ltrbwh, fact, track)

        else:

            # render annotations and footer
            alt_trans = None
            return self._mk_scene_alt_jpeg2(
                scene_dh_blob, alt_trans, body_ltrbwh, plate_ltrbwh, fact, track)

    def _mk_scene_alt_jpeg2(self, scene_dh_blob, scene_dh_trans, body_ltrbwh, plate_ltrbwh, fact, track):
        """
        :type scene_dh_blob: _RgbaImBlob
        :type scene_dh_trans: _ImTrans | None
        :type body_ltrbwh: _LtrbWh | None
        :type plate_ltrbwh: _LtrbWh
        :type fact: _models.Fact
        :type track: _models.TrTrack
        :rtype: str
        """
        footer_args = self._args.scene_alt.footer.eval(scene_dh_blob.ts, fact)

        # render footer
        scene_dh_mv = memoryview(scene_dh_blob.buffer)
        footer_offset = int(scene_dh_blob.stride * scene_dh_blob.h)
        with Image.frombuffer(
                'RGBA', (scene_dh_blob.w, scene_dh_blob.dh), scene_dh_mv[footer_offset:],
                'raw', 'RGBA', scene_dh_blob.stride, 1) as scene_alt_dh_im:

            scene_alt_dh_im.readonly = 0
            w, h = scene_alt_dh_im.size
            ImageDraw.Draw(scene_alt_dh_im).rectangle([0, 0, w, h], fill='black')
            _draw_footer(
                scene_alt_dh_im, footer_args, self._scene_alt_font,
                self._args.scene_alt.footer.font_size)

        # render annotations
        with Image.frombuffer(
                'RGBA', (scene_dh_blob.w, scene_dh_blob.h + scene_dh_blob.dh), scene_dh_mv,
                'raw', 'RGBA', scene_dh_blob.stride, 1) as scene_alt_dh_im:

            scene_alt_dh_im.readonly = 0

            if self._args.scene_alt.anno.plate:
                _draw_rect(scene_alt_dh_im, plate_ltrbwh, scene_dh_trans)

            if self._args.scene_alt.anno.body and body_ltrbwh:
                _draw_rect(scene_alt_dh_im, body_ltrbwh, scene_dh_trans)

            if self._args.scene_alt.anno.track and track.lr_track:
                plate_ltrbwh_on_scene = _LtrbWh.deserialize(fact.plate_rect, scene_dh_blob.wh)
                _draw_track(scene_alt_dh_im, plate_ltrbwh_on_scene, track.lr_track, scene_dh_blob.w, scene_dh_blob.h)

            return _mk_jpeg_str(scene_alt_dh_im)

    def process(self, scene_im_ref, tr_track, fact):
        """
        :type scene_im_ref: _models.LrImageRef
        :type tr_track: _models.TrTrack
        :type fact: _models.Fact
        :rtype: _models.FactImages
        """
        scene_blob = self._mk_scene_dh_blob(scene_im_ref, self._args.scene_alt.footer.height)

        plate_ltrbwh = _LtrbWh.deserialize(fact.plate_rect, scene_blob.wh)
        body_ltrbwh = None

        try:
            body_ltrbwh = FactImProc._get_body_ltrbwh(plate_ltrbwh, scene_blob.wh, self._args.body.wh_ratio)
        except ValueError as x:
            eventlog.log_debug('fail: get body (%s)' % x)

        fact_images = _models.FactImages(ts=scene_im_ref.ts)

        # define scene.measurement
        target_w = self._args.scene.width if self._args.scene.width < scene_blob.w else scene_blob.w
        fact_images.scene_blob = FactImProc._mk_scene_jpeg(scene_blob, target_w)

        # define vehicle.plate
        target_wh = self._args.plate_w, self._args.plate_w / self._args.plate_wh_ratio
        fact_images.plate_blob = FactImProc._mk_plate_jpeg(scene_blob, plate_ltrbwh, target_wh)

        if body_ltrbwh:
            # define vehicle.body.thumb
            target_wh = self._args.body_thumb_w, self._args.body_thumb_w / self._args.body.wh_ratio
            fact_images.body_thumb_blob = FactImProc._mk_body_thumb_jpeg(scene_blob, body_ltrbwh, target_wh)

        else:
            fake_body_ltrbwh = FactImProc._get_fake_body_ltrbwh(plate_ltrbwh, scene_blob.wh, self._args.body.wh_ratio)
            if fake_body_ltrbwh:
                target_wh = fake_body_ltrbwh.w, fake_body_ltrbwh.h
                fact_images.body_thumb_blob = FactImProc._mk_body_thumb_jpeg(scene_blob, fake_body_ltrbwh, target_wh)

        if body_ltrbwh:
            # define vehicle.body
            target_wh = self._args.body.width, self._args.body.width / self._args.body.wh_ratio
            fact_images.body_blob, body_trans_mat, body_size = self._mk_body_jpeg(
                scene_blob, body_ltrbwh, plate_ltrbwh, fact, target_wh)

            fact_images.body_meta = {
                'transform': body_trans_mat,
                'size': body_size
            }

        # define scene.measurement.alt
        target_w = self._args.scene_alt.width if self._args.scene_alt.width < scene_blob.w else scene_blob.w
        fact_images.scene_alt_blob = self._mk_scene_alt_jpeg(
            scene_blob, body_ltrbwh, plate_ltrbwh, fact, tr_track, target_w)

        return fact_images
