import edge.eventlog as eventlog
from edge.lane_detector import LaneDetector

import _models
import _counters
from _args import KernelArgs, SpeedRulesArgs
from _event import EventConsumer
from _mediastore import create_mediastore, create_stub_mediastore
from _plate_absent_db import PlateAbsentDb, StubPlateAbsentDb
from _raas import Raas, StubRaas
from _react import ReactProducer
from _rules import IRule, FactRule, RuleError
from _im_proc import FactImProc
from _gpsd import Gpsd, StubGpsd


class BaseKernel:
    def __init__(self, react_producer, event_consumer, plate_absent_db, fact_rule, speed_rule_args):
        """
        :type react_producer: ReactProducer
        :type event_consumer: EventConsumer
        :type plate_absent_db: IPlateAbsentDb
        :type fact_rule: IRule
        :type speed_rule_args: SpeedRulesArgs
        """
        self._react_producer = react_producer
        self._react_producer.set_track_handler(self._handle_track_react)
        self._event_consumer = event_consumer
        self._plate_absent_db = plate_absent_db

        self._fact_rule = fact_rule
        self._speed_rule_args = speed_rule_args

    def _handle_track_react(self, tr_track):
        """
        :type tr_track: _models.TrTrack
        """
        call_chain = self._fact_rule.check(tr_track)
        call_chain.addCallback(self._handle_fact_done)
        call_chain.addErrback(BaseKernel._handle_fact_fail)

    def _handle_fact_done(self, (fact, fact_images)):
        """
        :type fact: _models.Fact
        :type fact_images: _models.FactImages
        """
        if self._speed_rule_args:
            lane_limits = self._speed_rule_args.get_speed_enforcement(fact.lane)

            if fact.speed is not None:
                fact.set_speed_enforcement(lane_limits[0], lane_limits[1])

        if not fact.plate:
            self._plate_absent_db.handle_fact(fact, fact_images)

        self._event_consumer.consume_fact(fact, fact_images)

    @staticmethod
    def _handle_fact_fail(failure):
        """
        :type failure: twisted.python.failure.Failure
        """
        if failure.check([RuleError]):
            eventlog.log_debug('fail: process track')
        else:
            eventlog.log_debug('fail: process track (%s)' % str(failure))


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, args):
        """
        :type args: KernelArgs
        """
        counters = _counters.StatCounters(
            twisted_reactor, args.counters,
            stat_gpsd=bool(args.gpsd.geodata))

        react_producer = ReactProducer(twisted_reactor, counters.react, args.track_react_endpoint)
        event_consumer = EventConsumer.create(twisted_reactor, counters.event, args.event_consumer)

        if args.plate_absent_db:
            plate_absent_db = PlateAbsentDb(args.plate_absent_db)
        else:
            plate_absent_db = StubPlateAbsentDb()

        raas = Raas(twisted_reactor, args.raas) if args.raas else StubRaas()

        lane_detector = LaneDetector(
            args.lane_detector.id_to_lane_list,
            args.lane_detector.img_width,
            args.lane_detector.img_height)

        if args.ms:
            ms = create_mediastore(twisted_reactor, args.ms)
        else:
            ms = create_stub_mediastore()

        if args.gpsd.geodata:
            gpsd = StubGpsd(args.gpsd.geodata)
        else:
            gpsd = Gpsd(twisted_reactor, counters.gpsd, args.gpsd.endpoint)

        im_proc = FactImProc(args.im_proc)
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', im_proc.destroy)

        fact_rule = FactRule(ms, raas, gpsd, lane_detector, im_proc)

        BaseKernel.__init__(
            self, react_producer, event_consumer, plate_absent_db,
            fact_rule, args.speed_rule_args)

        self._twisted_reactor = twisted_reactor
        self._counters = counters

    def run(self):
        self._twisted_reactor.run()
