import Queue
import json
import sys
from twisted.internet.defer import Deferred, succeed
from twisted.internet.protocol import ReconnectingClientFactory, connectionDone
from twisted.protocols.basic import LineOnlyReceiver

import edge.eventlog as eventlog

from _args import MediaStoreArgs


class MediaStoreError(RuntimeError):
    pass


class IMediaStore:
    def __init__(self):
        pass

    def search_image_ts(self, min_ts, max_ts, limit):
        """
        :type min_ts: int
        :type max_ts: int
        :type limit: int
        :return: callbacks with (list of int); errbacks with MediaStoreError
        :rtype: Deferred
        :raise: MediaStoreError
        """
        raise NotImplemented

    def search_image_url(self, image_ts):
        """
        :type image_ts: int
        :return: callbacks with (str | None); errbacks with MediaStoreError
        :rtype: Deferred
        :raise: MediaStoreError
        """
        raise NotImplemented


class _StubMediaStore(IMediaStore):
    def __init__(self):
        IMediaStore.__init__(self)

    def search_image_ts(self, *args, **kwargs):
        """
        :rtype: Deferred
        """
        return succeed(result=[])

    def search_image_url(self, *args,  **kwargs):
        """
        :rtype: Deferred
        """
        return succeed(result=None)


def create_mediastore(twisted_reactor, args):
    """
    :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX | twisted.internet.interfaces.IReactorTime
    :type args: MediaStoreArgs
    :rtype: IMediaStore
    """
    return _MediaStore(twisted_reactor, args)


def create_stub_mediastore():
    """
    :rtype: IMediaStore
    """
    return _StubMediaStore()


class _Task:
    def __init__(self, query_jnode, call_chain):
        """
        :type query_jnode: dict
        :type call_chain: Deferred
        """
        self.query_jnode = query_jnode
        self.call_chain = call_chain


class _IMsProtocol:
    def __init__(self):
        pass

    def tx_query(self, query_jnode):
        """
        :type query_jnode: dict
        """
        raise NotImplemented


class _IMsProtocolHandler:
    def __init__(self):
        pass

    def proto_connection_made(self, protocol):
        """
        :type protocol: _IMsProtocol
        """
        raise NotImplemented

    def proto_connection_lost(self):
        raise NotImplemented

    def proto_rx_reply(self, reply_jnode):
        """
        :type reply_jnode: dict
        """
        raise NotImplemented


class _MsUdsProtocol(LineOnlyReceiver):
    MAX_LENGTH = sys.maxint
    encoding = 'utf-8'
    delimiter = b'\0'

    def __init__(self):
        pass

    def connectionMade(self):
        raise NotImplemented

    # Called when an established connection is lost.
    def connectionLost(self, reason=connectionDone):
        raise NotImplemented

    def lineReceived(self, raw_reply):
        """
        :type raw_reply: str
        """
        reply_jnode = None
        try:
            reply_jnode = json.loads(raw_reply.decode(self.encoding))
        except ValueError:
            eventlog.log_debug_last_exc()

        self.rx_reply(reply_jnode)

    def tx_query(self, query_jnode):
        """
        :type query_jnode: dict
        """
        raw_query = json.dumps(query_jnode).encode(self.encoding)
        self.sendLine(raw_query)

    def rx_reply(self, reply_jnode):
        """
        :type reply_jnode: dict
        """
        raise NotImplemented


class _MsProtocol(_MsUdsProtocol, _IMsProtocol):
    """
    :type _wait_reply: bool
    """
    def __init__(self, handler):
        """
        :type handler: _IMsProtocolHandler
        """
        _MsUdsProtocol.__init__(self)

        self._handler = handler
        self._wait_reply = False

    def connectionMade(self):
        assert not self._wait_reply
        self._handler.proto_connection_made(self)

    def connectionLost(self, reason=connectionDone):
        self._wait_reply = False
        self._handler.proto_connection_lost()

    def tx_query(self, query_jnode):
        """
        :type query_jnode: dict
        """
        assert not self._wait_reply

        self._wait_reply = True
        _MsUdsProtocol.tx_query(self, query_jnode)

    def rx_reply(self, reply_jnode):
        """
        :type reply_jnode: dict | None
        """
        assert self._wait_reply
        self._wait_reply = False
        self._handler.proto_rx_reply(reply_jnode)


class _MediaStore(ReconnectingClientFactory, IMediaStore, _IMsProtocolHandler):
    """
    :type _protocol: _MsProtocol
    :type _task: _Task | None
    """

    # adjust ReconnectingClientFactory
    maxDelay = 5
    factor = 1.6180339887498948

    def __init__(self, twisted_reactor, args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorUNIX | twisted.internet.interfaces.IReactorTime
        :type args: MediaStoreArgs
        """
        IMediaStore.__init__(self)
        _IMsProtocolHandler.__init__(self)

        self._twisted_reactor = twisted_reactor
        self._channel = args.channel
        self._protocol = None

        self._task_queue = Queue.Queue(maxsize=2)
        self._task = None

        twisted_reactor.connectUNIX(args.endpoint, self)

    def buildProtocol(self, addr):
        """
        Called when a connection has been established to addr.

        If None is returned, the connection is assumed to have been refused,
        and the Port will close the connection.

        :param addr: The address of the newly-established connection
        :type addr: twisted.internet.interfaces.IAddress

        :return: None if the connection was refused, otherwise an object
                 providing twisted.internet.interfaces.IProtocol
        :rtype: twisted.internet.interfaces.IProtocol | None
        """
        eventlog.log_debug('')

        self.resetDelay()
        protocol = _MsProtocol(self)
        protocol.factory = self
        return protocol

    def startedConnecting(self, connector):
        """Called when a connection has been started.

        You can call connector.stopConnecting() to stop the connection attempt.

        :param connector: Object used to interface between connections and protocols.
        :type connector: twisted.internet.interfaces.IConnector
        """
        eventlog.log_debug('')
        ReconnectingClientFactory.startedConnecting(self, connector)

    def clientConnectionFailed(self, connector, reason):
        """Called when a connection has failed to connect.

        It may be useful to call connector.connect() - this will reconnect.

        :param connector: Object used to interface between connections and protocols.
        :type connector: twisted.internet.interfaces.IConnector
        :type reason: twisted.python.failure.Failure
        """
        eventlog.log_debug('')
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)

    def clientConnectionLost(self, connector, reason):
        """Called when an established connection is lost.

        It may be useful to call connector.connect() - this will reconnect.

        :param connector: Object used to interface between connections and protocols.
        :type connector: twisted.internet.interfaces.IConnector
        :type reason: twisted.python.failure.Failure
        """
        eventlog.log_debug('')
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)

    def proto_connection_made(self, protocol):
        """
        :type protocol: _IMsProtocol
        """
        eventlog.log_debug('')
        self._protocol = protocol

        if self._task:
            # retry to handle the last task
            next_task = self._task
            self._task = None
            self._protocol.tx_query(next_task.query_jnode)

        else:
            # try to handle the next task.
            self._twisted_reactor.callLater(0, self._handle_next_task)

    def proto_connection_lost(self):
        # keep the last task until next connection
        eventlog.log_debug('')
        self._protocol = None

    def proto_rx_reply(self, reply_jnode):
        """
        :type reply_jnode: dict
        """
        call_chain = self._task.call_chain
        self._task = None

        self._twisted_reactor.callLater(0, self._handle_next_task)
        call_chain.callback(reply_jnode)

    def _handle_next_task(self):
        if self._task or not self._protocol:
            return

        try:
            self._task = self._task_queue.get_nowait()
            self._protocol.tx_query(self._task.query_jnode)

        except Queue.Empty:
            pass

    def search_image_ts(self, min_ts, max_ts, limit):
        """
        :type min_ts: int
        :type max_ts: int
        :type limit: int

        :return: callbacks with (list of int); errbacks with MediaStoreError
        :rtype: Deferred
        :raise: MediaStoreError
        """
        query_jnode = {
            'Action': 'SearchFrame',
            'Channel': self._channel,
            'StartTs': min_ts,
            'StopTs': max_ts,
            'Limit': limit,
            'Context': None
        }

        def handle_done(reply_jnode):
            """
            :type reply_jnode: dict
            :rtype: list of int
            :raise: MediaStoreError
            """
            status_done = 1
            status_fail = 2

            status = reply_jnode.get('Status', status_fail)
            if status == status_done:
                image_ts_list = reply_jnode['FrameTs']
                return image_ts_list

            eventlog.log_debug('fail: search-image-ts (status=%s)' % status)
            raise MediaStoreError()

        call_chain = Deferred()
        call_chain.addCallback(callback=handle_done)

        try:
            self._task_queue.put_nowait(_Task(query_jnode, call_chain))
            self._twisted_reactor.callLater(0, self._handle_next_task)
        except Queue.Full:
            eventlog.log_debug('fail: queue overflow detected')
            raise MediaStoreError()

        return call_chain

    def search_image_url(self, image_ts):
        """
        :type image_ts: int
        :return: callbacks with (str | None); errbacks with MediaStoreError
        :rtype: Deferred
        :raise: MediaStoreError
        """
        query_jnode = {
            'Action': 'LoadFrame',
            'Channel': self._channel,
            'FrameTs': image_ts,
            'Context': None
        }

        def handle_done(reply_jnode):
            """
            :type reply_jnode: dict
            :rtype: str | None
            :raise: MediaStoreError
            """
            status_present = 1
            status_absent = 2
            status_fail = 3

            status = reply_jnode.get('Status', status_fail)
            if status == status_present:
                return reply_jnode['Frame']

            elif status == status_absent:
                eventlog.log_debug('fail: search-image-url (image absent)')
                return None

            eventlog.log_debug('fail: search-image-url (status=%s)' % status)
            raise MediaStoreError('invalid reply status')

        call_chain = Deferred()
        call_chain.addCallbacks(callback=handle_done)

        try:
            self._task_queue.put_nowait(_Task(query_jnode, call_chain))
            self._twisted_reactor.callLater(0, self._handle_next_task)
        except Queue.Full:
            eventlog.log_debug('fail: queue overflow detected')
            raise MediaStoreError()

        return call_chain
