import uuid

from edge.types import ImageRef


class Geodata:
    def __init__(self, latitude, longitude, altitude, speed):
        """
        :param latitude: Latitude in degrees: +/- signifies North/South
        :type latitude: int | float
        :param longitude: Longitude in degrees: +/- signifies East/West
        :type longitude: int | float
        :param altitude: Altitude in meters
        :type altitude: float | None
        :param speed: Speed over ground, meters per second
        :type speed: float | None
        """

        # http://catb.org/gpsd/gpsd_json.html
        self.latitude = latitude
        self.longitude = longitude
        self.altitude = altitude
        self.speed = speed

    def serialize(self):
        """
        :return [self.latitude, self.longitude, self.altitude, self.speed]
        :return [self.latitude, self.longitude, self.altitude]
        :return [self.latitude, self.longitude]
        :rtype: list of float
        """
        geodata_list = [
            self.latitude,
            self.longitude,
            self.altitude
        ]

        if self.speed is not None:
            geodata_list.append(self.speed)

        elif self.altitude is None:
            geodata_list.pop()

        # return [self.latitude, self.longitude, self.altitude | None, self.speed | None]
        return geodata_list

    def serialize2(self):
        """
        :rtype: str
        """
        # <latChar><DDD.uuuuuu> <longChar><ddd.uuuuuu>
        lat_char = 'N' if self.latitude >= 0 else 'S'
        long_char = 'E' if self.longitude >= 0 else 'W'
        geodata_str = '%s%.6f %s%.6f' % (lat_char, abs(self.latitude), long_char, abs(self.longitude))
        return geodata_str


class PlateRect:
    LEFT_INDEX = 0
    TOP_INDEX = 1
    RIGHT_INDEX = 2
    BOTTOM_INDEX = 3

    def __init__(self, left, top, right, bottom):
        """
        :type left: float
        :type top: float
        :type right: float
        :type bottom: float
        """
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

    def check_visible(self):
        return self.left > 0.0 and self.right < 1.0 and self.top > 0.0 and self.bottom < 1.0

    @staticmethod
    def deserialize(lr_rect_jnode):
        """
        :type lr_rect_jnode: list of float
        :rtype: PlateRect
        """
        left = float(lr_rect_jnode[PlateRect.LEFT_INDEX])
        top = float(lr_rect_jnode[PlateRect.TOP_INDEX])
        right = float(lr_rect_jnode[PlateRect.RIGHT_INDEX])
        bottom = float(lr_rect_jnode[PlateRect.BOTTOM_INDEX])

        plate_rect = PlateRect(left, top, right, bottom)
        return plate_rect

    @staticmethod
    def serialize_lr(plate_rect):
        """
        :type plate_rect: PlateRect
        :rtype: list of float
        """
        return [plate_rect.left, plate_rect.top, plate_rect.right, plate_rect.bottom]


class Plate:
    def __init__(self, validity, text, template):
        """
        :type validity: int
        :type text: str
        :type template: int
        """
        self.validity = validity
        self.text = text
        self.template = template

    @staticmethod
    def deserialize_lr(plate_jnode):
        """
        :type plate_jnode: dict
        :rtype: Plate
        """
        validity = plate_jnode['V']
        text = plate_jnode['P']
        template = plate_jnode['T']

        plate = Plate(validity, text, template)
        return plate

    @staticmethod
    def deserialize_raas(raas_plate_jnode):
        """
        :type raas_plate_jnode: dict
        :rtype: Plate
        """
        validity = raas_plate_jnode['weight']
        text = raas_plate_jnode['number']
        country = int(raas_plate_jnode['country']) << 16
        template = country | int(raas_plate_jnode['template'])

        raas_plate = Plate(validity, text, template)
        return raas_plate


class RaasSceneItem:
    def __init__(self, variants):
        """
        :type variants: list of RaasPlate
        """
        self.variants = variants

    @staticmethod
    def deserialize(raas_scene_item_jnode):
        """
        :type raas_scene_item_jnode: dict
        :rtype: RaasSceneItem
        """
        variants = [Plate.deserialize_raas(raas_scene_item_jnode)]
        raas_scene_item = RaasSceneItem(variants)
        return raas_scene_item


class RaasScene:
    def __init__(self, items):
        """
        :type items: list of RaasSceneItem
        """
        self.items = items

    def try_get_first_scene_item_plate(self):
        """
        :rtype: Plate | None
        """
        for scene_item in self.items:
            max_validity_variant = max(scene_item.variants, key=lambda entry: entry.validity)
            return max_validity_variant

        return None

    @staticmethod
    def deserialize(raas_scene_jnode):
        """
        :type raas_scene_jnode: dict
        :rtype: RaasScene
        """
        if 0 != raas_scene_jnode['result'] and 1 != raas_scene_jnode['version']:
            raise ValueError

        raas_scene_items = map(lambda raas_scene_item_jnode: RaasSceneItem.deserialize(raas_scene_item_jnode),
                               raas_scene_jnode['hypos'])
        raas_scene = RaasScene(raas_scene_items)
        return raas_scene


class LrTrackItem:
    def __init__(self, ts, plate_weight, plate_rect):
        """
        :type ts: int
        :type plate_weight: int
        :type plate_rect: PlateRect
        """
        self.ts = ts
        self.plate_weight = plate_weight
        self.plate_rect = plate_rect

    @staticmethod
    def deserialize(lr_track_item_jnode):
        """
        :type lr_track_item_jnode: dict
        :rtype: LrTrackItem
        """
        ts = lr_track_item_jnode['Ts']
        weight = lr_track_item_jnode['W']
        plate_rect = PlateRect.deserialize(lr_track_item_jnode['R'])

        lr_track_item = LrTrackItem(ts, weight, plate_rect)
        return lr_track_item


class LrImageRef(ImageRef):
    def __init__(self, ts=None, url=None, codec=None, width=None, height=None,  rect=None):
        """
        :type ts: int | None
        :type url: str | None
        :type codec: str | None
        :type width: int | None
        :type height: int | None
        :type rect: PlateRect | None
        """
        ImageRef.__init__(self, ts, url, codec, width, height)
        self.rect = rect

    @staticmethod
    def deserialize(lr_best_image_jnode):
        """
        :type lr_best_image_jnode: dict
        :rtype: LrImageRef
        """
        ts = lr_best_image_jnode['Ts']
        url = str(lr_best_image_jnode['Url'])
        left, top, right, bottom = lr_best_image_jnode['R']

        if left < 0.0:
            left = 0.0

        if top < 0.0:
            top = 0.0

        if right > 1.0:
            right = 1.0

        if bottom > 1.0:
            bottom = 1.0

        plate_rect = PlateRect(left, top, right, bottom)

        raw_wh = lr_best_image_jnode['Wh']
        w = int(raw_wh[0])
        h = int(raw_wh[1])
        codec = lr_best_image_jnode['Codec']
        lr_image = LrImageRef(ts=ts, url=url, codec=codec, width=w, height=h, rect=plate_rect)
        return lr_image


class LrTrack:
    def __init__(self, catch_ts, lost_ts, direction, plate, plate_trace, best_image, src_filename):
        """
        :type catch_ts: int
        :type lost_ts: int
        :type direction: int
        :type plate: Plate
        :type plate_trace: list of LrTrackItem
        :type best_image: LrImageRef | None
        :type src_filename: str
        """
        self.result = None
        self.catch_ts = catch_ts
        self.lost_ts = lost_ts
        self.direction = direction
        self.plate = plate
        self.plate_trace = plate_trace
        self.best_image = best_image
        self.src_file_name = src_filename

    @staticmethod
    def deserialize(lr_track_jnode, src_filename=None):
        """
        :type lr_track_jnode: dict
        :type src_filename: str | None
        :rtype: LrTrack
        """
        direction = lr_track_jnode['Dir']
        plate = Plate.deserialize_lr(lr_track_jnode['Plate'])
        plate_trace = map(lambda lr_track_item_jnode: LrTrackItem.deserialize(lr_track_item_jnode),
                          lr_track_jnode['PlateTrace'])

        catch_ts = plate_trace[0].ts
        lost_ts = plate_trace[-1].ts

        lr_track_best_image_jnode = lr_track_jnode.get('Image')
        best_image = LrImageRef.deserialize(lr_track_best_image_jnode) if lr_track_best_image_jnode else None

        lr_track = LrTrack(catch_ts, lost_ts, direction, plate, plate_trace, best_image, src_filename)
        return lr_track


class RaTrackItem:
    def __init__(self, ts, x, y, speed):
        """
        :type ts: int
        :type x: float
        :type y: float
        :type speed: int
        """
        self.ts = ts
        self.x = x
        self.y = y
        self.speed = speed

    @staticmethod
    def deserialize(ra_track_item_jnode):
        """
        :type ra_track_item_jnode: dict
        :rtype: RaTrackItem
        """
        ts = ra_track_item_jnode['Ts']
        x, y = ra_track_item_jnode['Xy']
        speed_mps = ra_track_item_jnode['S']

        ra_track_item = RaTrackItem(ts, x, y, speed_mps)
        return ra_track_item


class RaTrack:
    def __init__(self, catch_ts, lost_ts, length, trace):
        """
        :type catch_ts: int
        :type lost_ts: int
        :type length: int
        :type trace: list of RaTrackItem
        """
        self.catch_ts = catch_ts
        self.lost_ts = lost_ts
        self.length = length
        self.trace = trace

    def get_median_speed_kmph(self):
        """
        :rtype: int | None
        """
        speed_list = map(lambda ra_scene_item: ra_scene_item.speed, self.trace)

        speed_list_len = len(speed_list)
        if speed_list_len > 0:
            speed_list.sort()
            index = speed_list_len / 2
            return int(speed_list[index])

        return None

    @staticmethod
    def deserialize(ra_track_jnode):
        """
        :type ra_track_jnode: dict
        :rtype: RaTrack
        """
        catch_ts = ra_track_jnode['CatchTs']
        lost_ts = ra_track_jnode['LostTs']
        length = ra_track_jnode['Length']
        trace = map(lambda ra_track_item_jnode: RaTrackItem.deserialize(ra_track_item_jnode),
                    ra_track_jnode['Trace'])

        ra_track = RaTrack(catch_ts, lost_ts, length, trace)
        return ra_track


class TrTrack:
    def __init__(self, lr_track, ra_track):
        """
        :type lr_track: LrTrack | None
        :type ra_track: RaTrack | None
        """
        self.lr_track = lr_track
        self.ra_track = ra_track

    @staticmethod
    def deserialize(tr_track_jnode):
        """
        :type tr_track_jnode: dict
        :rtype: TrTrack
        """
        lr_track_jnode = tr_track_jnode.get('Lr')
        lr_track = LrTrack.deserialize(lr_track_jnode) if lr_track_jnode else None

        ra_track_jnode = tr_track_jnode.get('Ra')
        ra_track = RaTrack.deserialize(ra_track_jnode) if ra_track_jnode else None

        tr_track = TrTrack(lr_track, ra_track)
        return tr_track


class FactImages:
    """
    :type plate_blob: str | None
    :type plate_meta: dict | None

    :type body_thumb_blob: str | None
    :type body_thumb_meta: dict | None

    :type body_blob: str | None
    :type body_meta: dict | None

    :type scene_blob: str | None
    :type scene_meta: dict | None

    :type scene_alt_blob: str | None
    :type scene_alt_meta: dict | None
    """
    def __init__(self, ts):
        """
        :type ts: int
        """
        self.ts = ts

        self.plate_blob = None
        self.plate_meta = None

        self.body_thumb_blob = None
        self.body_thumb_meta = None

        self.body_blob = None
        self.body_meta = None

        self.scene_blob = None
        self.scene_meta = None

        self.scene_alt_blob = None
        self.scene_alt_meta = None


class Fact:
    """
    :type speed_limit: int
    :type max_overspeed: int
    """

    def __init__(self, plate, direction, speed, length, lane, plate_rect, catch_ts, lost_ts, geodata):
        """
        :type plate: Plate | None
        :type direction: int
        :type speed: int | None
        :type length: int
        :type lane: int
        :type plate_rect: PlateRect
        :type catch_ts: int
        :type lost_ts: int
        :type geodata : Geodata | None
        """
        self.plate = plate
        self.direction = direction
        self.speed = speed
        self.length = length

        self._guid = str(uuid.uuid4())
        self.lane = lane
        self.plate_rect = plate_rect

        self.speed_limit = None
        self.max_overspeed = None

        self.catch_ts = catch_ts
        self.lost_ts = lost_ts
        self.geodata = geodata

    def set_lane(self, lane):
        """
        :type lane: int
        """
        self.lane = lane

    def set_plate(self, plate):
        """
        :type plate: Plate
        """
        self.plate = plate

    def set_geodata(self, geodata):
        """
        :type geodata: Geodata | None
        """
        self.geodata = geodata

    def set_speed_enforcement(self, speed_limit, max_overspeed):
        """
        :type speed_limit: int
        :type max_overspeed: int
        """
        self.speed_limit = speed_limit
        self.max_overspeed = max_overspeed

    @staticmethod
    def serialize(fact, images_ts, images_jnode):
        """
        :type fact: Fact
        :type images_jnode: list of dict
        :type images_ts: int
        :rtype: dict
        """
        if not fact.plate_rect:
            raise ValueError

        fact_jnode = {
            'fact': {
                'vehicle': {
                    'lane': fact.lane,
                    'dir': fact.direction,
                    'lenght': fact.length
                },
                'best-ts': images_ts,
                'catch-ts': fact.catch_ts,
                'lost-ts': fact.lost_ts,
                'uuid': fact._guid
            },
            'images': images_jnode,
            'alarm': {

            },

        }
        if fact.geodata is not None:
            fact_jnode['fact']['geolocation'] = fact.geodata.serialize()

        if fact.plate:
            plate_jnode = {
                'rect': PlateRect.serialize_lr(fact.plate_rect),
                'validity': fact.plate.validity,
                'text': fact.plate.text,
                'type': [(fact.plate.template >> 20) & 0xFFF,
                         (fact.plate.template >> 10) & 0x3FF,
                         fact.plate.template & 0x3FF]
            }

            fact_jnode['fact']['vehicle']['plate'] = plate_jnode

        else:
            fact_jnode['fact']['vehicle'] = {
                'plate': {
                    'rect': PlateRect.serialize_lr(fact.plate_rect),
                    'validity': 0,
                    'text': '',
                    'type': [-1, -1, -1]
                }
            }

        if fact.speed:
            fact_jnode['fact']['vehicle'].update({'speed': fact.speed})

        if fact.speed_limit:
            fact_jnode['fact']['speed-limit'] = fact.speed_limit

        if fact.max_overspeed:
            fact_jnode['fact']['max-overspeed'] = fact.max_overspeed

        return fact_jnode
