import ujson
from io import BytesIO
from twisted.internet.defer import Deferred, succeed
from twisted.internet.protocol import Protocol, connectionDone
from twisted.web.client import Agent, Response
from twisted.web.client import FileBodyProducer
from twisted.web.http_headers import Headers
from PIL import Image

import edge.eventlog as eventlog

from _args import RaasArgs
from _models import LrImageRef, PlateRect, RaasScene


class RaasError(RuntimeError):
    pass


class IRaas:
    def __init__(self):
        pass

    def analyze_scene(self, scene_ref, plate_rect):
        """
        :type scene_ref: LrImageRef
        :type plate_rect: PlateRect
        :return: callbacks to RaasScene; errbacks to RaasError
        :rtype: Deferred
        :raise: RaasError
        """
        raise NotImplemented


class StubRaas(IRaas):
    # noinspection PyUnusedLocal
    def __init__(self, *args, **kwargs):
        IRaas.__init__(self)
        self._scene = RaasScene(items=[])

    def analyze_scene(self, *args, **kwargs):
        """
        :return: callbacks to RaasScene
        :rtype: Deferred
        """
        return succeed(self._scene)


class _ResponseBodyProto(Protocol):
    def __init__(self, handler):
        """
        :type handler: Deferred - callbacks to str
        """
        self._handler = handler
        self._body_buffer = ''

    def dataReceived(self, body_data):
        """
        :type body_data: str
        """
        self._body_buffer += body_data

    def connectionLost(self, reason=connectionDone):
        self._handler.callback(self._body_buffer[: -1])  # strip '\0'


class Raas(IRaas):
    def __init__(self, twisted_reactor, args):
        """
        :type args: RaasArgs
        """
        IRaas.__init__(self)

        self._agent = Agent(twisted_reactor)

        self._base_url = 'http://%s:%s/alpr/%s/' % (args.raas_host, args.raas_port, args.raas_user)
        self._password = args.raas_pass

    @staticmethod
    def _upscale_plate_rect(plate_rect, (scene_w, scene_h)):
        """
        :type plate_rect: PlateRect
        :type scene_w: int
        :type scene_h: int
        :rtype: (int, int, int, int)
        """
        px_margin = int((plate_rect.bottom - plate_rect.top) * scene_h)

        px_left = int(plate_rect.left * scene_w) - px_margin
        px_top = int(plate_rect.top * scene_h) - px_margin
        px_right = int(plate_rect.right * scene_w) + px_margin
        px_bottom = int(plate_rect.bottom * scene_h) + px_margin
        return px_left, px_top, px_right, px_bottom

    @staticmethod
    def _mk_signature(plate_image_bytes_io, secret):
        from hashlib import md5

        md5_hash = md5(plate_image_bytes_io.getvalue())
        md5_hash.update(BytesIO(secret).getvalue())
        signature = md5_hash.hexdigest()

        return signature

    def analyze_scene(self, scene_ref, plate_rect):
        """
        :type scene_ref: LrImageRef
        :type plate_rect: PlateRect
        :return: callbacks to RaasScene; errbacks to RaasError
        :rtype: Deferred
        :raise: RaasError
        """
        if scene_ref.codec != 'JPEG':
            raise NotImplemented

        scene_im = Image.open(scene_ref.url)
        plate_ltrb = Raas._upscale_plate_rect(plate_rect, scene_im.size)
        plate_image = scene_im.crop(plate_ltrb)  # TODO: crop via stride

        mem_file = BytesIO()
        plate_image.save(mem_file, 'JPEG', quality=90)
        plate_image_signature = Raas._mk_signature(mem_file, self._password)

        mem_file.seek(0)
        qbody_producer = FileBodyProducer(mem_file)

        call_chain = self._agent.request(
            'POST', self._base_url + plate_image_signature,
            Headers({'User-Agent': ['EDGE'], 'Content-Type': ['image/jpeg']}),
            qbody_producer)

        call_chain.addCallbacks(
            callback=Raas._handle_reply_http_header_done,
            errback=Raas._handle_reply_http_header_fail)

        call_chain.addCallbacks(
            callback=Raas._handle_reply_http_body_done,
            errback=Raas._handle_reply_http_body_fail)

        return call_chain

    @staticmethod
    def _handle_reply_http_header_done(response):
        """
        :type response: Response
        :return: callbacks to str
        :rtype: Deferred
        :raise: RaasError
        """
        if response.code != 200:
            eventlog.log_debug('fail: invalid-reply (http-code=%s)' % response.code)
            raise RaasError()

        call_chain = Deferred()
        response.deliverBody(_ResponseBodyProto(call_chain))
        return call_chain

    @staticmethod
    def _handle_reply_http_header_fail(failure):
        """
        :type failure: twisted.python.failure.Failure
        :raise: RaasError
        """
        eventlog.log_debug('fail: invalid-query (%s)' % str(failure))
        raise RaasError

    @staticmethod
    def _handle_reply_http_body_done(response_body):
        """
        :type response_body: str
        :rtype: RaasScene
        """
        raas_scene_jnode = ujson.loads(response_body)
        raas_scene = RaasScene.deserialize(raas_scene_jnode)
        return raas_scene

    @staticmethod
    def _handle_reply_http_body_fail(failure):
        """
        :type failure: twisted.python.failure.Failure
        :raise: RaasError
        """
        if not failure.check([RaasError]):
            eventlog.log_debug('fail: load response-query (%s)' % str(failure))

        raise RaasError
