import ujson
from twisted.internet import inotify
from twisted.python.filepath import FilePath

import edge.eventlog as eventlog
from edge.fs_helpers import silent_mkdir_r, silent_rm_r

from _counters import ReactProducerCounters
import _models


class IReactProducer:
    def __init__(self):
        pass

    def set_track_handler(self, react_handler):
        """
        :type react_handler: (, _models.TrTrack) -> None
        :rtype: bool
        """
        raise NotImplemented


# noinspection PyUnusedLocal
def stub_track_react_handler(tr_track):
    """
    :type tr_track: _models.TrTrack
    """
    print 'stub_track_react_handler(tr_track)'


def _load_react_args_jnode(react_args_filepath):
    """
    :type react_args_filepath: str | None
    :rtype: dict
    """
    react_jnode = None
    if not react_args_filepath:
        return react_jnode

    # noinspection PyBroadException
    try:
        with open(react_args_filepath) as json_file:
            react_jnode = ujson.load(json_file)

    except IOError as x:
        eventlog.log_debug('fail: read %s (%s)' % (react_args_filepath, x))

    except ValueError as x:
        eventlog.log_debug('fail: deserialize %s (%s)' % (react_args_filepath, x))

    except:
        eventlog.log_debug_last_exc()

    return react_jnode


class ReactProducer(IReactProducer):
    def __init__(self, twisted_reactor, counters, react_endpoint):
        """
        :type counters: ReactProducerCounters
        :type react_endpoint: str
        """
        IReactProducer.__init__(self)

        self._twisted_reactor = twisted_reactor

        self._counters = counters
        self._endpoint = react_endpoint

        self._react_handler = None
        self._inotify = inotify.INotify()
        self._inotify.startReading()

    def set_track_handler(self, react_handler):
        """
        :type react_handler: (, _models.TrTrack) -> bool
        :rtype: bool
        """
        # Allow to set the handler only once.
        if react_handler and not self._react_handler:
            self._react_handler = react_handler
            self._reset_inotify_watch()
            return True

        return False

    def _reset_inotify_watch(self):
        try:
            self._inotify.ignore(FilePath(self._endpoint))
        except KeyError:
            pass

        silent_mkdir_r(self._endpoint)

        try:
            self._inotify.watch(
                FilePath(self._endpoint),
                mask=inotify.IN_MOVED_TO | inotify.IN_CLOSE_WRITE | inotify.IN_DELETE_SELF | inotify.IN_MOVE_SELF,
                callbacks=[self._handle_react])
        except inotify.INotifyError as x:
            eventlog.log_debug('fail: reset inotify watch (%s)' % x)

    # noinspection PyUnusedLocal
    def _handle_react(self, ignored, filepath, mask):
        if (mask & inotify.IN_DELETE_SELF) or (mask & inotify.IN_MOVE_SELF):
            self._twisted_reactor.callLater(0, self._reset_inotify_watch)
            eventlog.log_warning('endpoint was changed to %s' % self._endpoint)
            return

        tr_track = None
        try:
            react_jnode = _load_react_args_jnode(filepath.path)
            try:
                tr_track = _models.TrTrack.deserialize(react_jnode)
            except TypeError, x:
                return

        except ValueError as x:
            eventlog.log_debug('fail: deserialize %s (%s)' % (filepath.path, x))
            return

        finally:
            silent_rm_r(filepath.path)

        # noinspection PyBroadException
        try:
            self._counters.inc_track(tr_track)

            if tr_track.lr_track:
                plate_trace = tr_track.lr_track.plate_trace
                self._counters.inc_plate_rect(
                    plate_trace[0].plate_rect, plate_trace[-1].plate_rect)

            self._react_handler(tr_track)

        except Exception:
            eventlog.log_debug_last_exc()
