import time
from twisted.internet.defer import Deferred, fail, succeed

import _models
import edge.eventlog as eventlog
from edge.lane_detector import LaneDetector

from _raas import IRaas, RaasError
from _gpsd import IGpsd
from _im_proc import FactImProc
from _mediastore import IMediaStore, MediaStoreError
from _best_helpers import get_best_lr_track_item, get_best_ra_ts_and_rect


class RuleError(RuntimeError):
    pass


class IRule:
    def __init__(self):
        pass

    def check(self, tr_track):
        """
        :type tr_track: _models.TrTrack
        :return: callbacks to (_models.Fact, _models.FactImages); errbacks to RuleError
        :rtype: Deferred
        """
        raise NotImplemented


class FactRule(IRule):
    def __init__(self, ms, raas, gpsd, lane_detector, im_proc):
        """
        :type ms: IMediaStore
        :type raas: IRaas
        :type gpsd: IGpsd
        :type lane_detector: LaneDetector
        :type im_proc: FactImProc
        """
        self._ms = ms
        self._raas = raas
        self._gpsd = gpsd
        self._lane_detector = lane_detector
        self._im_proc = im_proc

        IRule.__init__(self)

    def check(self, tr_track):
        """
        :type tr_track: _models.TrTrack
        :return: callbacks to (_models.Fact, _models.FactImages); errbacks to RuleError
        :rtype: Deferred
        """

        if tr_track.lr_track:
            catch_ts = tr_track.lr_track.catch_ts
            lost_ts = tr_track.lr_track.lost_ts

        elif tr_track.ra_track:
            catch_ts = tr_track.ra_track.catch_ts
            lost_ts = tr_track.ra_track.lost_ts

        else:
            raise NotImplemented

        call_chain = self._ms.search_image_ts(catch_ts, lost_ts, limit=1000)
        call_chain.addCallbacks(
            callback=self._handle_ms_search_ts_done, callbackArgs=(tr_track, catch_ts, lost_ts),
            errback=self._handle_ms_search_ts_fail, errbackArgs=(tr_track, catch_ts, lost_ts))

        return call_chain

    def _check_lr_image(self, tr_track, catch_ts, lost_ts):
        """
        :type tr_track: _models.TrTrack
        :type catch_ts: int
        :type lost_ts: int
        :return:
        :rtype: Deferred
        """
        if tr_track.ra_track:
            speed = tr_track.ra_track.get_median_speed_kmph()
            length = tr_track.ra_track.length

        else:
            speed = None
            length = 0

        geodata = self._gpsd.get_geodata(tr_track.lr_track.best_image.ts)

        fact = _models.Fact(
            plate=tr_track.lr_track.plate,
            direction=tr_track.lr_track.direction,
            speed=speed,
            length=length,
            lane=0,
            plate_rect=tr_track.lr_track.best_image.rect,
            catch_ts=catch_ts,
            lost_ts=lost_ts,
            geodata=geodata)

        if not tr_track.lr_track.best_image:  # TODO
            return succeed(result=(fact, None))

        lane_set = self._lane_detector.get_lane2(
            fact.plate_rect.left, fact.plate_rect.top,
            fact.plate_rect.right, fact.plate_rect.bottom)

        if len(lane_set) == 1:
            fact.lane = lane_set.pop()

        call_chain = self._raas.analyze_scene(tr_track.lr_track.best_image, fact.plate_rect)
        call_chain.addCallbacks(
            callback=self._handle_raas_done, callbackArgs=(fact, tr_track.lr_track.best_image, tr_track),
            errback=self._handle_raas_fail, errbackArgs=(fact, tr_track.lr_track.best_image, tr_track))

        return call_chain

    # noinspection PyUnusedLocal
    def _handle_ms_search_ts_done(self, image_ts_list, tr_track, catch_ts, lost_ts):
        """
        :type image_ts_list: list of int
        :type tr_track: _models.TrTrack
        :type catch_ts: int
        :type lost_ts: int
        :rtype: Deferred | None
        """
        if not image_ts_list:
            # mediastore didn't find images or we are using stub mediastore
            return self._check_lr_image(tr_track, catch_ts, lost_ts)

        # mediastore found some images
        if tr_track.lr_track:
            lr_track = tr_track.lr_track

            best_lr_track_item = get_best_lr_track_item(lr_track.plate_trace, image_ts_list)

            if not best_lr_track_item:
                eventlog.log_warning('fail: define best-ts')
                return fail(result=None)

            speed = tr_track.ra_track.get_median_speed_kmph() if tr_track.ra_track else None
            length = tr_track.ra_track.length if tr_track.ra_track else 0

            image_ts = best_lr_track_item.ts
            geodata = self._gpsd.get_geodata(image_ts)

            fact = _models.Fact(
                plate=lr_track.plate,
                direction=lr_track.direction,
                speed=speed,
                length=length,
                lane=0,  # will be defined later
                plate_rect=best_lr_track_item.plate_rect,
                catch_ts=catch_ts,
                lost_ts=lost_ts,
                geodata=geodata)

        elif tr_track.ra_track:
            ra_track = tr_track.ra_track

            best_ra_ts_rect = get_best_ra_ts_and_rect(ra_track.trace, image_ts_list)
            if not best_ra_ts_rect:
                eventlog.log_warning('fail: define best-ts from ra-track (len=%s)' % len(ra_track.trace))
                return fail(result=None)

            image_ts, fake_radar_plate_rect = best_ra_ts_rect
            speed = tr_track.ra_track.get_median_speed_kmph()
            geodata = self._gpsd.get_geodata(image_ts)

            fact = _models.Fact(
                plate=None,
                direction=1,  # TODO:
                speed=speed,
                length=tr_track.ra_track.length,
                lane=0,  # will be defined later
                plate_rect=fake_radar_plate_rect,
                catch_ts=catch_ts,
                lost_ts=lost_ts,
                geodata=geodata)

        else:
            return fail(result=None)

        call_chain = self._ms.search_image_url(image_ts)
        call_chain.addCallbacks(
            callback=self._handle_ms_load_image_done, callbackArgs=(image_ts, fact, tr_track),
            errback=self._handle_ms_load_image_fail, errbackArgs=(tr_track, catch_ts, lost_ts))

        return call_chain

    # noinspection PyUnusedLocal
    def _handle_ms_search_ts_fail(self, failure, tr_track, catch_ts, lost_ts):
        """
        :type failure: twisted.python.failure.Failure
        :type tr_track: _models.TrTrack
        :type catch_ts: int
        :type lost_ts: int
        :rtype: None
        """
        if failure.check([MediaStoreError]):
            eventlog.log_debug('fail: search-image-ts')
        else:
            eventlog.log_debug('fail: search-image-ts (%s)' % str(failure))

        # fallback to an image from the lr-track
        self._check_lr_image(tr_track, catch_ts, lost_ts)

    def _handle_ms_load_image_done(self, image_url, image_ts, fact, tr_track):
        """
        :type image_url: str
        :type image_ts: int
        :type fact: _models.Fact
        :type tr_track: _models.TrTrack
        :rtype: Deferred
        """
        scene_ref = _models.LrImageRef(ts=image_ts, url=image_url, codec='JPEG')

        lane_set = self._lane_detector.get_lane2(
            fact.plate_rect.left, fact.plate_rect.top,
            fact.plate_rect.right, fact.plate_rect.bottom)

        if len(lane_set) == 1:
            fact.set_lane(lane_set.pop())

        call_chain = self._raas.analyze_scene(scene_ref, fact.plate_rect)
        call_chain.addCallbacks(
            callback=self._handle_raas_done, callbackArgs=(fact, scene_ref, tr_track),
            errback=self._handle_raas_fail, errbackArgs=(fact, scene_ref, tr_track))

        return call_chain

    # noinspection PyUnusedLocal
    def _handle_ms_load_image_fail(self, failure, tr_track, catch_ts, lost_ts):
        """
        :type tr_track: _models.TrTrack
        :type catch_ts: int
        :typee lost_ts: int
        :rtype: None
        """
        eventlog.log_debug('fail: load image url, trying to use from lr-track')
        self._check_lr_image(tr_track, catch_ts, lost_ts)

    def _handle_raas_done(self, raas_scene, fact, scene_ref, tr_track):
        """
        :type raas_scene: _models.RaasScene
        :type fact: _models.Fact
        :type scene_ref: _models.LrImageRef
        :type tr_track: _models.TrTrack
        :rtype: (_models.Fact, _models.FactImages)
        """
        # The raas_scene is always empty if Raas is not defined in configuration.
        raas_plate = raas_scene.try_get_first_scene_item_plate()
        if raas_plate:
            plate = _models.Plate(raas_plate.validity, raas_plate.text, raas_plate.template)
            fact.set_plate(plate)

        start = time.time()
        fact_images = self._im_proc.process(scene_ref, tr_track, fact)
        eventlog.log_debug(time.time() - start)

        return fact, fact_images

    # noinspection PyUnusedLocal
    def _handle_raas_fail(self, failure, fact, scene_ref, tr_track):
        """
        :type failure: twisted.python.failure.Failure
        :type fact: _models.Fact
        :type scene_ref: _models.LrImageRef
        :type tr_track: _models.TrTrack
        :rtype: (_models.Fact, _models.FactImages)
        """
        if failure.check([RaasError]):
            eventlog.log_debug('fail: analayze image by raas')
        else:
            eventlog.log_debug('fail: analayze image by raas (%s)' % str(failure))

        start = time.time()
        fact_images = self._im_proc.process(scene_ref, tr_track, fact)
        eventlog.log_debug(time.time() - start)

        return fact, fact_images
