import errno
import os

import edge.eventlog as eventlog

import _models
from _args import SceneEventArgs, TrackEventArgs, ImageQueueArgs
from _stat_counters import TrackEventCounters, SceneEventCounters
from copy import deepcopy
from edge.queue import ImageFileQueue
from _models import LrImageRef
import edge.fs_helpers as fs_helpers


class _ToutEntry:
    """
    :type tr_track: _models.TrTrack
    :type rm_delayed_call: twisted.internet.interfaces.IDelayedCall
    """
    def __init__(self, track, image_ref, rm_delayed_call):
        """
        :type track: _models.TrTrack
        :type image_ref: ImageRef | None
        :type rm_delayed_call: twisted.internet.interfaces.IDelayedCall
        """
        self.tr_track = track
        self.image_ref = image_ref
        self.rm_delayed_call = rm_delayed_call


class ITrackEventConsumer:
    def __init__(self):
        pass

    def consume_track(self, tr_track, plate_text=None):
        """
        :type tr_track: _models.TrTrack
        :type plate_text : str | None
        """
        raise NotImplemented


class ISceneEventConsumer:
    def __init__(self):
        pass

    def consume_scene(self, scene_image, tr_scene):
        """
        :type scene_image: _models.SceneImage
        :type tr_scene: _models.TrScene
        """
        raise NotImplemented


class TrackEventConsumer(ITrackEventConsumer):
    """
    :type _plate_text_to_tout_entry_map: dict[str, _ToutEntry]
    :type _debounce_tout: int
    """

    def __init__(self, twisted_reactor, track_event_args, stat_counters, image_store_args):
        """
        :type twisted_reactor: twisted.internet.interfaces.IReactorTime
        :type track_event_args: TrackEventArgs
        :type stat_counters: TrackEventCounters
        :type image_store_args: ImageQueueArgs | None
        """
        ITrackEventConsumer.__init__(self)

        self._stat_counters = stat_counters

        self._event_dirs = track_event_args.event_dirs
        for event_dirpath in self._event_dirs:
            fs_helpers.silent_mkdir_r(event_dirpath)

        self._event_file_prefix = os.getpid()
        self._event_file_seq = 0
        self._plate_text_to_tout_entry_map = {}
        self._debounce_tout = track_event_args.debounce_tout
        self._twisted_reactor = twisted_reactor

        self.image_file_queue = None
        if image_store_args:
            self.image_file_queue = ImageFileQueue(twisted_reactor,
                                                   None,
                                                   image_store_args.endpoint,
                                                   image_store_args.min_file_count,
                                                   image_store_args.max_file_count,
                                                   image_store_args.log_level)

    def _handle_track_debounce_tout(self, plate_text):
        """
        :type plate_text: str
        """
        tout_entry = self._plate_text_to_tout_entry_map.get(plate_text)
        try:
            self._consume_track(tout_entry.tr_track, tout_entry.image_ref)
        finally:
            self._plate_text_to_tout_entry_map.pop(plate_text, None)

    def consume_track(self, tr_track, plate_text=None):
        """
        :type tr_track: _models.TrTrack
        :type plate_text : str | None
        """
        if plate_text is None:
            # ra_track
            self._consume_track(tr_track, None)
            return

        entry = self._plate_text_to_tout_entry_map.get(plate_text, None)
        if entry:
            self._stat_counters.inc_dublicate_count()
            # longest track - is better
            best_track = None
            if len(tr_track.lr_track_jnode['PlateTrace']) > len(entry.tr_track.lr_track_jnode['PlateTrace']):
                best_track = deepcopy(tr_track)
            else:
                best_track = deepcopy(entry.tr_track)

            # entry.tr_track always appears earlier than tr_track
            best_track.lr_track_jnode['PlateTrace'] = entry.tr_track.lr_track_jnode['PlateTrace'] + \
                                                      tr_track.lr_track_jnode['PlateTrace']

            best_track.lr_track_jnode['Id'] = list(set(entry.tr_track.lr_track_jnode['Id'] +
                                                       tr_track.lr_track_jnode['Id']))
            # add new image to store
            image_jnode = best_track.lr_track_jnode.get('Image')
            image_ref = None
            if image_jnode is not None and self.image_file_queue is not None:
                image_ref = LrImageRef.deserialize(image_jnode)
                store_img_ref = self.image_file_queue.add_image_ref(image_ref)
                image_ref.url = store_img_ref.url

                # delete old one
                self.image_file_queue.remove_image(entry.image_ref)

            entry.image_ref = image_ref
            entry.tr_track = best_track
            # TODO : check ra_track's of entry.tr_track and tr_track (is it possible to have 2 r-track for 1 real car ?)
            entry.rm_delayed_call.reset(secondsFromNow=self._debounce_tout)

        #
        else:
            rm_delayed_call = self._twisted_reactor.callLater(
                self._debounce_tout, self._handle_track_debounce_tout, plate_text)

            image_jnode = tr_track.lr_track_jnode.get('Image')
            image_ref = None
            if image_jnode is not None:
                image_ref = LrImageRef.deserialize(image_jnode)
                store_img_ref = self.image_file_queue.add_image_ref(image_ref)
                image_ref.url = store_img_ref.url

            if plate_text is not None:
                self._plate_text_to_tout_entry_map[plate_text] = _ToutEntry(tr_track, image_ref, rm_delayed_call)

    def _consume_track(self, tr_track, lr_image_ref):
        """
        :type tr_track: _models.TrTrack
        :type lr_image_ref: _models.LrImageRef | None
        """
        self._stat_counters.inc_track_count(tr_track)

        # get the event's filename.
        event_filename = '%s-%d' % (self._event_file_prefix, self._event_file_seq)
        self._event_file_seq += 1

        # send the event to target directories.
        # noinspection PyBroadException
        try:
            event_args = _models.TrTrack.serialize(tr_track, lr_image_ref).encode('utf-8')

            for event_dir in self._event_dirs:
                try_count = 2
                track_dir_present = True
                event_abspath = os.path.join(event_dir, event_filename)
                while try_count > 0 and track_dir_present:
                    try:
                        with open(event_abspath, 'w') as event_file:
                            event_file.write(event_args)

                        break

                    except IOError as e:
                        if e.errno == errno.ENOENT:
                            track_dir_present = False

                    if not track_dir_present:
                        fs_helpers.silent_mkdir_r(event_dir)
                        track_dir_present = True
                    try_count -= 1

        except:
            eventlog.log_debug_last_exc()


class SceneEventConsumer(ISceneEventConsumer):
    def __init__(self, ident, scene_event_args, stat_counters):
        """
        :type ident: int
        :type scene_event_args: SceneEventArgs
        :type stat_counters: SceneEventCounters
        """
        ISceneEventConsumer.__init__(self)
        eventlog.log_debug('ident:%s dirs:%s' % (ident, repr(scene_event_args.event_dirs)))

        self._stat_counters = stat_counters

        self._ident = ident
        self._event_dirs = scene_event_args.event_dirs
        self._skip_rate = scene_event_args.skip_rate
        self._frame_counter = 0

        map(fs_helpers.silent_mkdir_r, self._event_dirs)

    def consume_scene(self, scene_image, tr_scene):
        """
        :type scene_image: _models.SceneImage
        :type tr_scene: _models.TrScene
        """

        self._frame_counter += 1
        self._frame_counter %= self._skip_rate

        if self._frame_counter:
            return

        # get the event's filename.
        event_filename = str(scene_image.image_ts)

        # send the event to target directories.
        # noinspection PyBroadException
        try:
            event_args = tr_scene.serialize(scene_image)
            for event_dir in self._event_dirs:
                try_count = 2
                scene_dir_present = True
                event_abspath = os.path.join(event_dir, event_filename)
                while try_count > 0 and scene_dir_present:
                    try:
                        with open(event_abspath, 'w') as event_file:
                            event_file.write(event_args)

                        self._stat_counters.inc_scene_count(self._ident)
                        break

                    except IOError as x:
                        if x.errno == errno.ENOENT:
                            scene_dir_present = False

                    if not scene_dir_present:
                        fs_helpers.silent_mkdir_r(event_dir)
                        scene_dir_present = True

                    try_count -= 1

        except:
            eventlog.log_debug_last_exc()
