import edge.eventlog as eventlog

import _event_consumer
import _models
from _args import KernelArgs
from _lr_ra_oid_match_tout_store import LrRaOidMatchToutStore
from _ra_track_tout_store import RaTrackToutStore
from _react_producer import SceneReactProducer
from _react_producer import TrackReactProducer
from _stat_counters import StatCounters


class BaseKernel:
    def __init__(self, scene_react_producer, track_react_producer, scene_event_consumers_list,
                 track_event_consumer, lr_ra_oid_match_tout_store, ra_track_tout_store, min_ra_track_len, lr_upscale):

        self._min_ra_track_len = min_ra_track_len

        self._scene_react_producer = scene_react_producer
        self._scene_react_producer.set_addscene_react_handler(self._handle_scene_react)

        self._track_react_producer = track_react_producer
        self._track_react_producer.set_addtrack_react_handler(self._handle_track_react)

        self._lr_ra_oid_match_tout_store = lr_ra_oid_match_tout_store
        self._ra_track_tout_store = ra_track_tout_store
        self._ra_track_tout_store.set_ra_track_tout_handler(self._handle_ra_track_tout)

        self._scene_event_consumers_list = scene_event_consumers_list
        self._track_event_consumer = track_event_consumer

        self._lr_upscale = lr_upscale

    def _handle_scene_react(self, scene_image, lr_scene, ra_scene):
        """
        :type scene_image: _models.SceneImage | None
        :type lr_scene: _models.LrScene | None
        :type ra_scene: _models.RaScene | None
        """
        # noinspection PyBroadException
        try:
            # Update _ra_track_tout_store
            if ra_scene:
                self._ra_track_tout_store.add_ra_scene(ra_scene)

            # Create the tr_scene from an avaliable lr_scene and ra_scene.
            if lr_scene and ra_scene:
                tr_scene = _models.TrScene.create_lr_ra(self._lr_upscale, lr_scene.scene, ra_scene.scene)
                self._lr_ra_oid_match_tout_store.add(lr_scene.source_id, tr_scene.tr_items)

            elif lr_scene:
                tr_scene = _models.TrScene.create_lr(self._lr_upscale, lr_scene.scene)

            elif ra_scene:
                tr_scene = _models.TrScene.create_ra(self._lr_upscale, ra_scene.scene)

            else:
                tr_scene = _models.TrScene.create_empty(self._lr_upscale)

            if scene_image:
                for consumer in self._scene_event_consumers_list:
                    consumer.consume_scene(scene_image, tr_scene)

        except:
            eventlog.log_debug_last_exc()

    def _handle_track_react(self, lr_track):
        """
        :type lr_track: _models.LrTrack
        """
        ra_track = self._lr_ra_oid_match_tout_store.try_get_most_mached_ra_track(
            lr_track.source_id, lr_track.oids,
            lambda ra_oid: self._ra_track_tout_store.try_get_ra_track(ra_oid))

        if ra_track:
            if len(ra_track) >= self._min_ra_track_len:
                self._ra_track_tout_store.remove_ra_track(ra_track.oid)
                tr_track = _models.TrTrack(lr_track.src_jnode, ra_track)

            else:
                tr_track = _models.TrTrack(lr_track.src_jnode, None)

        else:
            tr_track = _models.TrTrack(lr_track.src_jnode, None)

        self._track_event_consumer.consume_track(tr_track, plate_text=lr_track.plate_text)

    def _handle_ra_track_tout(self, ra_track):
        """
        :type ra_track: _models.RaTrack
        """
        if len(ra_track) >= self._min_ra_track_len:
            tr_track = _models.TrTrack(None, ra_track)
            self._track_event_consumer.consume_track(tr_track)


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, kernel_args):
        """
        :type kernel_args: KernelArgs
        """
        stat_counters = StatCounters(
            twisted_reactor,
            scene_consumer_count=len(kernel_args.scene_event_args_list),
            stat_filepath=kernel_args.stat_counters_args.stat_filepath,
            stat_tout_sec=kernel_args.stat_counters_args.stat_tout_sec)

        track_react_producer = TrackReactProducer(
            twisted_reactor, kernel_args.track_react_args, stat_counters.track_react)

        scene_react_producer = SceneReactProducer(
            twisted_reactor, kernel_args.scene_react_args, stat_counters.scene_react)

        scene_event_consumers = list()
        ident = 0
        for scene_event_args in kernel_args.scene_event_args_list:
            scene_event_consumers.append(
                _event_consumer.SceneEventConsumer(ident, scene_event_args, stat_counters.scene_event))
            ident += 1

        track_event_consumer = _event_consumer.TrackEventConsumer(
            twisted_reactor,
            kernel_args.track_event_args, stat_counters.track_event, kernel_args.image_store_args)

        ra_track_tout_store = RaTrackToutStore(twisted_reactor, kernel_args.ra_track_store_tout)
        lr_ra_oid_match_tout_store = LrRaOidMatchToutStore(twisted_reactor, kernel_args.lr_ra_oid_match_store_tout)

        BaseKernel.__init__(
            self,
            scene_react_producer,
            track_react_producer,
            scene_event_consumers,
            track_event_consumer,
            lr_ra_oid_match_tout_store,
            ra_track_tout_store,
            kernel_args.min_ra_track_len,
            kernel_args.lr_upscale_args)

        self._twisted_reactor = twisted_reactor
        self._stat_counters = stat_counters

    def run(self):

        self._twisted_reactor.run()
