import sys
import ujson

import edge.eventlog as eventlog

from edge.types import ImageRef
DIR_UNKNOWN = 1
DIR_APPROACHING = 11
DIR_RECEDING = 12


class RaSceneItemRaw:
    def __init__(self, x, y, z, vx, vy, vz):
        """
        :type x: float
        :type y: float
        :type z: float
        :type vx: float
        :type vy: float
        :type vz: float
        """
        self.x = x
        self.y = y
        self.z = z
        self.vx = vx
        self.vy = vy
        self.vz = vz

    @staticmethod
    def deserialize(ra_scene_item_raw_jnode):
        x = float(ra_scene_item_raw_jnode['x'])
        y = float(ra_scene_item_raw_jnode['y'])
        z = float(ra_scene_item_raw_jnode['z'])

        vx = float(ra_scene_item_raw_jnode['vx'])
        vy = float(ra_scene_item_raw_jnode['vy'])
        vz = float(ra_scene_item_raw_jnode['vz'])

        ra_scene_item_raw = RaSceneItemRaw(x, y, z, vx, vy, vz)
        return ra_scene_item_raw

    def get_direction(self):
        delta = 1e-5
        if -delta < self.vz < +delta:
            return DIR_UNKNOWN

        elif self.vz > 0.0:
            return DIR_APPROACHING

        else:
            return DIR_RECEDING


class RaSceneItemImg:
    X_INDEX = 0
    Y_INDEX = 1

    def __init__(self, x, y):
        """
        :type x: float
        :type y: float
        """
        self.x = x
        self.y = y

    @staticmethod
    def deserialize(ra_scene_item_img_jnode):
        """
        :type ra_scene_item_img_jnode: dict
        :rtype: RaSceneItemImg
        """
        img_x = float(ra_scene_item_img_jnode[RaSceneItemImg.X_INDEX])
        img_y = float(ra_scene_item_img_jnode[RaSceneItemImg.Y_INDEX])

        ra_scene_item_img = RaSceneItemImg(img_x, img_y)
        return ra_scene_item_img


class RaSceneItem:
    def __init__(self, oid, ts, length, speed, direction, img, stat):
        """
        :type oid: int
        :type ts: int
        :type length: int
        :type speed: int
        :type direction: int
        :type stat: int
        :type img: RaSceneItemImg
        """
        self.oid = oid
        self.ts = ts
        self.len = length
        self.speed = speed
        self.direction = direction
        self.img = img
        self.stat = stat

    @staticmethod
    def deserialize(ra_scene_item_jnode, scene_ts):
        """
        :type ra_scene_item_jnode: dict
        :type scene_ts: int
        :rtype: RaSceneItem
        """
        img = RaSceneItemImg.deserialize(ra_scene_item_jnode['img-lt'])

        if 0.0 < img.x < 1.0 and 0.0 < img.y < 1.0:
            stat = int(ra_scene_item_jnode['stat'])
            oid = ra_scene_item_jnode['oid']
            length = int(ra_scene_item_jnode['len'])
            speed = int(ra_scene_item_jnode['kmh'])
            direction = int(ra_scene_item_jnode['dir'])
            ra_scene_item = RaSceneItem(oid, scene_ts, length, speed, direction, img, stat)
            return ra_scene_item

        return None


class RaScene:
    _SCENE_VALID = 1
    _SCENE_INVALID = 2

    def __init__(self, ra_scene_items, status):
        """
        :type ra_scene_items: list of RaSceneItem
        :type status: int
        """
        self.scene = ra_scene_items
        self.status = status

    @staticmethod
    def deserialize(ra_scene_jnode):
        """
        :type ra_scene_jnode: dict
        :rtype: RaScene
        """
        if not ra_scene_jnode:
            raise ValueError

        expected_ts = int(ra_scene_jnode['strobe-ts'])
        status = int(ra_scene_jnode['stat'])
        img_scene_items = list()

        if status == RaScene._SCENE_VALID:
            all_scene_items = map(lambda ra_scene_item_jnode: RaSceneItem.deserialize(ra_scene_item_jnode, expected_ts),
                                  ra_scene_jnode['scene'])

            img_scene_items = filter(None, all_scene_items)

        ra_scene = RaScene(img_scene_items, status)
        return ra_scene


class RaTrack:
    def __init__(self, ra_scene_item):
        """
        :type ra_scene_item: RaSceneItem
        """
        self.catch_ts = sys.maxint
        self.lost_ts = 0
        self.trace = []
        self.max_length = 0

        self.oid = ra_scene_item.oid
        self.append(ra_scene_item)

    def __len__(self):
        return len(self.trace)

    def append(self, ra_scene_item):
        """
        :type ra_scene_item: RaSceneItem
        """
        if ra_scene_item.ts < self.catch_ts:
            self.catch_ts = ra_scene_item.ts

        if ra_scene_item.ts > self.lost_ts:
            self.lost_ts = ra_scene_item.ts

        if ra_scene_item.len > self.max_length:
            self.max_length = ra_scene_item.len

        self.trace.append({"Ts": ra_scene_item.ts,
                           "Xy": [ra_scene_item.img.x, ra_scene_item.img.y],
                           "S": ra_scene_item.speed})


class LrRect:
    LEFT_INDEX = 0
    TOP_INDEX = 1
    RIGHT_INDEX = 2
    BOTTOM_INDEX = 3

    def __init__(self, left, top, right, bottom):
        """
        :type left: float
        :type top: float
        :type right: float
        :type bottom: float
        """
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

    def upscale(self, lr_upscale):
        """
        :type lr_upscale: _kernel_param_set._LrUpscale
        :rtype: LrRect
        """
        width = self.right - self.left
        height = self.bottom - self.top

        center_x = (self.left + self.right) / 2.0
        center_y = (self.top + self.bottom) / 2.0

        new_upper_left_x = center_x - (width * lr_upscale.left)
        new_lower_right_x = center_x + (width * lr_upscale.right)

        new_upper_left_y = center_y - (height * lr_upscale.top)
        new_lower_right_y = center_y + (height * lr_upscale.bottom)

        scaled_rect = LrRect(new_upper_left_x, new_upper_left_y, new_lower_right_x, new_lower_right_y)
        return scaled_rect

    @staticmethod
    def deserialize(rect_jnode):
        """
        :type rect_jnode: list of float
        :rtype: LrRect
        """
        left = float(rect_jnode[LrRect.LEFT_INDEX])
        top = float(rect_jnode[LrRect.TOP_INDEX])
        right = float(rect_jnode[LrRect.RIGHT_INDEX])
        bottom = float(rect_jnode[LrRect.BOTTOM_INDEX])

        lr_rect = LrRect(left, top, right, bottom)
        return lr_rect

    @staticmethod
    def serialize(lr_rect):
        """
        :type lr_rect: LrRect
        :rtype: list of float
        """
        return [lr_rect.left, lr_rect.top, lr_rect.right, lr_rect.bottom]


def check_point_in_rect(point_x, point_y, lr_rect):
    """
    :type point_x: float
    :type point_y: float
    :type lr_rect: LrRect
    :rtype: bool
    """
    return lr_rect.left < point_x < lr_rect.right and lr_rect.top < point_y < lr_rect.bottom


class LrPlate:
    def __init__(self, validity, plate_text, plate_template, rect):
        """
        :type validity: int
        :type plate_text: str
        :type plate_template: int
        :type rect: LrRect
        """
        self.validity = validity
        self.plate_text = plate_text
        self.plate_template = plate_template
        self.rect = rect

    @staticmethod
    def deserialize(lr_plate_jnode):
        """
        :type lr_plate_jnode: dict
        :rtype: LrPlate
        """
        validity = lr_plate_jnode['V']
        plate_text = lr_plate_jnode['P']

        plate_template = lr_plate_jnode['T']
        rect = LrRect.deserialize(lr_plate_jnode['R'])

        lr_plate = LrPlate(validity, plate_text, plate_template, rect)
        return lr_plate


class LrSceneExtra:
    def __init__(self, image_ts, image_url, image_codec, image_width, image_height):
        """
        :type image_ts: int
        :type image_url: str
        :type image_codec: str
        :type image_width: int
        :type image_height: int
        """
        self.image_ts = image_ts
        self.image_url = image_url
        self.image_codec = image_codec
        self.image_width = image_width
        self.image_height = image_height

    @staticmethod
    def deserialize(lr_scene_extra_jnode):
        """
        :type lr_scene_extra_jnode: dict
        :rtype: LrSceneExtra
        """
        image_ts = int(lr_scene_extra_jnode['ImageTs'])
        image_url = lr_scene_extra_jnode['ImageUrl']
        image_codec = lr_scene_extra_jnode['ImageCodec']

        image_wh = lr_scene_extra_jnode['ImageWh']
        image_width = image_wh[0]
        image_height = image_wh[1]

        lr_scene_extra = LrSceneExtra(image_ts, image_url, image_codec, image_width, image_height)
        return lr_scene_extra


class LrImageRef(ImageRef):
    def __init__(self, ts=None, url=None, codec=None, width=None, height=None,  rect=None):
        """
        :type ts: int | None
        :type url: str | None
        :type codec: str | None
        :type width: int | None
        :type height: int | None
        :type rect: list of float
        """
        ImageRef.__init__(self, ts, url, codec, width, height)
        self.rect = rect

    @staticmethod
    def deserialize(lr_best_image_jnode):
        """
        :type lr_best_image_jnode: dict
        :rtype: LrImageRef
        """
        ts = lr_best_image_jnode['Ts']
        url = str(lr_best_image_jnode['Url'])
        left, top, right, bottom = lr_best_image_jnode['R']

        if left < 0.0:
            left = 0.0

        if top < 0.0:
            top = 0.0

        if right > 1.0:
            right = 1.0

        if bottom > 1.0:
            bottom = 1.0

        plate_rect = [left, top, right, bottom]

        raw_wh = lr_best_image_jnode['Wh']
        w = int(raw_wh[0])
        h = int(raw_wh[1])
        codec = lr_best_image_jnode['Codec']
        lr_image = LrImageRef(ts=ts, url=url, codec=codec, width=w, height=h, rect=plate_rect)
        return lr_image


class LrSceneItem:
    def __init__(self, oid, direction, best_variant, history, scene_ts):
        """
        :type oid: int
        :type direction: int
        :type best_variant: LrPlate
        :type history: list
        :type scene_ts: int
        """
        self.oid = oid
        self.direction = direction
        self.best_variant = best_variant
        self.history = history

        self.ts = scene_ts

    @staticmethod
    def deserialize(lr_scene_item_jnode, scene_ts):
        """
        :type lr_scene_item_jnode: dict
        :type scene_ts: int
        :rtype: LrSceneItem
        """
        oid = int(lr_scene_item_jnode['Id'])
        direction = lr_scene_item_jnode['Dir']
        plate = LrPlate.deserialize(lr_scene_item_jnode['Plate'])
        plate_trace = lr_scene_item_jnode['PlateTrace']

        lr_scene_item = LrSceneItem(oid, direction, plate, plate_trace, scene_ts)
        return lr_scene_item


class LrScene:
    def __init__(self, source_id, scene, extra, stat):
        """
        :type source_id: list of int
        :type scene: list of LrSceneItem
        :type extra: LrSceneExtra
        :type stat: int
        """
        self.source_id = source_id
        self.scene = scene
        self.extra = extra
        self.stat = stat

    @staticmethod
    def deserialize(lr_scene_jnode):
        """
        :type lr_scene_jnode: dict
        :rtype: LrScene
        """
        stat = lr_scene_jnode['Stat']
        source_jnode = lr_scene_jnode['Source']

        extra_jnode = lr_scene_jnode.get('Extra')
        if not extra_jnode:
            eventlog.log_debug("fail: extra-jnode is absent")
            raise ValueError

        extra = LrSceneExtra.deserialize(extra_jnode)

        lr_scene_items_jnode = lr_scene_jnode['Scene']
        lr_scene_items = map(lambda lr_scene_item_jnode: LrSceneItem.deserialize(lr_scene_item_jnode, extra.image_ts),
                             lr_scene_items_jnode)

        lr_scene = LrScene(source_jnode, lr_scene_items, extra, stat)
        return lr_scene


class LrTrackItem:
    def __init__(self, ts, weight, rect):
        """
        :type ts: int
        :type weight: int
        :type rect: list of double
        """
        self.Ts = ts
        self.weight = weight
        self.rect = rect

    @staticmethod
    def deserialize(lr_track_item_jnode):
        """
        :type lr_track_item_jnode: dict
        :rtype: LrTrackItem
        """
        ts = lr_track_item_jnode['Ts']
        weight = lr_track_item_jnode['W']
        rect = lr_track_item_jnode['R']

        lr_track_item = LrTrackItem(ts, weight, rect)
        return lr_track_item


class LrTrack:
    def __init__(self, source_id, oids, src_jnode, plate_text,  src_filename=None):
        """
        :type source_id: list of int
        :type oids: list of int
        :type src_jnode: dict
        :type plate_text: str
        :type src_filename: str
        """
        self.source_id = source_id
        self.oids = oids
        self.src_jnode = src_jnode
        self.src_file_name = src_filename
        self.plate_text = plate_text

    @staticmethod
    def deserialize(lr_track_jnode, src_filename=None):
        """
        :type lr_track_jnode: dict
        :type src_filename: str | None
        :rtype: LrTrack
        """
        source = lr_track_jnode['Source']
        oids = lr_track_jnode['Id']
        plate_text = lr_track_jnode['Plate']['P']

        lr_track = LrTrack(source, oids, lr_track_jnode, plate_text, src_filename)
        return lr_track


class SceneImage:
    def __init__(self, image_ts, image_url, image_codec, image_width, image_height):
        """
        :type image_ts: int
        :type image_url: str
        :type image_codec: str
        :type image_width: int
        :type image_height: int
        """
        self.image_ts = image_ts
        self.image_url = image_url
        self.image_codec = image_codec
        self.image_width = image_width
        self.image_height = image_height

    @staticmethod
    def deserialize_mg_scene(mg_scene_jnode):
        """
        :type mg_scene_jnode: dict
        :rtype: SceneImage
        """
        image_ts = int(mg_scene_jnode['Env']['ExpStartTs'])

        frame_jnode = mg_scene_jnode['Frame']
        image_url = frame_jnode['Url']
        image_codec = frame_jnode['Codec']
        image_width = frame_jnode['W']
        image_height = frame_jnode['H']

        mg_scene = SceneImage(image_ts, image_url, image_codec, image_width, image_height)
        return mg_scene

    @staticmethod
    def deserialize_mc_scene(mc_scene_jnode):
        """
        :type mc_scene_jnode: dict
        :rtype: SceneImage
        """
        status_success = 1
        # status_fail = 2
        if mc_scene_jnode['Status'] != status_success:
            raise ValueError('Failed to deserialize mc_scene')

        image_ts = int(mc_scene_jnode['Ts'])

        image_url = mc_scene_jnode['Url']
        image_codec = mc_scene_jnode['Codec']
        image_width, image_height = mc_scene_jnode['SizeWh']

        mg_scene = SceneImage(image_ts, image_url, image_codec, image_width, image_height)
        return mg_scene


class TrSceneItem:
    def __init__(self, lr_scene_item, ra_scene_item):
        """
        :type lr_scene_item: LrSceneItem | None
        :type ra_scene_item: RaSceneItem | None
        """
        if not lr_scene_item and not ra_scene_item:
            raise ValueError

        self.lr_scene_item = lr_scene_item
        self.ra_scene_item = ra_scene_item


class TrScene:
    def __init__(self, lr_upscale, tr_items, lr_items, ra_items):
        """
        :type lr_upscale: _kernel_param_set._LrUpscale
        :type tr_items: list of TrSceneItem
        :type lr_items: list of LrSceneItem
        :type ra_items: list of RaSceneItem
        """
        self._lr_upscale = lr_upscale

        self.tr_items = tr_items  # list of matched lr_scene_item and ra_scene_item
        self.lr_items = lr_items  # list of not matched lr_scene_items
        self.ra_items = ra_items  # list of not matched ra_scene_items

    @staticmethod
    def create_lr_ra(lr_upscale, lr_scene_items, ra_scene_items):
        """
        :type lr_upscale: _kernel_param_set._LrUpscale
        :type lr_scene_items: list of LrSceneItem
        :type ra_scene_items: list of RaSceneItem
        :rtype: TrScene
        """
        unused_lr_scene_items_set = set(map(lambda lr_item: lr_item.oid, lr_scene_items))
        unused_ra_scene_items_set = set(map(lambda ra_item: ra_item.oid, ra_scene_items))

        tr_scene_tr_items = []
        for lr_scene_item in lr_scene_items:
            # reshape vehicle's plate react -> body rect

            match_rect = lr_scene_item.best_variant.rect.upscale(lr_upscale)
            for ra_scene_item in ra_scene_items:
                if not check_point_in_rect(ra_scene_item.img.x, ra_scene_item.img.y, match_rect):
                    continue

                unused_lr_scene_items_set.discard(lr_scene_item.oid)
                unused_ra_scene_items_set.discard(ra_scene_item.oid)

                tr_scene_item = TrSceneItem(lr_scene_item, ra_scene_item)
                tr_scene_tr_items.append(tr_scene_item)

        # Construct a list from those elements of iterable for which function returns true.
        tr_scene_lr_items = filter(lambda lr_item: lr_item.oid in unused_lr_scene_items_set, lr_scene_items)
        tr_scene_ra_items = filter(lambda ra_item: ra_item.oid in unused_ra_scene_items_set, ra_scene_items)

        tr_scene = TrScene(lr_upscale, tr_scene_tr_items, tr_scene_lr_items, tr_scene_ra_items)
        return tr_scene

    @staticmethod
    def create_lr(lr_upscale, lr_scene_items):
        """
        :type lr_upscale: _kernel_param_set._LrUpscale
        :type lr_scene_items: list of LrSceneItem
        :rtype: TrScene
        """
        tr_scene = TrScene(lr_upscale, [], lr_scene_items, [])
        return tr_scene

    @staticmethod
    def create_ra(lr_upscale, ra_scene_items):
        """
        :type lr_upscale: _kernel_param_set._LrUpscale
        :type ra_scene_items: list of RaSceneItem
        :rtype: TrScene
        """
        tr_scene = TrScene(lr_upscale, [], [], ra_scene_items)
        return tr_scene

    @staticmethod
    def create_empty(lr_upscale):
        """
        :type lr_upscale: _kernel_param_set._LrUpscale
        :rtype: TrScene
        """
        tr_scene = TrScene(lr_upscale, [], [], [])
        return tr_scene

    @staticmethod
    def _serialize_lr_scene_item(lr_scene_item):
        """
        :type lr_scene_item: LrSceneItem
        """
        lr_jnode = {
            'Id': lr_scene_item.oid,
            'Dir': lr_scene_item.direction,
            'Plate': {
                'P': lr_scene_item.best_variant.plate_text,
                'R': lr_scene_item.best_variant.rect.serialize(lr_scene_item.best_variant.rect)
            },
            'PlateTrace': lr_scene_item.history
        }

        return lr_jnode

    @staticmethod
    def _serialize_ra_scene_item(ra_scene_item):
        """
        :type ra_scene_item: RaSceneItem
        """
        ra_jnode = {
            'oid': ra_scene_item.oid,
            'len': ra_scene_item.len,
            'speed': ra_scene_item.speed,
            'dir': ra_scene_item.direction,
            'xy': [ra_scene_item.img.x, ra_scene_item.img.y]
        }

        return ra_jnode

    def serialize(self, scene_image):
        """
        :type scene_image: _models.SceneImage
        :rtype: str
        """

        tr_scene_jnode = {
            'Image': {
                'Ts': scene_image.image_ts,
                'Url': scene_image.image_url,
                'Codec': scene_image.image_codec,
                'Wh': [scene_image.image_width, scene_image.image_height]
            },
            'LrUpscaleLtrb': self._lr_upscale.serialize(),
            'LrRa': map(lambda tr_item: {'Lr': TrScene._serialize_lr_scene_item(tr_item.lr_scene_item),
                                         'Ra': TrScene._serialize_ra_scene_item(tr_item.ra_scene_item)},
                        self.tr_items),
            'Lr': map(TrScene._serialize_lr_scene_item, self.lr_items),
            'Ra': map(TrScene._serialize_ra_scene_item, self.ra_items)
        }

        #
        # https://pypi.python.org/pypi/ujson
        #
        # ensure_ascii limits output to ASCII and escapes all extended characters above 127. Default is true.
        # If your end format supports UTF-8 setting this option to false is highly recommended to save space.
        raw_tr_scene = ujson.dumps(tr_scene_jnode, ensure_ascii=False)
        return raw_tr_scene


class TrTrack:
    def __init__(self, lr_track_jnode, ra_track):
        """
        :type lr_track_jnode: dict | None
        :type ra_track: RaTrack | None
        """
        if not lr_track_jnode and not ra_track:
            raise ValueError

        self.lr_track_jnode = lr_track_jnode
        self.ra_track = ra_track

    @staticmethod
    def serialize(tr_track, lr_image_ref):
        """
        :type tr_track: TrTrack
        :type lr_image_ref: LrImageRef
        :rtype: str
        """
        tr_track_jnode = dict()
        if tr_track.lr_track_jnode:
            tr_track_jnode['Lr'] = tr_track.lr_track_jnode
            # change url to url from tr store
            if lr_image_ref is not None:
                tr_track_jnode['Lr']['Image']['Url'] = lr_image_ref.url
                tr_track_jnode['Lr']['Image']['R'] = lr_image_ref.rect

        ra_track = tr_track.ra_track
        if tr_track.ra_track:
            tr_track_jnode['Ra'] = {"CatchTs": ra_track.catch_ts, "LostTs": ra_track.lost_ts,
                                    "Length": ra_track.max_length, "Trace": ra_track.trace}

        raw_tr_track = ujson.dumps(tr_track_jnode)
        return raw_tr_track
